@@ -198,8 +198,6 @@ base_configs="$base_configs --with-extra-charsets=complex "
 base_configs="$base_configs --enable-thread-safe-client "
 base_configs="$base_configs --with-big-tables $maintainer_mode"
 base_configs="$base_configs --with-plugin-aria --with-aria-tmp-tables --with-plugin-s3=STATIC"
-# Following is to get tokudb to work
-base_configs="$base_configs --with-jemalloc=NO"
 
 if test -d "$path/../cmd-line-utils/readline"
 then
@@ -303,7 +301,7 @@ gcov_compile_flags="$gcov_compile_flags -DMYSQL_SERVER_SUFFIX=-gcov -DHAVE_gcov"
 
 #
 # The following plugins doesn't work on 32 bit systems
-disable_64_bit_plugins="--without-plugin-tokudb --without-plugin-rocksdb"
+disable_64_bit_plugins="--without-plugin-rocksdb"
 
 
 # GCC4 needs -fprofile-arcs -ftest-coverage on the linker command line (as well
