@@ -101,8 +101,9 @@ enum options_client
   OPT_SSL_CRL, OPT_SSL_CRLPATH,
   OPT_IGNORE_DATA,
   OPT_PRINT_ROW_COUNT, OPT_PRINT_ROW_EVENT_POSITIONS,
-  OPT_CHECK_IF_UPGRADE_NEEDED,
   OPT_SHUTDOWN_WAIT_FOR_SLAVES,
+  OPT_COPY_S3_TABLES,
+  OPT_PRINT_TABLE_METADATA,
   OPT_MAX_CLIENT_OPTION /* should be always the last */
 };
 
@@ -125,3 +126,13 @@ enum options_client
   Name of the performance schema database.
 */
 #define PERFORMANCE_SCHEMA_DB_NAME "performance_schema"
+
+/**
+  First mariadb version supporting the sys schema.
+*/
+#define FIRST_SYS_SCHEMA_VERSION 100600
+
+/**
+  Name of the sys schema database.
+*/
+#define SYS_SCHEMA_DB_NAME "sys"
