@@ -168,8 +168,7 @@ emb_advanced_command(MYSQL *mysql, enum enum_server_command command,
     arg_length= header_length;
   }
 
-  result= dispatch_command(command, thd, (char *) arg, arg_length, FALSE,
-                           FALSE);
+  result= dispatch_command(command, thd, (char *) arg, arg_length);
   thd->cur_data= 0;
   thd->mysys_var= NULL;
 
@@ -660,7 +659,7 @@ void end_embedded_server()
 }
 
 
-void init_embedded_mysql(MYSQL *mysql, ulong client_flag)
+void init_embedded_mysql(MYSQL *mysql, int client_flag)
 {
   THD *thd = (THD *)mysql->thd;
   thd->mysql= mysql;
@@ -680,7 +679,7 @@ void init_embedded_mysql(MYSQL *mysql, ulong client_flag)
   create_new_thread(), and prepare_new_connection_state().  This should
   be refactored to avoid code duplication.
 */
-void *create_embedded_thd(ulong client_flag)
+void *create_embedded_thd(int client_flag)
 {
   THD * thd= new THD(next_thread_id());
 
@@ -696,7 +695,7 @@ void *create_embedded_thd(ulong client_flag)
   thd->set_command(COM_SLEEP);
   thd->set_time();
   thd->init_for_queries();
-  thd->client_capabilities= client_flag | MARIADB_CLIENT_EXTENDED_METADATA;
+  thd->client_capabilities= client_flag;
   thd->real_id= pthread_self();
 
   thd->db= null_clex_str;
@@ -827,7 +826,7 @@ int check_embedded_connection(MYSQL *mysql, const char *db)
   /* acl_authenticate() takes the data from thd->net->read_pos */
   thd->net.read_pos= (uchar*)buf;
 
-  if (acl_authenticate(thd, (uint) (end - buf)))
+  if (acl_authenticate(thd, 0, end - buf))
   {
     my_free(thd->security_ctx->user);
     goto err;
@@ -1259,7 +1258,7 @@ bool Protocol_binary::write()
 bool Protocol::net_send_ok(THD *thd,
             uint server_status, uint statement_warn_count,
             ulonglong affected_rows, ulonglong id, const char *message,
-            bool, bool)
+            bool)
 {
   DBUG_ENTER("emb_net_send_ok");
   MYSQL_DATA *data;
