@@ -5,7 +5,6 @@ use warnings;
 use Text::Wrap;
 use Cwd;
 use My::Platform;
-use mtr_report;
 
 # 1. options to support:
 #       --xxx[=ARGS]
@@ -75,7 +74,7 @@ my %debuggers = (
     options => '-f -o {log} {exe} {args}',
   },
   rr => {
-    options => '_RR_TRACE_DIR={log} rr record {exe} {args}',
+    options => '_RR_TRACE_DIR={log} rr record {exe} {args} --loose-skip-innodb-use-native-aio',
     run => 'env',
     pre => sub {
       ::mtr_error('rr requires kernel.perf_event_paranoid <= 1')
@@ -106,10 +105,6 @@ EEE
 
 my %opts;
 my %opt_vals;
-my $debugger;
-my $boot_debugger;
-my $client_debugger;
-
 my $help = "\n\nOptions for running debuggers\n\n";
 
 for my $k (sort keys %debuggers) {
@@ -144,7 +139,7 @@ sub do_args($$$$$) {
   my $v = $debuggers{$k};
 
   # on windows mtr args are quoted (for system), otherwise not (for exec)
-  sub quote($) { $_[0] =~ /[; ]/ ? "\"$_[0]\"" : $_[0] }
+  sub quote($) { $_[0] =~ / / ? "\"$_[0]\"" : $_[0] }
   sub unquote($) { $_[0] =~ s/^"(.*)"$/$1/; $_[0] }
   sub quote_from_mtr($) { IS_WINDOWS() ? $_[0] : quote($_[0]) }
   sub unquote_for_mtr($) { IS_WINDOWS() ? $_[0] : unquote($_[0]) }
@@ -152,8 +147,7 @@ sub do_args($$$$$) {
   my %vars = (
     vardir => $::opt_vardir,
     exe => $$exe,
-    args => join(' ', map { quote_from_mtr $_ } @$$args,
-                 '--loose-debug-gdb', '--loose-skip-stack-trace'),
+    args => join(' ', map { quote_from_mtr $_ } @$$args, '--gdb'),
     input => $input,
     script => "$::opt_vardir/tmp/${k}init.$type",
     log => "$::opt_vardir/log/$type.$k",
@@ -164,9 +158,9 @@ sub do_args($$$$$) {
 
   my $script = join "\n", @params;
   if ($v->{script}) {
-    ::mtr_tonewfile($vars{script}, subst($v->{script}, %vars)."\n".$script);
+    ::mtr_tofile($vars{script}, subst($v->{script}, %vars)."\n".$script);
   } elsif ($script) {
-    mtr_error "$k is not using a script file, nowhere to write the script \n---\n$script\n---";
+    die "$k is not using a script file, nowhere to write the script \n---\n$script\n---\n";
   }
 
   my $options = subst($v->{options}, %vars);
@@ -191,61 +185,24 @@ sub help() { $help }
 sub fix_options(@) {
   my $re=join '|', keys %opts;
   $re =~ s/=s//g;
-  # FIXME: what is '=;'? What about ':s' to denote optional argument in register_opt()
   map { $_ . (/^--($re)$/ and '=;') } @_;
 }
 
 sub pre_setup() {
   my $used;
-  my %options;
-  my %client_options;
-  my %boot_options;
-
-  my $embedded= $::opt_embedded_server ? ' with --embedded' : '';
-
   for my $k (keys %debuggers) {
     for my $opt ($k, "manual-$k", "boot-$k", "client-$k") {
-      my $val= $opt_vals{$opt};
-      if ($val) {
+      if ($opt_vals{$opt})
+      {
         $used = 1;
         if ($debuggers{$k}->{pre}) {
           $debuggers{$k}->{pre}->();
           delete $debuggers{$k}->{pre};
         }
-        if ($opt eq $k) {
-          $options{$opt}= $val;
-          $client_options{$opt}= $val
-            if $embedded;
-        } elsif ($opt eq "manual-$k") {
-          $options{$opt}= $val;
-        } elsif ($opt eq "boot-$k") {
-          $boot_options{$opt}= $val;
-        } elsif ($opt eq "client-$k") {
-          $client_options{$opt}= $val;
-        }
       }
     }
   }
 
-  if ((keys %options) > 1) {
-    mtr_error "Multiple debuggers specified: ",
-        join (" ", map { "--$_" } keys %options);
-  }
-
-  if ((keys %boot_options) > 1) {
-    mtr_error "Multiple boot debuggers specified: ",
-        join (" ", map { "--$_" } keys %boot_options);
-  }
-
-  if ((keys %client_options) > 1) {
-    mtr_error "Multiple client debuggers specified: ",
-        join (" ", map { "--$_" } keys %client_options);
-  }
-
-  $debugger= (keys %options)[0];
-  $boot_debugger= (keys %boot_options)[0];
-  $client_debugger= (keys %client_options)[0];
-
   if ($used) {
     $ENV{ASAN_OPTIONS}= 'abort_on_error=1:'.($ENV{ASAN_OPTIONS} || '');
     ::mtr_error("Can't use --extern when using debugger") if $ENV{USE_RUNNING_SERVER};
@@ -261,20 +218,49 @@ sub pre_setup() {
 
 sub setup_boot_args($$$) {
   my ($args, $exe, $input) = @_;
-  do_args($args, $exe, $input, 'bootstrap', $boot_debugger)
-    if defined $boot_debugger;
+  my $found;
+
+  for my $k (keys %debuggers) {
+    if ($opt_vals{"boot-$k"}) {
+      die "--boot-$k and --$found cannot be used at the same time\n" if $found;
+
+      $found="boot-$k";
+      do_args($args, $exe, $input, 'bootstrap', $found);
+    }
+  }
 }
 
 sub setup_client_args($$) {
   my ($args, $exe) = @_;
-  do_args($args, $exe, IS_WINDOWS() ? 'NUL' : '/dev/null', 'client', $client_debugger)
-    if defined $client_debugger;
+  my $found;
+  my $embedded = $::opt_embedded_server ? ' with --embedded' : '';
+
+  for my $k (keys %debuggers) {
+    my @opt_names=("client-$k");
+    push @opt_names, $k if $embedded;
+    for my $opt (@opt_names) {
+      if ($opt_vals{$opt}) {
+        die "--$opt and --$found cannot be used at the same time$embedded\n" if $found;
+        $found=$opt;
+        do_args($args, $exe, IS_WINDOWS() ? 'NUL' : '/dev/null', 'client', $found);
+      }
+    }
+  }
 }
 
 sub setup_args($$$) {
   my ($args, $exe, $type) = @_;
-  do_args($args, $exe, IS_WINDOWS() ? 'NUL' : '/dev/null', $type, $debugger)
-    if defined $debugger;
+  my $found;
+
+  for my $k (keys %debuggers) {
+    for my $opt ($k, "manual-$k") {
+      if ($opt_vals{$opt}) {
+        die "--$opt and --$found cannot be used at the same time\n" if $found;
+        $found=$opt;
+        do_args($args, $exe, IS_WINDOWS() ? 'NUL' : '/dev/null', $type, $found);
+      }
+    }
+  }
 }
 
 1;
