@@ -216,8 +216,8 @@ static int check_lock(struct st_lock_list *list, const char* lock_type,
                 lock_type, where);
         return 1;
       }
-      if ((read_lock && data->type > TL_READ_NO_INSERT) ||
-          (!read_lock && data->type <= TL_READ_NO_INSERT))
+      if ((read_lock && data->type >= TL_FIRST_WRITE) ||
+          (!read_lock && data->type < TL_FIRST_WRITE))
       {
 	fprintf(stderr,
 		"Warning: Found %s lock in %s queue at %s: %s\n",
@@ -357,7 +357,8 @@ static void check_locks(THR_LOCK *lock, const char *where,
                data && count < MAX_LOCKS;
                data=data->next)
           {
-            if (data->type != TL_WRITE_CONCURRENT_INSERT)
+            if (data->type != TL_WRITE_CONCURRENT_INSERT &&
+                data->type != TL_WRITE_ALLOW_WRITE)
             {
               fprintf(stderr,
                       "Warning at '%s': Found TL_WRITE_CONCURRENT_INSERT lock mixed with other write lock: %d\n",
@@ -467,7 +468,7 @@ void thr_lock_delete(THR_LOCK *lock)
 
 void thr_lock_info_init(THR_LOCK_INFO *info, struct st_my_thread_var *tmp)
 {
-  if (tmp)
+  if (!tmp)
     tmp= my_thread_var;
   info->thread=    tmp->pthread_self;
   info->thread_id= tmp->id;
@@ -633,9 +634,10 @@ wait_for_lock(struct st_lock_list *wait, THR_LOCK_DATA *data,
   else
   {
     result= THR_LOCK_SUCCESS;
-    if (data->lock->get_status)
-      (*data->lock->get_status)(data->status_param,
-                                data->type == TL_WRITE_CONCURRENT_INSERT);
+    if (data->lock->get_status &&
+        (*data->lock->get_status)(data->status_param,
+                                  data->type == TL_WRITE_CONCURRENT_INSERT))
+        result= THR_LOCK_ABORTED;
     check_locks(data->lock,"got wait_for_lock", data->type, 0);
   }
   mysql_mutex_unlock(&data->lock->mutex);
@@ -763,9 +765,9 @@ thr_lock(THR_LOCK_DATA *data, THR_LOCK_INFO *owner, ulong lock_wait_timeout)
   DBUG_PRINT("lock",("data:%p  thread:%lu  lock:%p  type: %d",
                      data, (ulong) data->owner->thread_id,
                      lock, (int) lock_type));
-  check_locks(lock,(uint) lock_type <= (uint) TL_READ_NO_INSERT ?
+  check_locks(lock,(uint) lock_type < (uint) TL_FIRST_WRITE ?
 	      "enter read_lock" : "enter write_lock", lock_type, 0);
-  if ((int) lock_type <= (int) TL_READ_NO_INSERT)
+  if ((int) lock_type < (int) TL_FIRST_WRITE)
   {
     /* Request for READ lock */
     if (lock->write.data)
@@ -810,8 +812,8 @@ thr_lock(THR_LOCK_DATA *data, THR_LOCK_INFO *owner, ulong lock_wait_timeout)
 	if (lock_type == TL_READ_NO_INSERT)
 	  lock->read_no_write_count++;
 	check_locks(lock,"read lock with old write lock", lock_type, 0);
-	if (lock->get_status)
-	  (*lock->get_status)(data->status_param, 0);
+	if ((lock->get_status) && (*lock->get_status)(data->status_param, 0))
+          result= THR_LOCK_ABORTED;
 	statistic_increment(locks_immediate,&THR_LOCK_lock);
 	goto end;
       }
@@ -834,8 +836,8 @@ thr_lock(THR_LOCK_DATA *data, THR_LOCK_INFO *owner, ulong lock_wait_timeout)
       if (lock_type == TL_READ_NO_INSERT)
 	lock->read_no_write_count++;
       check_locks(lock,"read lock with no write locks", lock_type, 0);
-      if (lock->get_status)
-	(*lock->get_status)(data->status_param, 0);
+      if ((lock->get_status) && (*lock->get_status)(data->status_param, 0))
+        result= THR_LOCK_ABORTED;
       statistic_increment(locks_immediate,&THR_LOCK_lock);
       goto end;
     }
@@ -950,9 +952,10 @@ thr_lock(THR_LOCK_DATA *data, THR_LOCK_INFO *owner, ulong lock_wait_timeout)
 	data->prev=lock->write.last;
 	lock->write.last= &data->next;
 	check_locks(lock,"second write lock", lock_type, 0);
-	if (lock->get_status)
-	  (*lock->get_status)(data->status_param,
-                              lock_type == TL_WRITE_CONCURRENT_INSERT);
+	if ((lock->get_status) &&
+            (*lock->get_status)(data->status_param,
+                                lock_type == TL_WRITE_CONCURRENT_INSERT))
+          result= THR_LOCK_ABORTED;
 	statistic_increment(locks_immediate,&THR_LOCK_lock);
 	goto end;
       }
@@ -985,8 +988,9 @@ thr_lock(THR_LOCK_DATA *data, THR_LOCK_INFO *owner, ulong lock_wait_timeout)
 	  (*lock->write.last)=data;		/* Add as current write lock */
 	  data->prev=lock->write.last;
 	  lock->write.last= &data->next;
-	  if (lock->get_status)
-	    (*lock->get_status)(data->status_param, concurrent_insert);
+	  if ((lock->get_status) &&
+              (*lock->get_status)(data->status_param, concurrent_insert))
+            result= THR_LOCK_ABORTED;
 	  check_locks(lock,"only write lock", lock_type, 0);
 	  statistic_increment(locks_immediate,&THR_LOCK_lock);
 	  goto end;
@@ -1540,7 +1544,7 @@ void thr_downgrade_write_lock(THR_LOCK_DATA *in_data,
                               enum thr_lock_type new_lock_type)
 {
   THR_LOCK *lock=in_data->lock;
-#ifndef DBUG_OFF
+#ifdef DBUG_ASSERT_EXISTS
   enum thr_lock_type old_lock_type= in_data->type;
 #endif
   DBUG_ENTER("thr_downgrade_write_only_lock");
@@ -1580,6 +1584,7 @@ my_bool thr_upgrade_write_delay_lock(THR_LOCK_DATA *data,
   {
     if (!lock->read.data)			/* No read locks */
     {						/* We have the lock */
+      /* For this function, get_status is not allowed to fail */
       if (data->lock->get_status)
 	(*data->lock->get_status)(data->status_param, 0);
       mysql_mutex_unlock(&lock->mutex);
@@ -1683,31 +1688,35 @@ void thr_print_locks(void)
   uint count=0;
 
   mysql_mutex_lock(&THR_LOCK_lock);
-  puts("Current locks:");
+  puts("Current active THR (table level locks):");
   for (list= thr_lock_thread_list; list && count++ < MAX_THREADS;
        list= list_rest(list))
   {
     THR_LOCK *lock=(THR_LOCK*) list->data;
     mysql_mutex_lock(&lock->mutex);
-    printf("lock:%p:", lock);
-    if ((lock->write_wait.data || lock->read_wait.data) &&
-	(! lock->read.data && ! lock->write.data))
-      printf(" WARNING: ");
-    if (lock->write.data)
-      printf(" write");
-    if (lock->write_wait.data)
-      printf(" write_wait");
-    if (lock->read.data)
-      printf(" read");
-    if (lock->read_wait.data)
-      printf(" read_wait");
-    puts("");
-    thr_print_lock("write",&lock->write);
-    thr_print_lock("write_wait",&lock->write_wait);
-    thr_print_lock("read",&lock->read);
-    thr_print_lock("read_wait",&lock->read_wait);
+    if ((lock->write.data || lock->read.data ||
+         lock->write_wait.data || lock->read_wait.data))
+    {
+      printf("lock: %p:", lock);
+      if ((lock->write_wait.data || lock->read_wait.data) &&
+          (! lock->read.data && ! lock->write.data))
+        printf(" WARNING: ");
+      if (lock->write.data)
+        printf(" write");
+      if (lock->write_wait.data)
+        printf(" write_wait");
+      if (lock->read.data)
+        printf(" read");
+      if (lock->read_wait.data)
+        printf(" read_wait");
+      puts("");
+      thr_print_lock("write",&lock->write);
+      thr_print_lock("write_wait",&lock->write_wait);
+      thr_print_lock("read",&lock->read);
+      thr_print_lock("read_wait",&lock->read_wait);
+      puts("");
+    }
     mysql_mutex_unlock(&lock->mutex);
-    puts("");
   }
   fflush(stdout);
   mysql_mutex_unlock(&THR_LOCK_lock);
@@ -1776,9 +1785,10 @@ static ulong sum=0;
 
 /* The following functions is for WRITE_CONCURRENT_INSERT */
 
-static void test_get_status(void* param __attribute__((unused)),
-                            my_bool concurrent_insert __attribute__((unused)))
+static my_bool test_get_status(void* param __attribute__((unused)),
+                               my_bool concurrent_insert __attribute__((unused)))
 {
+  return 0;
 }
 
 static void test_update_status(void* param __attribute__((unused)))
