@@ -1,6 +1,6 @@
 #!/bin/bash -ue
 # Copyright (C) 2013 Percona Inc
-# Copyright (C) 2017-2019 MariaDB
+# Copyright (C) 2017-2021 MariaDB
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -37,6 +37,7 @@ REMOTEIP=""
 tcert=""
 tpem=""
 tkey=""
+tmode="DISABLED"
 sockopt=""
 progress=""
 ttime=0
@@ -70,6 +71,8 @@ xtmpdir=""
 scomp=""
 sdecomp=""
 
+readonly SECRET_TAG="secret"
+
 # Required for backup locks
 # For backup locks it is 1 sent by joiner
 # 5.6.21 PXC and later can't donate to an older joiner
@@ -186,27 +189,48 @@ get_transfer()
             if nc -h 2>&1 | grep -q ncat;then 
                 # Ncat
                 tcmd="nc -l ${TSST_PORT}"
-            elif nc -h 2>&1 | grep -q -- '-d\>';then
+            elif nc -h 2>&1 | grep -qw -- '-d\>';then
                 # Debian netcat
-                tcmd="nc -dl ${TSST_PORT}"
+                if [ $WSREP_SST_OPT_HOST_IPv6 -eq 1 ];then
+                    # When host is not explicitly specified (when only the port
+                    # is specified) netcat can only bind to an IPv4 address if
+                    # the "-6" option is not explicitly specified:
+                    tcmd="nc -dl -6 ${TSST_PORT}"
+                else
+                    tcmd="nc -dl ${TSST_PORT}"
+                fi
             else
                 # traditional netcat
                 tcmd="nc -l -p ${TSST_PORT}"
             fi
         else
+            # Check to see if netcat supports the '-N' flag.
+            # -N Shutdown the network socket after EOF on stdin
+            # If it supports the '-N' flag, then we need to use the '-N'
+            # flag, otherwise the transfer will stay open after the file
+            # transfer and cause the command to timeout.
+            # Older versions of netcat did not need this flag and will
+            # return an error if the flag is used.
+            #
+            tcmd_extra=""
+            if nc -h 2>&1 | grep -qw -- -N; then
+                tcmd_extra+="-N"
+		wsrep_log_info "Using nc -N"
+            fi
+
+            # netcat doesn't understand [] around IPv6 address
             if nc -h 2>&1 | grep -q ncat;then
                 # Ncat
-                tcmd="nc ${REMOTEIP} ${TSST_PORT}"
-            elif nc -h 2>&1 | grep -q -- '-d\>';then
+                wsrep_log_info "Using Ncat as streamer"
+                tcmd="nc ${tcmd_extra} ${WSREP_SST_OPT_HOST_UNESCAPED} ${TSST_PORT}"
+            elif nc -h 2>&1 | grep -qw -- '-d\>';then
                 # Debian netcat
-                if nc -h 2>&1 | grep -q -- '-N\>';then
-                   tcmd="nc -N ${REMOTEIP} ${TSST_PORT}"
-                else
-                   tcmd="nc ${REMOTEIP} ${TSST_PORT}"
-                fi
+                wsrep_log_info "Using Debian netcat as streamer"
+                tcmd="nc ${tcmd_extra} ${WSREP_SST_OPT_HOST_UNESCAPED} ${TSST_PORT}"
             else
                 # traditional netcat
-                tcmd="nc -q0 ${REMOTEIP} ${TSST_PORT}"
+                wsrep_log_info "Using traditional netcat as streamer"
+                tcmd="nc -q0 ${tcmd_extra} ${WSREP_SST_OPT_HOST_UNESCAPED} ${TSST_PORT}"
             fi
         fi
     else
@@ -240,24 +264,30 @@ get_transfer()
                 exit 22
             fi
             stagemsg+="-OpenSSL-Encrypted-3"
-            if [[ "$WSREP_SST_OPT_ROLE"  == "joiner" ]];then
-                if [[ -z $tcert ]];then
+            if [[ -z $tcert ]];then
+                # no verification
+                if [[ "$WSREP_SST_OPT_ROLE"  == "joiner" ]];then
                     wsrep_log_info "Decrypting with cert=${tpem}, key=${tkey}, verify=0"
                     tcmd="socat -u openssl-listen:${TSST_PORT},reuseaddr,cert=${tpem},key=${tkey},verify=0${sockopt} stdio"
                 else
-                    wsrep_log_info "Decrypting with cert=${tpem}, key=${tkey}, cafile=${tcert}"
-                    tcmd="socat -u openssl-listen:${TSST_PORT},reuseaddr,cert=${tpem},key=${tkey},cafile=${tcert}${sockopt} stdio"
-                fi
-            else
-                if [[ -z $tcert ]];then
                     wsrep_log_info "Encrypting with cert=${tpem}, key=${tkey}, verify=0"
                     tcmd="socat -u stdio openssl-connect:${REMOTEIP}:${TSST_PORT},cert=${tpem},key=${tkey},verify=0${sockopt}"
+                fi
+            else
+                # CA verification
+                if [[ "$WSREP_SST_OPT_ROLE"  == "joiner" ]];then
+                    wsrep_log_info "Decrypting with cert=${tpem}, key=${tkey}, cafile=${tcert}"
+                    tcmd="socat -u openssl-listen:${TSST_PORT},reuseaddr,cert=${tpem},key=${tkey},cafile=${tcert}${sockopt} stdio"
                 else
+                    if [ -n "$WSREP_SST_OPT_REMOTE_USER" ]; then
+                        CN_option=",commonname=$WSREP_SST_OPT_REMOTE_USER"
+                    else
+                        CN_option=""
+                    fi
                     wsrep_log_info "Encrypting with cert=${tpem}, key=${tkey}, cafile=${tcert}"
-                    tcmd="socat -u stdio openssl-connect:${REMOTEIP}:${TSST_PORT},cert=${tpem},key=${tkey},cafile=${tcert}${sockopt}"
+                    tcmd="socat -u stdio openssl-connect:${REMOTEIP}:${TSST_PORT},cert=${tpem},key=${tkey},cafile=${tcert}${CN_option}${sockopt}"
                 fi
             fi
-
         else 
             if [[ "$WSREP_SST_OPT_ROLE"  == "joiner" ]];then
                 tcmd="socat -u TCP-LISTEN:${TSST_PORT},reuseaddr${sockopt} stdio"
@@ -266,7 +296,6 @@ get_transfer()
             fi
         fi
     fi
-
 }
 
 parse_cnf()
@@ -274,10 +303,17 @@ parse_cnf()
     local group=$1
     local var=$2
     # print the default settings for given group using my_print_default.
+    # remove possible 'loose' variable name prefix
     # normalize the variable names specified in cnf file (user can use _ or - for example log-bin or log_bin)
     # then grep for needed variable
     # finally get the variable value (if variables has been specified multiple time use the last value only)
-    reval=$($MY_PRINT_DEFAULTS $group | awk -F= '{if ($1 ~ /_/) { gsub(/_/,"-",$1); print $1"="$2 } else { print $0 }}' | grep -- "--$var=" | cut -d= -f2- | tail -1)
+    reval=$($MY_PRINT_DEFAULTS $group | \
+            awk -F= '{ sub(/^--loose/,"-",$0); \
+                       if ($1 ~ /_/) \
+                           { gsub(/_/,"-",$1); print $1"="$2 } \
+                       else \
+                           { print $0 }}' | \
+            grep -- "--$var=" | cut -d= -f2- | tail -1)
     if [[ -z $reval ]];then 
         [[ -n $3 ]] && reval=$3
     fi
@@ -327,6 +363,24 @@ adjust_progress()
     fi
 }
 
+check_server_ssl_config()
+{
+    local section=$1
+    tcert=$(parse_cnf $section ssl-ca "")
+    tpem=$(parse_cnf $section ssl-cert "")
+    tkey=$(parse_cnf $section ssl-key "")
+    if [ 0 -eq $encrypt -a -n "$tpem" -a -n "$tkey" ]
+    then
+         encrypt=3 # enable cert/key SSL encyption
+
+         # avoid CA verification if not set explicitly:
+         # nodes may happen to have different CA if self-generated
+         # zeroing up tcert does the trick
+         local mode=$(parse_cnf SST ssl-mode "")
+         [[ ${mode} = *VERIFY* ]] || tcert=""
+    fi
+}
+
 read_cnf()
 {
     sfmt=$(parse_cnf sst streamfmt "xbstream")
@@ -335,6 +389,26 @@ read_cnf()
     tpem=$(parse_cnf sst tcert "")
     tkey=$(parse_cnf sst tkey "")
     encrypt=$(parse_cnf sst encrypt 0)
+    tmode=$(parse_cnf sst ssl-mode "DISABLED" | tr [:lower:] [:upper:])
+
+    if [ -z "$tpem" -a -z "$tkey" -a -z "$tcert" ]
+    then # no old-style SSL config in [sst]
+        if [ "$tmode" != "DISABLED" ]
+        then # backward-incompatible behavior
+            check_server_ssl_config "sst"
+            if [ -z "$tpem" -a -z "$tkey" -a -z "$tcert" ]
+            then # no new-stype SSL config in [sst], try server-wide SSL config
+                check_server_ssl_config "mysqld.$WSREP_SST_OPT_SUFFIX_VALUE"
+                if [ -z "$tpem" -a -z "$tkey" -a -z "$tcert" ]
+                then
+                    check_server_ssl_config "mysqld"
+                fi
+            fi
+        fi
+    fi
+    wsrep_log_info "SSL configuration: CA='"$tcert"', CERT='"$tpem"'," \
+                   "KEY='"$tkey"', MODE='"$tmode"', encrypt="$encrypt
+
     sockopt=$(parse_cnf sst sockopt "")
     progress=$(parse_cnf sst progress "")
     ttime=$(parse_cnf sst time 0)
@@ -359,7 +433,7 @@ read_cnf()
     iopts=$(parse_cnf sst inno-backup-opts "")
     iapts=$(parse_cnf sst inno-apply-opts "")
     impts=$(parse_cnf sst inno-move-opts "")
-    stimeout=$(parse_cnf sst sst-initial-timeout 100)
+    stimeout=$(parse_cnf sst sst-initial-timeout 300)
     ssyslog=$(parse_cnf sst sst-syslog 0)
     ssystag=$(parse_cnf mysqld_safe syslog-tag "${SST_SYSLOG_TAG:-}")
     ssystag+="-"
@@ -379,7 +453,7 @@ read_cnf()
 
     if [[ $encrypt -eq 1 ]]; then
         wsrep_log_error "Xtrabackup-based encryption is currently not" \
-            "supported with MariaBackup"
+                        "supported with MariaBackup"
         exit 2
     fi
 }
@@ -523,25 +597,11 @@ kill_xtrabackup()
 
 setup_ports()
 {
+    SST_PORT=${WSREP_SST_OPT_ADDR_PORT}
     if [[ "$WSREP_SST_OPT_ROLE"  == "donor" ]];then
-        if [ "${WSREP_SST_OPT_ADDR#\[}" != "$WSREP_SST_OPT_ADDR" ]; then
-            remain=$(echo $WSREP_SST_OPT_ADDR | awk -F '\\][:/]' '{ print $2 }')
-            REMOTEIP=$(echo $WSREP_SST_OPT_ADDR | awk -F '\\]:' '{ print $1 }')"]"
-            SST_PORT=$(echo $remain | awk -F '[:/]' '{ print $1 }')
-            lsn=$(echo $remain | awk -F '[:/]' '{ print $3 }')
-            sst_ver=$(echo $remain | awk -F '[:/]' '{ print $4 }')
-        else
-            SST_PORT=$(echo $WSREP_SST_OPT_ADDR | awk -F '[:/]' '{ print $2 }')
-            REMOTEIP=$(echo $WSREP_SST_OPT_ADDR | awk -F ':' '{ print $1 }')
-            lsn=$(echo $WSREP_SST_OPT_ADDR | awk -F '[:/]' '{ print $4 }')
-            sst_ver=$(echo $WSREP_SST_OPT_ADDR | awk -F '[:/]' '{ print $5 }')
-        fi
-    else
-        if [ "${WSREP_SST_OPT_ADDR#\[}" != "$WSREP_SST_OPT_ADDR" ]; then
-            SST_PORT=$(echo ${WSREP_SST_OPT_ADDR} | awk -F '\\]:' '{ print $2 }')
-        else
-            SST_PORT=$(echo ${WSREP_SST_OPT_ADDR} | awk -F ':' '{ print $2 }')
-        fi
+        REMOTEIP=${WSREP_SST_OPT_HOST}
+        lsn=${WSREP_SST_OPT_LSN}
+        sst_ver=${WSREP_SST_OPT_SST_VER}
     fi
 }
 
@@ -620,7 +680,8 @@ recv_joiner()
     popd 1>/dev/null 
 
     if [[ ${RC[0]} -eq 124 ]];then 
-        wsrep_log_error "Possible timeout in receving first data from donor in gtid stage"
+        wsrep_log_error "Possible timeout in receiving first data from " \
+                        "donor in gtid stage: exit codes: ${RC[@]}"
         exit 32
     fi
 
@@ -632,12 +693,27 @@ recv_joiner()
         fi
     done
 
-    if [[ $checkf -eq 1 && ! -r "${MAGIC_FILE}" ]];then
-        # this message should cause joiner to abort
-        wsrep_log_error "xtrabackup process ended without creating '${MAGIC_FILE}'"
-        wsrep_log_info "Contents of datadir" 
-        wsrep_log_info "$(ls -l ${dir}/*)"
-        exit 32
+    if [[ $checkf -eq 1 ]]; then
+        if [[ ! -r "${MAGIC_FILE}" ]];then
+            # this message should cause joiner to abort
+            wsrep_log_error "receiving process ended without creating " \
+                            "'${MAGIC_FILE}'"
+            wsrep_log_info "Contents of datadir" 
+            wsrep_log_info "$(ls -l ${dir}/*)"
+            exit 32
+        fi
+
+        # check donor supplied secret
+        SECRET=$(grep "$SECRET_TAG " ${MAGIC_FILE} 2>/dev/null | cut -d ' ' -f 2)
+        if [[ $SECRET != $MY_SECRET ]]; then
+            wsrep_log_error "Donor does not know my secret!"
+            wsrep_log_info "Donor:'$SECRET', my:'$MY_SECRET'"
+            exit 32
+        fi
+
+        # remove secret from magic file
+        grep -v "$SECRET_TAG " ${MAGIC_FILE} > ${MAGIC_FILE}.new
+        mv ${MAGIC_FILE}.new ${MAGIC_FILE}
     fi
 }
 
@@ -653,10 +729,9 @@ send_donor()
     set -e
     popd 1>/dev/null 
 
-
     for ecode in "${RC[@]}";do 
         if [[ $ecode -ne 0 ]];then 
-            wsrep_log_error "Error while getting data from donor node: " \
+            wsrep_log_error "Error while sending data to joiner node: " \
                             "exit codes: ${RC[@]}"
             exit 32
         fi
@@ -700,9 +775,11 @@ if ${INNOBACKUPEX_BIN} /tmp --help 2>/dev/null | grep -q -- '--version-check'; t
     disver="--no-version-check"
 fi
 
+iopts+=" --databases-exclude=\"lost+found\""
+
 if [[ ${FORCE_FTWRL:-0} -eq 1 ]];then 
     wsrep_log_info "Forcing FTWRL due to environment variable FORCE_FTWRL equal to $FORCE_FTWRL"
-    iopts+=" --no-backup-locks "
+    iopts+=" --no-backup-locks"
 fi
 
 INNOEXTRA=
@@ -712,7 +789,8 @@ INNODB_DATA_HOME_DIR=${INNODB_DATA_HOME_DIR:-""}
 if [ ! -z "$INNODB_DATA_HOME_DIR_ARG" ]; then
     INNODB_DATA_HOME_DIR=$INNODB_DATA_HOME_DIR_ARG
 fi
-# if INNODB_DATA_HOME_DIR env. variable is not set, try to get it from my.cnf
+# if no command line arg and INNODB_DATA_HOME_DIR environment variable
+# is not set, try to get it from my.cnf:
 if [ -z "$INNODB_DATA_HOME_DIR" ]; then
     INNODB_DATA_HOME_DIR=$(parse_cnf mysqld$WSREP_SST_OPT_SUFFIX_VALUE innodb-data-home-dir '')
 fi
@@ -840,7 +918,7 @@ then
               -z $(parse_cnf --mysqld tmpdir "") && \
               -z $(parse_cnf xtrabackup tmpdir "") ]]; then
             xtmpdir=$(mktemp -d)
-            tmpopts=" --tmpdir=$xtmpdir"
+            tmpopts="--tmpdir=$xtmpdir"
             wsrep_log_info "Using $xtmpdir as xtrabackup temporary directory"
         fi
 
@@ -853,10 +931,10 @@ then
         fi
 
         if [ -n "${WSREP_SST_OPT_PSWD:-}" ]; then
-           INNOEXTRA+=" --password=$WSREP_SST_OPT_PSWD"
+            export MYSQL_PWD=$WSREP_SST_OPT_PSWD
         elif [[ $usrst -eq 1 ]];then
-           # Empty password, used for testing, debugging etc.
-           INNOEXTRA+=" --password="
+            # Empty password, used for testing, debugging etc.
+            unset MYSQL_PWD
         fi
 
         check_extra
@@ -867,6 +945,11 @@ then
         # (separated by a space).
         echo "${WSREP_SST_OPT_GTID} ${WSREP_SST_OPT_GTID_DOMAIN_ID}" > "${MAGIC_FILE}"
 
+        if [[ -n ${WSREP_SST_OPT_REMOTE_PSWD} ]]; then
+            # Let joiner know that we know its secret
+            echo "$SECRET_TAG ${WSREP_SST_OPT_REMOTE_PSWD}" >> ${MAGIC_FILE}
+        fi
+
         ttcmd="$tcmd"
 
         if [[ $encrypt -eq 1 ]];then
@@ -951,17 +1034,25 @@ then
 
     ib_home_dir=$INNODB_DATA_HOME_DIR
 
-    # Try to set ib_log_dir from the command line:
-    ib_log_dir=$INNODB_LOG_GROUP_HOME_ARG
-    if [ -z "$ib_log_dir" ]; then
-        ib_log_dir=$(parse_cnf mysqld$WSREP_SST_OPT_SUFFIX_VALUE innodb-log-group-home-dir "")
+    WSREP_LOG_DIR=${WSREP_LOG_DIR:-""}
+    # Try to set WSREP_LOG_DIR from the command line:
+    if [ ! -z "$INNODB_LOG_GROUP_HOME_ARG" ]; then
+        WSREP_LOG_DIR=$INNODB_LOG_GROUP_HOME_ARG
     fi
-    if [ -z "$ib_log_dir" ]; then
-        ib_log_dir=$(parse_cnf --mysqld innodb-log-group-home-dir "")
+    # if no command line arg and WSREP_LOG_DIR is not set,
+    # try to get it from my.cnf:
+    if [ -z "$WSREP_LOG_DIR" ]; then
+        WSREP_LOG_DIR=$(parse_cnf mysqld$WSREP_SST_OPT_SUFFIX_VALUE innodb-log-group-home-dir '')
+    fi
+    if [ -z "$WSREP_LOG_DIR" ]; then
+        WSREP_LOG_DIR=$(parse_cnf --mysqld innodb-log-group-home-dir '')
     fi
 
+    ib_log_dir=$WSREP_LOG_DIR
+
     # Try to set ib_undo_dir from the command line:
-    ib_undo_dir=$INNODB_UNDO_DIR_ARG
+    ib_undo_dir=${INNODB_UNDO_DIR_ARG:-""}
+    # if no command line arg then try to get it from my.cnf:
     if [ -z "$ib_undo_dir" ]; then
         ib_undo_dir=$(parse_cnf mysqld$WSREP_SST_OPT_SUFFIX_VALUE innodb-undo-directory "")
     fi
@@ -971,7 +1062,6 @@ then
 
     stagemsg="Joiner-Recv"
 
-
     sencrypted=1
     nthreads=1
 
@@ -993,7 +1083,26 @@ then
         fi
     fi
 
-    wait_for_listen ${SST_PORT} ${ADDR} ${MODULE} &
+    if [[ "$tmode" = *"VERIFY"* ]]
+    then # backward-incompatible behavior
+        if [ -n "$tpem" ]
+        then
+            # find out my Common Name
+            wsrep_check_programs openssl
+            CN=$(openssl x509 -noout -subject -in $tpem | \
+                 tr "," "\n" | grep "CN =" | cut -d= -f2 | sed s/^\ // | \
+                 sed s/\ %//)
+        else
+            CN=""
+        fi
+        MY_SECRET=$(wsrep_gen_secret)
+        # Add authentication data to address
+        ADDR="$CN:$MY_SECRET@$ADDR"
+    else
+        MY_SECRET="" # for check down in recv_joiner()
+    fi # tmode == *VERIFY*
+
+    wait_for_listen ${SST_PORT} "${ADDR}" ${MODULE} &
 
     trap sig_joiner_cleanup HUP PIPE INT TERM
     trap cleanup_joiner EXIT
