@@ -3,7 +3,7 @@
 
 /*
    Copyright (c) 2005, 2012, Oracle and/or its affiliates.
-   Copyright (c) 2009, 2022, MariaDB Corporation.
+   Copyright (c) 2009, 2021, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -21,19 +21,8 @@
 #include "sql_partition.h"      /* part_id_range, partition_element */
 #include "queues.h"             /* QUEUE */
 
-struct Ordered_blob_storage
-{
-  String blob;
-  bool set_read_value;
-  Ordered_blob_storage() : set_read_value(false)
-  {}
-};
-
-#define PAR_EXT ".par"
 #define PARTITION_BYTES_IN_POS 2
-#define ORDERED_PART_NUM_OFFSET sizeof(Ordered_blob_storage **)
-#define ORDERED_REC_OFFSET (ORDERED_PART_NUM_OFFSET + PARTITION_BYTES_IN_POS)
-
+#define PAR_EXT ".par"
 
 /** Struct used for partition_name_hash */
 typedef struct st_part_name_def
@@ -536,6 +525,10 @@ public:
     Meta data routines to CREATE, DROP, RENAME table and often used at
     ALTER TABLE (update_create_info used from ALTER TABLE and SHOW ..).
 
+    update_table_comment is used in SHOW TABLE commands to provide a
+    chance for the handler to add any interesting comments to the table
+    comments not provided by the users comment.
+
     create_partitioning_metadata is called before opening a new handler object
     with openfrm to call create. It is used to create any local handler
     object needed in opening the object in openfrm
@@ -551,6 +544,7 @@ public:
     override;
   bool check_if_updates_are_ignored(const char *op) const override;
   void update_create_info(HA_CREATE_INFO *create_info) override;
+  char *update_table_comment(const char *comment) override;
   int change_partitions(HA_CREATE_INFO *create_info, const char *path,
                         ulonglong * const copied, ulonglong * const deleted,
                         const uchar *pack_frm_data, size_t pack_frm_len)
@@ -588,9 +582,8 @@ private:
     And one method to read it in.
   */
   bool create_handler_file(const char *name);
-  bool setup_engine_array(MEM_ROOT *mem_root, handlerton *first_engine);
+  bool setup_engine_array(MEM_ROOT *mem_root);
   bool read_par_file(const char *name);
-  handlerton *get_def_part_engine(const char *name);
   bool get_from_handler_file(const char *name, MEM_ROOT *mem_root,
                              bool is_clone);
   bool new_handlers_from_part_info(MEM_ROOT *mem_root);
@@ -945,7 +938,6 @@ private:
   int handle_ordered_next(uchar * buf, bool next_same);
   int handle_ordered_prev(uchar * buf);
   void return_top_record(uchar * buf);
-  void swap_blobs(uchar* rec_buf, Ordered_blob_storage ** storage, bool restore);
 public:
   /*
     -------------------------------------------------------------------------
@@ -1079,9 +1071,6 @@ public:
   */
   const char *index_type(uint inx) override;
 
-  /* The name of the table type that will be used for display purposes */
-  const char *table_type() const;
-
   /* The name of the row type used for the underlying tables. */
   enum row_type get_row_type() const override;
 
@@ -1107,6 +1096,10 @@ public:
     NOTE: This cannot be cached since it can depend on TRANSACTION ISOLATION
     LEVEL which is dynamic, see bug#39084.
 
+    HA_READ_RND_SAME:
+    Not currently used. (Means that the handler supports the rnd_same() call)
+    (MyISAM, HEAP)
+
     HA_TABLE_SCAN_ON_INDEX:
     Used to avoid scanning full tables on an index. If this flag is set then
     the handler always has a primary key (hidden if not defined) and this
@@ -1365,7 +1358,7 @@ public:
   void release_auto_increment() override;
 private:
   int reset_auto_increment(ulonglong value) override;
-  int update_next_auto_inc_val();
+  void update_next_auto_inc_val();
   virtual void lock_auto_increment()
   {
     /* lock already taken */
@@ -1404,17 +1397,15 @@ private:
     unlock_auto_increment();
   }
 
-  void check_insert_or_replace_autoincrement()
+  void check_insert_autoincrement()
   {
     /*
-      If we INSERT or REPLACE into the table having the AUTO_INCREMENT column,
+      If we INSERT into the table having the AUTO_INCREMENT column,
       we have to read all partitions for the next autoincrement value
       unless we already did it.
     */
     if (!part_share->auto_inc_initialized &&
-        (ha_thd()->lex->sql_command == SQLCOM_INSERT ||
-         ha_thd()->lex->sql_command == SQLCOM_INSERT_SELECT ||
-         ha_thd()->lex->sql_command == SQLCOM_REPLACE) &&
+        ha_thd()->lex->sql_command == SQLCOM_INSERT &&
         table->found_next_number_field)
       bitmap_set_all(&m_part_info->read_partitions);
   }
@@ -1616,6 +1607,7 @@ public:
     for (; part_id < part_id_end; ++part_id)
     {
       handler *file= m_file[part_id];
+      DBUG_ASSERT(bitmap_is_set(&(m_part_info->read_partitions), part_id));
       file->info(HA_STATUS_VARIABLE | HA_STATUS_NO_LOCK | HA_STATUS_OPEN);
       part_recs+= file->stats.records;
     }
