@@ -23,7 +23,7 @@
 /*
   Compare ASCII string against the string with the specified
   character set.
-  Only compares the equality, case insencitive.
+  Only compares the equality, case insensitive.
 */
 static bool eq_ascii_string(const CHARSET_INFO *cs,
                             const char *ascii,
@@ -247,15 +247,15 @@ error:
 
 
 #define report_json_error(js, je, n_param) \
-  report_json_error_ex(js, je, func_name(), n_param, \
+  report_json_error_ex(js->ptr(), je, func_name(), n_param, \
       Sql_condition::WARN_LEVEL_WARN)
 
-void report_json_error_ex(String *js, json_engine_t *je,
+void report_json_error_ex(const char *js, json_engine_t *je,
                           const char *fname, int n_param,
                           Sql_condition::enum_warning_level lv)
 {
   THD *thd= current_thd;
-  int position= (int)((const char *) je->s.c_str - js->ptr());
+  int position= (int)((const char *) je->s.c_str - js);
   uint code;
 
   n_param++;
@@ -285,16 +285,22 @@ void report_json_error_ex(String *js, json_engine_t *je,
 
   case JE_DEPTH:
     code= ER_JSON_DEPTH;
-    push_warning_printf(thd, lv, code, ER_THD(thd, code), JSON_DEPTH_LIMIT,
-                        n_param, fname, position);
+    if (lv == Sql_condition::WARN_LEVEL_ERROR)
+      my_error(code, MYF(0), JSON_DEPTH_LIMIT, n_param, fname, position);
+    else
+      push_warning_printf(thd, lv, code, ER_THD(thd, code), JSON_DEPTH_LIMIT,
+                          n_param, fname, position);
     return;
 
   default:
     return;
   }
 
-  push_warning_printf(thd, lv, code, ER_THD(thd, code),
-                      n_param, fname, position);
+  if (lv == Sql_condition::WARN_LEVEL_ERROR)
+    my_error(code, MYF(0), n_param, fname, position);
+  else
+    push_warning_printf(thd, lv, code, ER_THD(thd, code),
+                        n_param, fname, position);
 }
 
 
@@ -304,15 +310,15 @@ void report_json_error_ex(String *js, json_engine_t *je,
 #define TRIVIAL_PATH_NOT_ALLOWED 3
 
 #define report_path_error(js, je, n_param) \
-  report_path_error_ex(js, je, func_name(), n_param,\
+  report_path_error_ex(js->ptr(), je, func_name(), n_param,\
       Sql_condition::WARN_LEVEL_WARN)
 
-static void report_path_error_ex(String *ps, json_path_t *p,
-                                 const char *fname, int n_param,
-                                 Sql_condition::enum_warning_level lv)
+void report_path_error_ex(const char *ps, json_path_t *p,
+                          const char *fname, int n_param,
+                          Sql_condition::enum_warning_level lv)
 {
   THD *thd= current_thd;
-  int position= (int)((const char *) p->s.c_str - ps->ptr() + 1);
+  int position= (int)((const char *) p->s.c_str - ps + 1);
   uint code;
 
   n_param++;
@@ -331,8 +337,11 @@ static void report_path_error_ex(String *ps, json_path_t *p,
 
   case JE_DEPTH:
     code= ER_JSON_PATH_DEPTH;
-    push_warning_printf(thd, lv, code, ER_THD(thd, code),
-                        JSON_DEPTH_LIMIT, n_param, fname, position);
+    if (lv == Sql_condition::WARN_LEVEL_ERROR)
+      my_error(code, MYF(0), JSON_DEPTH_LIMIT, n_param, fname, position);
+    else
+      push_warning_printf(thd, lv, code, ER_THD(thd, code),
+                          JSON_DEPTH_LIMIT, n_param, fname, position);
     return;
 
   case NO_WILDCARD_ALLOWED:
@@ -347,12 +356,14 @@ static void report_path_error_ex(String *ps, json_path_t *p,
   default:
     return;
   }
-  push_warning_printf(thd, lv, code, ER_THD(thd, code),
-                      n_param, fname, position);
+  if (lv == Sql_condition::WARN_LEVEL_ERROR)
+    my_error(code, MYF(0), n_param, fname, position);
+  else
+    push_warning_printf(thd, lv, code, ER_THD(thd, code),
+                        n_param, fname, position);
 }
 
 
-
 /*
   Checks if the path has '.*' '[*]' or '**' constructions
   and sets the NO_WILDCARD_ALLOWED error if the case.
@@ -613,8 +624,6 @@ String *Item_func_json_unquote::read_json(json_engine_t *je)
   json_scan_start(je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
 
-  je->value_type= (enum json_value_types) -1; /* To report errors right. */
-
   if (json_read_value(je))
     goto error;
 
@@ -977,7 +986,8 @@ my_decimal *Item_func_json_extract::val_decimal(my_decimal *to)
       case JSON_VALUE_ARRAY:
       case JSON_VALUE_FALSE:
       case JSON_VALUE_NULL:
-        break;
+      case JSON_VALUE_UNINITALIZED:
+      break;
     };
   }
   int2my_decimal(E_DEC_FATAL_ERROR, 0, false/*unsigned_flag*/, to);
@@ -1115,7 +1125,7 @@ static int check_contains(json_engine_t *js, json_engine_t *value)
     if (value->value_type != JSON_VALUE_STRING)
       return FALSE;
     /*
-       TODO: make proper json-json comparison here that takes excapint
+       TODO: make proper json-json comparison here that takes excipient
              into account.
      */
     return value->value_len == js->value_len &&
@@ -1403,7 +1413,7 @@ longlong Item_func_json_contains_path::val_int()
     n_found= arg_count - 2;
   }
   else
-    n_found= 0; /* Jost to prevent 'uninitialized value' warnings */
+    n_found= 0; /* Just to prevent 'uninitialized value' warnings */
 
   result= 0;
   while (json_get_path_next(&je, &p) == 0)
@@ -3373,7 +3383,7 @@ bool Item_func_json_search::fix_length_and_dec()
 
   /*
     It's rather difficult to estimate the length of the result.
-    I belive arglen^2 is the reasonable upper limit.
+    I believe arglen^2 is the reasonable upper limit.
   */
   if (args[0]->max_length > SQR_MAX_BLOB_WIDTH)
     max_length= MAX_BLOB_WIDTH;
