@@ -41,6 +41,13 @@ public:
 };
 
 
+void report_path_error_ex(const char *ps, json_path_t *p,
+                          const char *fname, int n_param,
+                          Sql_condition::enum_warning_level lv);
+void report_json_error_ex(const char *js, json_engine_t *je,
+                          const char *fname, int n_param,
+                          Sql_condition::enum_warning_level lv);
+
 class Json_engine_scan: public json_engine_t
 {
 public:
@@ -589,7 +596,7 @@ public:
   void cleanup();
 
   enum Sumfunctype sum_func () const {return JSON_OBJECTAGG_FUNC;}
-  const char *func_name() const { return "json_objectagg("; }
+  const char *func_name() const { return "json_objectagg"; }
   const Type_handler *type_handler() const
   {
     if (too_big_for_varchar())
@@ -618,7 +625,6 @@ public:
   String* val_str(String* str);
   Item *copy_or_same(THD* thd);
   void no_rows_in_result() {}
-  void print(String *str, enum_query_type query_type);
   Item *get_copy(THD *thd)
   { return get_item_copy<Item_func_json_objectagg>(thd, this); }
 };
