@@ -28,8 +28,11 @@
 #pragma implementation				// gcc: Class implementation
 #endif
 
+#include "mariadb.h"
 #include "my_bit.h"
 #include "sql_select.h"
+#include "opt_trace.h"
+#include "my_json_writer.h"
 
 /*
   OVERVIEW
@@ -521,7 +524,8 @@ eliminate_tables_for_list(JOIN *join,
                           List<TABLE_LIST> *join_list,
                           table_map tables_in_list,
                           Item *on_expr,
-                          table_map tables_used_elsewhere);
+                          table_map tables_used_elsewhere,
+                          Json_writer_array* trace_eliminate_tables);
 static
 bool check_func_dependency(JOIN *join, 
                            table_map dep_tables,
@@ -540,7 +544,8 @@ static
 Dep_module_expr *merge_eq_mods(Dep_module_expr *start, 
                                  Dep_module_expr *new_fields, 
                                  Dep_module_expr *end, uint and_level);
-static void mark_as_eliminated(JOIN *join, TABLE_LIST *tbl);
+static void mark_as_eliminated(JOIN *join, TABLE_LIST *tbl,
+                               Json_writer_array* trace_eliminate_tables);
 static 
 void add_module_expr(Dep_analysis_context *dac, Dep_module_expr **eq_mod,
                      uint and_level, Dep_value_field *field_val, Item *right,
@@ -607,6 +612,8 @@ void eliminate_tables(JOIN *join)
   if (!optimizer_flag(thd, OPTIMIZER_SWITCH_TABLE_ELIMINATION))
     DBUG_VOID_RETURN; /* purecov: inspected */
 
+  Json_writer_object trace_wrapper(thd);
+
   /* Find the tables that are referred to from WHERE/HAVING */
   used_tables= (join->conds?  join->conds->used_tables() : 0) | 
                (join->having? join->having->used_tables() : 0);
@@ -616,12 +623,12 @@ void eliminate_tables(JOIN *join)
     we should also take into account tables mentioned in "val".
   */
   if (join->thd->lex->sql_command == SQLCOM_INSERT_SELECT &&
-      join->select_lex == &thd->lex->select_lex)
+      join->select_lex == thd->lex->first_select_lex())
   {
     List_iterator<Item> val_it(thd->lex->value_list);
     while ((item= val_it++))
     {
-      DBUG_ASSERT(item->fixed);
+      DBUG_ASSERT(item->is_fixed());
       used_tables |= item->used_tables();
     }
   }
@@ -630,6 +637,22 @@ void eliminate_tables(JOIN *join)
   List_iterator<Item> it(join->fields_list);
   while ((item= it++))
     used_tables |= item->used_tables();
+
+  {
+    /*
+      Table function JSON_TABLE() can have references to other tables. Do not
+      eliminate the tables that JSON_TABLE() refers to.
+      Note: the JSON_TABLE itself cannot be eliminated as it doesn't
+            have unique keys.
+    */
+    List_iterator<TABLE_LIST> it(join->select_lex->leaf_tables);
+    TABLE_LIST *tbl;
+    while ((tbl= it++))
+    {
+      if (tbl->table_function)
+        used_tables|= tbl->table_function->used_tables();
+    }
+  }
  
   /* Add tables referred to from ORDER BY and GROUP BY lists */
   ORDER *all_lists[]= { join->order, join->group_list};
@@ -639,7 +662,7 @@ void eliminate_tables(JOIN *join)
       used_tables |= (*(cur_list->item))->used_tables();
   }
   
-  if (join->select_lex == &thd->lex->select_lex)
+  if (join->select_lex == thd->lex->first_select_lex())
   {
 
     /* Multi-table UPDATE: don't eliminate tables referred from SET statement */
@@ -662,13 +685,14 @@ void eliminate_tables(JOIN *join)
       }
     }
   }
-  
+
   table_map all_tables= join->all_tables_map();
+  Json_writer_array trace_eliminated_tables(thd,"eliminated_tables");
   if (all_tables & ~used_tables)
   {
     /* There are some tables that we probably could eliminate. Try it. */
     eliminate_tables_for_list(join, join->join_list, all_tables, NULL,
-                              used_tables);
+                              used_tables, &trace_eliminated_tables);
   }
   DBUG_VOID_RETURN;
 }
@@ -711,7 +735,8 @@ void eliminate_tables(JOIN *join)
 static bool
 eliminate_tables_for_list(JOIN *join, List<TABLE_LIST> *join_list,
                           table_map list_tables, Item *on_expr,
-                          table_map tables_used_elsewhere)
+                          table_map tables_used_elsewhere,
+                          Json_writer_array *trace_eliminate_tables)
 {
   TABLE_LIST *tbl;
   List_iterator<TABLE_LIST> it(*join_list);
@@ -733,9 +758,10 @@ eliminate_tables_for_list(JOIN *join, List<TABLE_LIST> *join_list,
                                       &tbl->nested_join->join_list, 
                                       tbl->nested_join->used_tables, 
                                       tbl->on_expr,
-                                      outside_used_tables))
+                                      outside_used_tables,
+                                      trace_eliminate_tables))
         {
-          mark_as_eliminated(join, tbl);
+          mark_as_eliminated(join, tbl, trace_eliminate_tables);
         }
         else
           all_eliminated= FALSE;
@@ -747,7 +773,7 @@ eliminate_tables_for_list(JOIN *join, List<TABLE_LIST> *join_list,
             check_func_dependency(join, tbl->table->map, NULL, tbl, 
                                   tbl->on_expr))
         {
-          mark_as_eliminated(join, tbl);
+          mark_as_eliminated(join, tbl, trace_eliminate_tables);
         }
         else
           all_eliminated= FALSE;
@@ -1787,7 +1813,8 @@ Dep_module* Dep_value_field::get_next_unbound_module(Dep_analysis_context *dac,
   Mark one table or the whole join nest as eliminated.
 */
 
-static void mark_as_eliminated(JOIN *join, TABLE_LIST *tbl)
+static void mark_as_eliminated(JOIN *join, TABLE_LIST *tbl,
+                               Json_writer_array* trace_eliminate_tables)
 {
   TABLE *table;
   /*
@@ -1800,7 +1827,7 @@ static void mark_as_eliminated(JOIN *join, TABLE_LIST *tbl)
     TABLE_LIST *child;
     List_iterator<TABLE_LIST> it(tbl->nested_join->join_list);
     while ((child= it++))
-      mark_as_eliminated(join, child);
+      mark_as_eliminated(join, child, trace_eliminate_tables);
   }
   else if ((table= tbl->table))
   {
@@ -1809,7 +1836,9 @@ static void mark_as_eliminated(JOIN *join, TABLE_LIST *tbl)
     {
       DBUG_PRINT("info", ("Eliminated table %s", table->alias.c_ptr()));
       tab->type= JT_CONST;
+      tab->table->const_table= 1;
       join->eliminated_tables |= table->map;
+      trace_eliminate_tables->add(table->alias.c_ptr_safe());
       join->const_table_map|= table->map;
       set_position(join, join->const_tables++, tab, (KEYUSE*)0);
     }
@@ -1843,7 +1872,7 @@ void Dep_analysis_context::dbug_print_deps()
               (long)(eq_mod - equality_mods),
               str.c_ptr(),
               eq_mod->field->table->table->alias.c_ptr(),
-              eq_mod->field->field->field_name);
+              eq_mod->field->field->field_name.str);
     }
     else
     {
@@ -1867,7 +1896,7 @@ void Dep_analysis_context::dbug_print_deps()
       {
         fprintf(DBUG_FILE, "    field %s.%s ->",
                 table_dep->table->alias.c_ptr(),
-                field_dep->field->field_name);
+                field_dep->field->field_name.str);
         uint ofs= field_dep->bitmap_offset;
         for (uint bit= ofs; bit < ofs + n_equality_mods; bit++)
         {
