@@ -177,7 +177,7 @@ TABLE_FIELD_TYPE proc_table_fields[MYSQL_PROC_FIELD_COUNT] =
   },
   {
     { STRING_WITH_LEN("definer") },
-    { STRING_WITH_LEN("char(") },
+    { STRING_WITH_LEN("varchar(") },
     { STRING_WITH_LEN("utf8") }
   },
   {
@@ -1074,7 +1074,7 @@ sp_returns_type(THD *thd, String &result, const sp_head *sp)
   {
     result.append(STRING_WITH_LEN(" CHARSET "));
     result.append(field->charset()->csname);
-    if (Charset(field->charset()).can_have_collate_clause())
+    if (!(field->charset()->state & MY_CS_PRIMARY))
     {
       result.append(STRING_WITH_LEN(" COLLATE "));
       result.append(field->charset()->name);
@@ -1931,6 +1931,8 @@ bool
 Sp_handler::sp_show_create_routine(THD *thd,
                                    const Database_qualified_name *name) const
 {
+  sp_head *sp;
+
   DBUG_ENTER("sp_show_create_routine");
   DBUG_PRINT("enter", ("type: %s name: %.*s",
                        type_str(),
@@ -1943,28 +1945,20 @@ Sp_handler::sp_show_create_routine(THD *thd,
     It is "safe" to do as long as it doesn't affect the results
     of the binary log or the query cache, which currently it does not.
   */
-  sp_head *sp= 0;
+  if (sp_cache_routine(thd, name, false, &sp))
+    DBUG_RETURN(TRUE);
 
-  DBUG_EXECUTE_IF("cache_sp_in_show_create",
-    /* Some tests need just need a way to cache SP without other side-effects.*/
-    sp_cache_routine(thd, name, false, &sp);
-    sp->show_create_routine(thd, this);
-    DBUG_RETURN(false);
-  );
-
-  bool free_sp= db_find_routine(thd, name, &sp) == SP_OK;
-  bool ret= !sp || sp->show_create_routine(thd, this);
-  if (ret)
+  if (sp == NULL || sp->show_create_routine(thd, this))
   {
     /*
       If we have insufficient privileges, pretend the routine
       does not exist.
     */
     my_error(ER_SP_DOES_NOT_EXIST, MYF(0), type_str(), name->m_name.str);
+    DBUG_RETURN(TRUE);
   }
-  if (free_sp)
-    sp_head::destroy(sp);
-  DBUG_RETURN(ret);
+
+  DBUG_RETURN(FALSE);
 }
 
 
