@@ -1,5 +1,5 @@
 /* Copyright (c) 2010, 2015, Oracle and/or its affiliates.
-   Copyright (c) 2011, 2021, MariaDB
+   Copyright (c) 2011, 2020, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -32,9 +32,6 @@
 #include "sql_admin.h"
 #include "sql_statistics.h"
 #include "wsrep_mysqld.h"
-
-const LEX_CSTRING msg_status= {STRING_WITH_LEN("status")};
-
 /* Prepare, run and cleanup for mysql_recreate_table() */
 
 static bool admin_recreate_table(THD *thd, TABLE_LIST *table_list)
@@ -432,21 +429,48 @@ dbug_err:
   return open_error;
 }
 
-
-
-static void send_read_only_warning(THD *thd, const LEX_CSTRING *msg_status,
-                                   const LEX_CSTRING *table_name)
+#ifdef WITH_WSREP
+  /*
+     OPTIMIZE, REPAIR and ALTER may take MDL locks not only for the affected table, but
+     also for the table referenced by foreign key constraint.
+     This wsrep_toi_replication() function handles TOI replication for OPTIMIZE and REPAIR
+     so that certification keys for potential FK parent tables are also appended in the
+     write set.
+     ALTER TABLE case is handled elsewhere.
+  */
+static bool wsrep_toi_replication(THD *thd, TABLE_LIST *tables)
 {
-  Protocol *protocol= thd->protocol;
-  char buf[MYSQL_ERRMSG_SIZE];
-  size_t length;
-  length= my_snprintf(buf, sizeof(buf),
-                      ER_THD(thd, ER_OPEN_AS_READONLY),
-                      table_name->str);
-  protocol->store(msg_status, system_charset_info);
-  protocol->store(buf, length, system_charset_info);
-}
+  LEX *lex= thd->lex;
+  /* only handle OPTIMIZE and REPAIR here */
+  switch (lex->sql_command)
+  {
+  case SQLCOM_OPTIMIZE:
+  case SQLCOM_REPAIR:
+    break;
+  default:
+    return false;
+  }
 
+  close_thread_tables(thd);
+  wsrep::key_array keys;
+
+  wsrep_append_fk_parent_table(thd, tables, &keys);
+
+  /* now TOI replication, with no locks held */
+  if (keys.empty())
+  {
+    WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, tables);
+  } else {
+    WSREP_TO_ISOLATION_BEGIN_FK_TABLES(NULL, NULL, tables, &keys) {
+      return true;
+    }
+  }
+  return false;
+
+ wsrep_error_label:
+  return true;
+}
+#endif /* WITH_WSREP */
 
 /*
   RETURN VALUES
@@ -466,8 +490,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
                               int (handler::*operator_func)(THD *,
                                                             HA_CHECK_OPT *),
                               int (view_operator_func)(THD *, TABLE_LIST*,
-                                                       HA_CHECK_OPT *),
-                              bool is_cmd_replicated)
+                                                       HA_CHECK_OPT *))
 {
   TABLE_LIST *table;
   List<Item> field_list;
@@ -478,8 +501,6 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
   int compl_result_code;
   bool need_repair_or_alter= 0;
   wait_for_commit* suspended_wfc;
-  bool is_table_modified= false;
-
   DBUG_ENTER("mysql_admin_table");
   DBUG_PRINT("enter", ("extra_open_options: %u", extra_open_options));
 
@@ -518,11 +539,20 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
   close_thread_tables(thd);
   for (table= tables; table; table= table->next_local)
     table->table= NULL;
+#ifdef WITH_WSREP
+  if (WSREP(thd))
+  {
+    if(wsrep_toi_replication(thd, tables))
+    {
+      WSREP_INFO("wsrep TOI replication of has failed.");
+      goto err;
+    }
+  }
+#endif /* WITH_WSREP */
 
   for (table= tables; table; table= table->next_local)
   {
-    char table_name_buff[SAFE_NAME_LEN*2+2];
-    LEX_CSTRING table_name= { table_name_buff, 0};
+    char table_name[SAFE_NAME_LEN*2+2];
     const char *db= table->db.str;
     bool fatal_error=0;
     bool open_error;
@@ -530,10 +560,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
     bool open_for_modify= org_open_for_modify;
 
     DBUG_PRINT("admin", ("table: '%s'.'%s'", db, table->table_name.str));
-    DEBUG_SYNC(thd, "admin_command_kill_before_modify");
-
-    table_name.length= strxmov(table_name_buff, db, ".", table->table_name.str,
-                               NullS) - table_name_buff;
+    strxmov(table_name, db, ".", table->table_name.str, NullS);
     thd->open_options|= extra_open_options;
     table->lock_type= lock_type;
     /*
@@ -544,17 +571,9 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
     */
     table->mdl_request.set_type(lex->sql_command == SQLCOM_REPAIR
                                 ? MDL_SHARED_NO_READ_WRITE
-                                : lock_type >= TL_WRITE_ALLOW_WRITE
+                                : lock_type >= TL_FIRST_WRITE
                                 ? MDL_SHARED_WRITE : MDL_SHARED_READ);
 
-    if (thd->check_killed())
-    {
-      open_error= false;
-      fatal_error= true;
-      result_code= HA_ADMIN_FAILED;
-      goto send_result;
-    }
-
     /* open only one table from local list of command */
     while (1)
     {
@@ -619,12 +638,12 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
             size_t length;
             DBUG_PRINT("admin", ("sending non existent partition error"));
             protocol->prepare_for_resend();
-            protocol->store(&table_name, system_charset_info);
+            protocol->store(table_name, system_charset_info);
             protocol->store(operator_name, system_charset_info);
             protocol->store(STRING_WITH_LEN("error"), system_charset_info);
             length= my_snprintf(buff, sizeof(buff),
                                 ER_THD(thd, ER_DROP_PARTITION_NON_EXISTENT),
-                                table_name.str);
+                                table_name);
             protocol->store(buff, length, system_charset_info);
             if(protocol->write())
               goto err;
@@ -700,17 +719,20 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       goto send_result;
     }
 
-    if ((table->table->db_stat & HA_READ_ONLY) && open_for_modify &&
-        operator_func != &handler::ha_analyze)
+    if ((table->table->db_stat & HA_READ_ONLY) && open_for_modify)
     {
       /* purecov: begin inspected */
+      char buff[FN_REFLEN + MYSQL_ERRMSG_SIZE];
+      size_t length;
       enum_sql_command save_sql_command= lex->sql_command;
-      LEX_CSTRING error_clex_str= { STRING_WITH_LEN("error") };
       DBUG_PRINT("admin", ("sending error message"));
       protocol->prepare_for_resend();
-      protocol->store(&table_name, system_charset_info);
+      protocol->store(table_name, system_charset_info);
       protocol->store(operator_name, system_charset_info);
-      send_read_only_warning(thd, &error_clex_str, &table_name);
+      protocol->store(STRING_WITH_LEN("error"), system_charset_info);
+      length= my_snprintf(buff, sizeof(buff), ER_THD(thd, ER_OPEN_AS_READONLY),
+                          table_name);
+      protocol->store(buff, length, system_charset_info);
       trans_commit_stmt(thd);
       trans_commit(thd);
       close_thread_tables(thd);
@@ -765,7 +787,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       /* purecov: begin inspected */
       DBUG_PRINT("admin", ("sending crashed warning"));
       protocol->prepare_for_resend();
-      protocol->store(&table_name, system_charset_info);
+      protocol->store(table_name, system_charset_info);
       protocol->store(operator_name, system_charset_info);
       protocol->store(STRING_WITH_LEN("warning"), system_charset_info);
       protocol->store(STRING_WITH_LEN("Table is marked as crashed"),
@@ -836,15 +858,6 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
 
     if (compl_result_code == HA_ADMIN_OK && collect_eis)
     {
-      if (result_code == HA_ERR_TABLE_READONLY)
-      {
-        protocol->prepare_for_resend();
-        protocol->store(&table_name, system_charset_info);
-        protocol->store(operator_name, system_charset_info);
-        send_read_only_warning(thd, &msg_status, &table_name);
-        (void) protocol->write();
-        result_code= HA_ADMIN_OK;
-      }
       /*
         Here we close and reopen table in read mode because operation of
         collecting statistics is long and it will be better do not block
@@ -880,7 +893,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
             enum enum_field_types type= (*field_ptr)->type();
             if (type < MYSQL_TYPE_MEDIUM_BLOB ||
                 type > MYSQL_TYPE_BLOB)
-              tab->field[fields]->register_field_in_read_map();
+              bitmap_set_bit(tab->read_set, fields);
             else
               push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                                   ER_NO_EIS_FOR_FIELD,
@@ -908,7 +921,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
             enum enum_field_types type= tab->field[pos]->type();
             if (type < MYSQL_TYPE_MEDIUM_BLOB ||
                 type > MYSQL_TYPE_BLOB)
-              tab->field[pos]->register_field_in_read_map();
+              bitmap_set_bit(tab->read_set, pos);
             else
               push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                                   ER_NO_EIS_FOR_FIELD,
@@ -952,9 +965,9 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       else
       {
         protocol->prepare_for_resend();
-        protocol->store(&table_name, system_charset_info);
+        protocol->store(table_name, system_charset_info); 
         protocol->store(operator_name, system_charset_info);
-        protocol->store(&msg_status, system_charset_info);
+        protocol->store(STRING_WITH_LEN("status"), system_charset_info);
 	protocol->store(STRING_WITH_LEN("Engine-independent statistics collected"), 
                         system_charset_info);
         if (protocol->write())
@@ -981,7 +994,7 @@ send_result:
       while ((err= it++))
       {
         protocol->prepare_for_resend();
-        protocol->store(&table_name, system_charset_info);
+        protocol->store(table_name, system_charset_info);
         protocol->store((char*) operator_name, system_charset_info);
         protocol->store(warning_level_names[err->get_level()].str,
                         warning_level_names[err->get_level()].length,
@@ -993,7 +1006,7 @@ send_result:
       thd->get_stmt_da()->clear_warning_info(thd->query_id);
     }
     protocol->prepare_for_resend();
-    protocol->store(&table_name, system_charset_info);
+    protocol->store(table_name, system_charset_info);
     protocol->store(operator_name, system_charset_info);
 
 send_result_message:
@@ -1016,32 +1029,32 @@ send_result_message:
         char buf[MYSQL_ERRMSG_SIZE];
         size_t length= my_snprintf(buf, sizeof(buf),
                                    ER_THD(thd, ER_BAD_TABLE_ERROR),
-                                   table_name.str);
+                                   table_name);
         protocol->store(STRING_WITH_LEN("note"), system_charset_info);
         protocol->store(buf, length, system_charset_info);
       }
       break;
 
     case HA_ADMIN_OK:
-      protocol->store(&msg_status, system_charset_info);
+      protocol->store(STRING_WITH_LEN("status"), system_charset_info);
       protocol->store(STRING_WITH_LEN("OK"), system_charset_info);
       break;
 
     case HA_ADMIN_FAILED:
-      protocol->store(&msg_status, system_charset_info);
+      protocol->store(STRING_WITH_LEN("status"), system_charset_info);
       protocol->store(STRING_WITH_LEN("Operation failed"),
                       system_charset_info);
       break;
 
     case HA_ADMIN_REJECT:
-      protocol->store(&msg_status, system_charset_info);
+      protocol->store(STRING_WITH_LEN("status"), system_charset_info);
       protocol->store(STRING_WITH_LEN("Operation need committed state"),
                       system_charset_info);
       open_for_modify= FALSE;
       break;
 
     case HA_ADMIN_ALREADY_DONE:
-      protocol->store(&msg_status, system_charset_info);
+      protocol->store(STRING_WITH_LEN("status"), system_charset_info);
       protocol->store(STRING_WITH_LEN("Table is already up to date"),
                       system_charset_info);
       break;
@@ -1121,7 +1134,7 @@ send_result_message:
       }
       /* Start a new row for the final status row */
       protocol->prepare_for_resend();
-      protocol->store(&table_name, system_charset_info);
+      protocol->store(table_name, system_charset_info);
       protocol->store(operator_name, system_charset_info);
       if (result_code) // either mysql_recreate_table or analyze failed
       {
@@ -1142,7 +1155,7 @@ send_result_message:
               goto err;
             /* Start off another row for HA_ADMIN_FAILED */
             protocol->prepare_for_resend();
-            protocol->store(&table_name, system_charset_info);
+            protocol->store(table_name, system_charset_info);
             protocol->store(operator_name, system_charset_info);
           }
           thd->clear_error();
@@ -1186,11 +1199,7 @@ send_result_message:
       fatal_error=1;
       break;
     }
-    case HA_ERR_TABLE_READONLY:
-    {
-      send_read_only_warning(thd, &msg_status, &table_name);
-      break;
-    }
+
     default:				// Probably HA_ADMIN_INTERNAL_ERROR
       {
         char buf[MYSQL_ERRMSG_SIZE];
@@ -1203,13 +1212,6 @@ send_result_message:
         break;
       }
     }
-    /*
-      Admin commands acquire table locks and these locks are not detected by
-      parallel replication deadlock detection-and-handling mechanism. Hence
-      they must be marked as DDL so that they are not scheduled in parallel
-      with conflicting DMLs resulting in deadlock.
-    */
-    thd->transaction->stmt.mark_executed_table_admin_cmd();
     if (table->table && !table->view)
     {
       /*
@@ -1251,9 +1253,10 @@ send_result_message:
     }
     else
     {
-      if (trans_commit_stmt(thd))
+      if (trans_commit_stmt(thd) ||
+          (stmt_causes_implicit_commit(thd, CF_IMPLICIT_COMMIT_END) &&
+           trans_commit_implicit(thd)))
         goto err;
-      is_table_modified= true;
     }
     close_thread_tables(thd);
     thd->release_transactional_locks();
@@ -1276,16 +1279,6 @@ send_result_message:
 
     if (protocol->write())
       goto err;
-    DEBUG_SYNC(thd, "admin_command_kill_after_modify");
-  }
-  if (is_table_modified && is_cmd_replicated &&
-      (!opt_readonly || thd->slave_thread) && !thd->lex->no_write_to_binlog)
-  {
-    thd->get_stmt_da()->set_overwrite_status(true);
-    auto res= write_bin_log(thd, true, thd->query(), thd->query_length());
-    thd->get_stmt_da()->set_overwrite_status(false);
-    if (res)
-      goto err;
   }
 
   my_eof(thd);
@@ -1348,7 +1341,7 @@ bool mysql_assign_to_keycache(THD* thd, TABLE_LIST* tables,
   check_opt.key_cache= key_cache;
   DBUG_RETURN(mysql_admin_table(thd, tables, &check_opt,
 				"assign_to_keycache", TL_READ_NO_INSERT, 0, 0,
-				0, 0, &handler::assign_to_keycache, 0, false));
+				0, 0, &handler::assign_to_keycache, 0));
 }
 
 
@@ -1375,7 +1368,7 @@ bool mysql_preload_keys(THD* thd, TABLE_LIST* tables)
   */
   DBUG_RETURN(mysql_admin_table(thd, tables, 0,
 				"preload_keys", TL_READ_NO_INSERT, 0, 0, 0, 0,
-				&handler::preload_keys, 0, false));
+				&handler::preload_keys, 0));
 }
 
 
@@ -1396,8 +1389,17 @@ bool Sql_cmd_analyze_table::execute(THD *thd)
   WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, first_table);
   res= mysql_admin_table(thd, first_table, &m_lex->check_opt,
                          "analyze", lock_type, 1, 0, 0, 0,
-                         &handler::ha_analyze, 0, true);
-
+                         &handler::ha_analyze, 0);
+  /* ! we write after unlocking the table */
+  if (!res && !m_lex->no_write_to_binlog && (!opt_readonly || thd->slave_thread))
+  {
+    /*
+      Presumably, ANALYZE and binlog writing doesn't require synchronization
+    */
+    thd->get_stmt_da()->set_overwrite_status(true);
+    res= write_bin_log(thd, TRUE, thd->query(), thd->query_length());
+    thd->get_stmt_da()->set_overwrite_status(false);
+  }
   m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
 
@@ -1423,7 +1425,7 @@ bool Sql_cmd_check_table::execute(THD *thd)
 
   res= mysql_admin_table(thd, first_table, &m_lex->check_opt, "check",
                          lock_type, 0, 0, HA_OPEN_FOR_REPAIR, 0,
-                         &handler::ha_check, &view_check, false);
+                         &handler::ha_check, &view_check);
 
   m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
@@ -1444,19 +1446,24 @@ bool Sql_cmd_optimize_table::execute(THD *thd)
                          FALSE, UINT_MAX, FALSE))
     goto error; /* purecov: inspected */
 
-  WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, first_table);
   res= (specialflag & SPECIAL_NO_NEW_FUNC) ?
     mysql_recreate_table(thd, first_table, true) :
     mysql_admin_table(thd, first_table, &m_lex->check_opt,
                       "optimize", TL_WRITE, 1, 0, 0, 0,
-                      &handler::ha_optimize, 0, true);
-
+                      &handler::ha_optimize, 0);
+  /* ! we write after unlocking the table */
+  if (!res && !m_lex->no_write_to_binlog && (!opt_readonly || thd->slave_thread))
+  {
+    /*
+      Presumably, OPTIMIZE and binlog writing doesn't require synchronization
+    */
+    thd->get_stmt_da()->set_overwrite_status(true);
+    res= write_bin_log(thd, TRUE, thd->query(), thd->query_length());
+    thd->get_stmt_da()->set_overwrite_status(false);
+  }
   m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
 
-#ifdef WITH_WSREP
-wsrep_error_label:
-#endif /* WITH_WSREP */
 error:
   DBUG_RETURN(res);
 }
@@ -1472,20 +1479,25 @@ bool Sql_cmd_repair_table::execute(THD *thd)
   if (check_table_access(thd, SELECT_ACL | INSERT_ACL, first_table,
                          FALSE, UINT_MAX, FALSE))
     goto error; /* purecov: inspected */
-
-  WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, first_table);
   res= mysql_admin_table(thd, first_table, &m_lex->check_opt, "repair",
                          TL_WRITE, 1,
                          MY_TEST(m_lex->check_opt.sql_flags & TT_USEFRM),
                          HA_OPEN_FOR_REPAIR, &prepare_for_repair,
-                         &handler::ha_repair, &view_repair, true);
+                         &handler::ha_repair, &view_repair);
 
+  /* ! we write after unlocking the table */
+  if (!res && !m_lex->no_write_to_binlog && (!opt_readonly || thd->slave_thread))
+  {
+    /*
+      Presumably, REPAIR and binlog writing doesn't require synchronization
+    */
+    thd->get_stmt_da()->set_overwrite_status(true);
+    res= write_bin_log(thd, TRUE, thd->query(), thd->query_length());
+    thd->get_stmt_da()->set_overwrite_status(false);
+  }
   m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
 
-#ifdef WITH_WSREP
-wsrep_error_label:
-#endif /* WITH_WSREP */
 error:
   DBUG_RETURN(res);
 }
