@@ -27,6 +27,7 @@ Alter_info::Alter_info(const Alter_info &rhs, MEM_ROOT *mem_root)
   key_list(rhs.key_list, mem_root),
   alter_rename_key_list(rhs.alter_rename_key_list, mem_root),
   create_list(rhs.create_list, mem_root),
+  alter_index_ignorability_list(rhs.alter_index_ignorability_list, mem_root),
   check_constraint_list(rhs.check_constraint_list, mem_root),
   flags(rhs.flags), partition_flags(rhs.partition_flags),
   keys_onoff(rhs.keys_onoff),
@@ -259,7 +260,7 @@ Alter_table_ctx::Alter_table_ctx()
     db(null_clex_str), table_name(null_clex_str), alias(null_clex_str),
     new_db(null_clex_str), new_name(null_clex_str), new_alias(null_clex_str),
     fk_error_if_delete_row(false), fk_error_id(NULL),
-    fk_error_table(NULL), modified_primary_key(false)
+    fk_error_table(NULL)
 #ifdef DBUG_ASSERT_EXISTS
     , tmp_table(false)
 #endif
@@ -279,7 +280,7 @@ Alter_table_ctx::Alter_table_ctx(THD *thd, TABLE_LIST *table_list,
     tables_opened(tables_opened_arg),
     new_db(*new_db_arg), new_name(*new_name_arg),
     fk_error_if_delete_row(false), fk_error_id(NULL),
-    fk_error_table(NULL), modified_primary_key(false)
+    fk_error_table(NULL)
 #ifdef DBUG_ASSERT_EXISTS
     , tmp_table(false)
 #endif
@@ -503,24 +504,10 @@ bool Sql_cmd_alter_table::execute(THD *thd)
                                    lex->name.str ? lex->name.str
                                    : first_table->table_name.str,
                                    first_table, &alter_info, &keys,
-                                   used_engine ? &create_info : nullptr)
-    {
-      WSREP_WARN("ALTER TABLE isolation failure");
-      DBUG_RETURN(TRUE);
-    }
+                                   used_engine ? &create_info : nullptr);
 
-    /*
-      It makes sense to set auto_increment_* to defaults in TOI operations.
-      Must be done before wsrep_TOI_begin() since Query_log_event encapsulating
-      TOI statement and auto inc variables for wsrep replication is constructed
-      there. Variables are reset back in THD::reset_for_next_command() before
-      processing of next command.
-    */
-    if (wsrep_auto_increment_control)
-    {
-      thd->variables.auto_increment_offset = 1;
-      thd->variables.auto_increment_increment = 1;
-    }
+    thd->variables.auto_increment_offset = 1;
+    thd->variables.auto_increment_increment = 1;
   }
 #endif
 
@@ -559,6 +546,11 @@ bool Sql_cmd_alter_table::execute(THD *thd)
                             lex->ignore, lex->if_exists());
 
   DBUG_RETURN(result);
+#ifdef WITH_WSREP
+wsrep_error_label:
+  WSREP_WARN("ALTER TABLE isolation failure");
+  DBUG_RETURN(TRUE);
+#endif
 }
 
 bool Sql_cmd_discard_import_tablespace::execute(THD *thd)
