@@ -25,13 +25,13 @@
 #include "mariadb.h"                         /* NO_EMBEDDED_ACCESS_CHECKS */
 #include "sql_priv.h"
 #include "unireg.h"
+#include "sql_derived.h"
 #include "sql_select.h"
 #include "derived_handler.h"
 #include "sql_base.h"
 #include "sql_view.h"                         // check_duplicate_names
 #include "sql_acl.h"                          // SELECT_ACL
 #include "sql_class.h"
-#include "sql_derived.h"
 #include "sql_cte.h"
 #include "my_json_writer.h"
 #include "opt_trace.h"
@@ -73,6 +73,7 @@ bool
 mysql_handle_derived(LEX *lex, uint phases)
 {
   bool res= FALSE;
+  THD *thd= lex->thd;
   DBUG_ENTER("mysql_handle_derived");
   DBUG_PRINT("enter", ("phases: 0x%x", phases));
   if (!lex->derived_tables)
@@ -87,6 +88,8 @@ mysql_handle_derived(LEX *lex, uint phases)
       break;
     if (!(phases & phase_flag))
       continue;
+    if (phase_flag >= DT_CREATE && !thd->fill_derived_tables())
+      break;
 
     for (SELECT_LEX *sl= lex->all_selects_list;
 	 sl && !res;
@@ -170,6 +173,7 @@ bool
 mysql_handle_single_derived(LEX *lex, TABLE_LIST *derived, uint phases)
 {
   bool res= FALSE;
+  THD *thd= lex->thd;
   uint8 allowed_phases= (derived->is_merged_derived() ? DT_PHASES_MERGE :
                          DT_PHASES_MATERIALIZE);
   DBUG_ENTER("mysql_handle_single_derived");
@@ -196,6 +200,8 @@ mysql_handle_single_derived(LEX *lex, TABLE_LIST *derived, uint phases)
     if (phase_flag != DT_PREPARE &&
         !(allowed_phases & phase_flag))
       continue;
+    if (phase_flag >= DT_CREATE && !thd->fill_derived_tables())
+      break;
 
     if ((res= (*processors[phase])(lex->thd, lex, derived)))
       break;
@@ -376,6 +382,10 @@ bool mysql_derived_merge(THD *thd, LEX *lex, TABLE_LIST *derived)
     DBUG_RETURN(FALSE);
   }
 
+  if (thd->lex->sql_command == SQLCOM_UPDATE_MULTI ||
+      thd->lex->sql_command == SQLCOM_DELETE_MULTI)
+    thd->save_prep_leaf_list= TRUE;
+
   arena= thd->activate_stmt_arena_if_needed(&backup);  // For easier test
 
   if (!derived->merged_for_insert || 
@@ -449,7 +459,6 @@ bool mysql_derived_merge(THD *thd, LEX *lex, TABLE_LIST *derived)
       derived->on_expr= expr;
       derived->prep_on_expr= expr->copy_andor_structure(thd);
     }
-    thd->where= "on clause";
     if (derived->on_expr &&
         derived->on_expr->fix_fields_if_needed_for_bool(thd, &derived->on_expr))
     {
@@ -1358,67 +1367,6 @@ bool mysql_derived_reinit(THD *thd, LEX *lex, TABLE_LIST *derived)
 }
 
 
-/*
-  @brief
-    Given condition cond and transformer+argument, try transforming as many
-    conjuncts as possible.
-
-  @detail
-    The motivation of this function is to convert the condition that's being
-    pushed into a WHERE clause with derived_field_transformer_for_where or
-    with derived_grouping_field_transformer_for_where.
-    The transformer may fail for some sub-condition, in this case we want to
-    convert the most restrictive part of the condition that can be pushed.
-
-    This function only does it for top-level AND: conjuncts that could not be
-    converted are dropped.
-
-  @return
-    Converted condition, or NULL if nothing could be converted
-*/
-
-Item *transform_condition_or_part(THD *thd,
-                                  Item *cond,
-                                  Item_transformer transformer,
-                                  uchar *arg)
-{
-  if (cond->type() != Item::COND_ITEM ||
-      ((Item_cond*) cond)->functype() != Item_func::COND_AND_FUNC)
-  {
-    Item *new_item= cond->transform(thd, transformer, arg);
-      // Indicate that the condition is not pushable
-    if (!new_item)
-      cond->clear_extraction_flag();
-    return new_item;
-  }
-
-  List_iterator<Item> li(*((Item_cond*) cond)->argument_list());
-  Item *item;
-  while ((item=li++))
-  {
-    Item *new_item= item->transform(thd, transformer, arg);
-    if (!new_item)
-    {
-      // Indicate that the condition is not pushable
-      item->clear_extraction_flag();
-      li.remove();
-    }
-    else
-      li.replace(new_item);
-  }
-
-  switch (((Item_cond*) cond)->argument_list()->elements)
-  {
-  case 0:
-    return NULL;
-  case 1:
-    return ((Item_cond*) cond)->argument_list()->head();
-  default:
-    return cond;
-  }
-}
-
-
 /**
   @brief
     Extract condition that can be pushed into a derived table/view
@@ -1500,7 +1448,8 @@ bool pushdown_cond_for_derived(THD *thd, Item *cond, TABLE_LIST *derived)
     DBUG_RETURN(false);
 
   /* Do not push conditions into unit with global ORDER BY ... LIMIT */
-  if (unit->fake_select_lex && unit->fake_select_lex->explicit_limit)
+  if (unit->fake_select_lex &&
+      unit->fake_select_lex->limit_params.explicit_limit)
     DBUG_RETURN(false);
 
   /* Check whether any select of 'unit' allows condition pushdown */
