@@ -722,13 +722,6 @@ private:
   /** Reset the current row counter. Start counting from the first row. */
   void reset_current_row_for_warning() { m_current_row_for_warning= 1; }
 
-  ulong set_current_row_for_warning(ulong row)
-  {
-    ulong old_row= m_current_row_for_warning;
-    m_current_row_for_warning= row;
-    return old_row;
-  }
-
   /** Return the current counter value. */
   ulong current_row_for_warning() const { return m_current_row_for_warning; }
 
@@ -967,8 +960,6 @@ public:
     DA_EOF,
     /** Set whenever one calls my_ok() in PS bulk mode. */
     DA_OK_BULK,
-    /** Set whenever one calls my_eof() in PS bulk mode. */
-    DA_EOF_BULK,
     /** Set whenever one calls my_error() or my_message(). */
     DA_ERROR,
     /** Set in case of a custom response, such as one from COM_STMT_PREPARE. */
@@ -1028,20 +1019,9 @@ public:
   enum_diagnostics_status status() const { return m_status; }
 
   const char *message() const
-  {
-    DBUG_ASSERT(m_status == DA_ERROR || m_status == DA_OK ||
-                m_status == DA_OK_BULK || m_status == DA_EOF_BULK);
-    return m_message;
-  }
+  { DBUG_ASSERT(m_status == DA_ERROR || m_status == DA_OK ||
+                m_status == DA_OK_BULK); return m_message; }
 
-  bool skip_flush() const
-  {
-    DBUG_ASSERT(m_status == DA_OK || m_status == DA_OK_BULK);
-    return m_skip_flush;
-  }
-
-  void set_skip_flush()
-  { m_skip_flush= TRUE; }
 
   uint sql_errno() const
   {
@@ -1067,7 +1047,7 @@ public:
   uint statement_warn_count() const
   {
     DBUG_ASSERT(m_status == DA_OK || m_status == DA_OK_BULK ||
-                m_status == DA_EOF ||m_status == DA_EOF_BULK );
+                m_status == DA_EOF);
     return m_statement_warn_count;
   }
 
@@ -1142,9 +1122,6 @@ public:
   void opt_clear_warning_info(ulonglong query_id)
   { get_warning_info()->opt_clear(query_id); }
 
-  long set_current_row_for_warning(long row)
-  { return get_warning_info()->set_current_row_for_warning(row); }
-
   ulong current_row_for_warning() const
   { return get_warning_info()->current_row_for_warning(); }
 
@@ -1230,9 +1207,6 @@ private:
   /** Set to make set_error_status after set_{ok,eof}_status possible. */
   bool m_can_overwrite_status;
 
-  /** Skip flushing network buffer after writing OK (for COM_MULTI) */
-  bool m_skip_flush;
-
   /** Message buffer. Can be used by OK or ERROR status. */
   char m_message[MYSQL_ERRMSG_SIZE];
 
