@@ -604,29 +604,27 @@ void Explain_union::print_explain_json(Explain_query *query,
   else
     writer->add_member("union_result").start_object();
 
-  if (using_tmp)
+  // using_temporary_table
+  make_union_table_name(table_name_buffer);
+  writer->add_member("table_name").add_str(table_name_buffer);
+  writer->add_member("access_type").add_str("ALL"); // not very useful
+
+  /* r_loops (not present in tabular output) */
+  if (is_analyze)
   {
-    make_union_table_name(table_name_buffer);
-    writer->add_member("table_name").add_str(table_name_buffer);
-    writer->add_member("access_type").add_str("ALL"); // not very useful
-
-    /* r_loops (not present in tabular output) */
-    if (is_analyze)
-    {
-      writer->add_member("r_loops").add_ll(
-          fake_select_lex_tracker.get_loops());
-    }
-
-    /* `r_rows` */
-    if (is_analyze)
-    {
-      writer->add_member("r_rows");
-      if (fake_select_lex_tracker.has_scans())
-        writer->add_double(fake_select_lex_tracker.get_avg_rows());
-      else
-        writer->add_null();
-    }
+    writer->add_member("r_loops").add_ll(fake_select_lex_tracker.get_loops());
   }
+
+  /* `r_rows` */
+  if (is_analyze)
+  {
+    writer->add_member("r_rows");
+    if (fake_select_lex_tracker.has_scans())
+      writer->add_double(fake_select_lex_tracker.get_avg_rows());
+    else
+      writer->add_null();
+  }
+
   writer->add_member("query_specifications").start_array();
 
   for (int i= 0; i < (int) union_members.elements(); i++)
@@ -662,11 +660,7 @@ int Explain_node::print_explain_for_children(Explain_query *query,
   for (int i= 0; i < (int) children.elements(); i++)
   {
     Explain_node *node= query->get_node(children.at(i));
-    /*
-      Note: node may not be present because for certain kinds of subqueries,
-      the optimizer is not able to see that they were eliminated.
-    */
-    if (node && node->print_explain(query, output, explain_flags, is_analyze))
+    if (node->print_explain(query, output, explain_flags, is_analyze))
       return 1;
   }
   return 0;
@@ -710,15 +704,8 @@ void Explain_node::print_explain_json_for_children(Explain_query *query,
   for (int i= 0; i < (int) children.elements(); i++)
   {
     Explain_node *node= query->get_node(children.at(i));
-    
-    /*
-      Note: node may not be present because for certain kinds of subqueries,
-      the optimizer is not able to see that they were eliminated.
-    */
-    if (!node)
-      continue;
-
     /* Derived tables are printed inside Explain_table_access objects */
+    
     if (!is_connection_printable_in_json(node->connection_type))
       continue;
 
@@ -1648,6 +1635,9 @@ void Explain_table_access::tag_to_json(Json_writer *writer, enum explain_extra_t
     case ET_DISTINCT:
       writer->add_member("distinct").add_bool(true);
       break;
+    case ET_TABLE_FUNCTION:
+      writer->add_member("table_function").add_str("json_table");
+      break;
 
     default:
       DBUG_ASSERT(0);
@@ -1676,7 +1666,6 @@ void Explain_rowid_filter::print_explain_json(Explain_query *query,
   if (is_analyze)
   {
     writer->add_member("r_rows").add_double(tracker->get_container_elements());
-    writer->add_member("r_lookups").add_ll(tracker->get_container_lookups());
     writer->add_member("r_selectivity_pct").
       add_double(tracker->get_r_selectivity_pct() * 100.0);
     writer->add_member("r_buffer_size").
@@ -1936,7 +1925,7 @@ void Explain_table_access::print_explain_json(Explain_query *query,
 
 
 /*
-  Elements in this array match members of enum explain_extra_tag, defined in
+  Elements in this array match members of enum Extra_tag, defined in
   sql_explain.h
 */
 
@@ -2041,6 +2030,9 @@ void Explain_table_access::append_tag_name(String *str, enum explain_extra_tag t
       if (loose_scan_is_scanning)
         str->append(" (scanning)");
       break;
+    case ET_TABLE_FUNCTION:
+      str->append("Table function: json_table");
+      break;
     }
     default:
      str->append(extra_tag_text[tag]);
