@@ -551,7 +551,7 @@ bool mysql_create_or_drop_trigger(THD *thd, TABLE_LIST *tables, bool create)
 
 #ifdef WITH_WSREP
   if (WSREP(thd) &&
-      !wsrep_should_replicate_ddl(thd, table->s->db_type()->db_type))
+      !wsrep_should_replicate_ddl(thd, table->s->db_type()))
     goto wsrep_error_label;
 #endif
 
@@ -2270,7 +2270,7 @@ add_tables_and_routines_for_triggers(THD *thd,
                                      TABLE_LIST *table_list)
 {
   DBUG_ASSERT(static_cast<int>(table_list->lock_type) >=
-              static_cast<int>(TL_WRITE_ALLOW_WRITE));
+              static_cast<int>(TL_FIRST_WRITE));
 
   for (int i= 0; i < (int)TRG_EVENT_MAX; i++)
   {
