@@ -30,7 +30,6 @@
 #include "filesort.h"                           // filesort_free_buffers
 #include "sql_view.h"
 #include "sql_cte.h"
-#include "item_windowfunc.h"
 
 bool mysql_union(THD *thd, LEX *lex, select_result *result,
                  SELECT_LEX_UNIT *unit, ulong setup_tables_done_option)
@@ -251,27 +250,7 @@ bool select_unit::send_eof()
 
 int select_union_recursive::send_data(List<Item> &values)
 {
-
-  int rc;
-  bool save_abort_on_warning= thd->abort_on_warning;
-  enum_check_fields save_count_cuted_fields= thd->count_cuted_fields;
-  long save_counter;
-
-  /*
-    For recursive CTE's give warnings for wrong field info
-    However, we don't do that for CREATE TABLE ... SELECT or INSERT ... SELECT
-    as the upper level code for these handles setting of abort_on_warning
-    depending on if 'IGNORE' is used.
-  */
-  if (thd->lex->sql_command != SQLCOM_CREATE_TABLE &&
-      thd->lex->sql_command != SQLCOM_INSERT_SELECT)
-    thd->abort_on_warning= thd->is_strict_mode();
-  thd->count_cuted_fields= CHECK_FIELD_WARN;
-  save_counter= thd->get_stmt_da()->set_current_row_for_warning(++row_counter);
-  rc= select_unit::send_data(values);
-  thd->get_stmt_da()->set_current_row_for_warning(save_counter);
-  thd->count_cuted_fields= save_count_cuted_fields;
-  thd->abort_on_warning= save_abort_on_warning;
+  int rc= select_unit::send_data(values);
 
   if (rc == 0 &&
       write_err != HA_ERR_FOUND_DUPP_KEY &&
@@ -938,7 +917,6 @@ void select_union_recursive::cleanup()
     thd->rec_tables= tab;
     tbl->derived_result= 0;
   }
-  row_counter= 0;
 }
 
 
@@ -1105,7 +1083,8 @@ bool st_select_lex_unit::prepare_join(THD *thd_arg, SELECT_LEX *sl,
 
   thd_arg->lex->current_select= sl;
 
-  can_skip_order_by= is_union_select && !(sl->braces && sl->explicit_limit);
+  can_skip_order_by= is_union_select && !(sl->braces &&
+                                          sl->limit_params.explicit_limit);
 
   saved_error= join->prepare(sl->table_list.first,
                              (derived && derived->merged ? NULL : sl->where),
@@ -1361,7 +1340,7 @@ bool st_select_lex_unit::prepare(TABLE_LIST *derived_arg,
 	else
 	{
 	  sl->join->result= result;
-          lim.set_unlimited();
+          lim.clear();
 	  if (!sl->join->procedure &&
 	      result->prepare(sl->join->fields_list, this))
 	  {
@@ -1471,7 +1450,7 @@ bool st_select_lex_unit::prepare(TABLE_LIST *derived_arg,
         if (fake_select_lex)
 	{
           if (fake_select_lex->order_list.first ||
-              fake_select_lex->explicit_limit)
+              fake_select_lex->limit_params.explicit_limit)
           {
 	    my_error(ER_NOT_SUPPORTED_YET, MYF(0),
                      "global ORDER_BY/LIMIT in recursive CTE spec");
@@ -1531,7 +1510,7 @@ bool st_select_lex_unit::prepare(TABLE_LIST *derived_arg,
       if (!unit->first_select()->next_select())
       {
         if (!unit->fake_select_lex)
-	{
+        {
           Query_arena *arena, backup_arena;
           arena= thd->activate_stmt_arena_if_needed(&backup_arena);
           bool rc= unit->add_fake_select_lex(thd);
@@ -1542,17 +1521,13 @@ bool st_select_lex_unit::prepare(TABLE_LIST *derived_arg,
         }
         SELECT_LEX *fake= unit->fake_select_lex;
         fake->order_list= sl->order_list;
-        fake->explicit_limit= sl->explicit_limit;
-        fake->select_limit= sl->select_limit;
-        fake->offset_limit= sl->offset_limit;
+        fake->limit_params= sl->limit_params;
         sl->order_list.empty();
-        sl->explicit_limit= 0;
-        sl->select_limit= 0;
-        sl->offset_limit= 0;
+        sl->limit_params.clear();
         if (describe)
           fake->options|= SELECT_DESCRIBE;
       }
-      else if (!sl->explicit_limit)
+      else if (!sl->limit_params.explicit_limit)
         sl->order_list.empty();
     }
   }
@@ -1709,8 +1684,7 @@ cont:
         Test if the aggregated data type is OK for a UNION element.
         E.g. in case of string data, DERIVATION_NONE is not allowed.
       */
-      if (type->type() == Item::TYPE_HOLDER && type->type_handler()->
-            union_element_finalize(static_cast<Item_type_holder*>(type)))
+      if (type->type_handler()->union_element_finalize(type))
         goto err;
     }
     
@@ -2584,8 +2558,7 @@ bool st_select_lex_unit::cleanup()
   {
     DBUG_RETURN(FALSE);
   }
-  if (with_element && with_element->is_recursive && union_result &&
-      with_element->rec_outer_references)
+  if (with_element && with_element->is_recursive && union_result)
   {
     select_union_recursive *result= with_element->rec_result;
     if (++result->cleanup_count == with_element->rec_outer_references)
@@ -2754,29 +2727,6 @@ static void cleanup_order(ORDER *order)
 }
 
 
-static void cleanup_window_funcs(List<Item_window_func> &win_funcs)
-{
-  List_iterator_fast<Item_window_func> it(win_funcs);
-  Item_window_func *win_func;
-  while ((win_func= it++))
-  {
-    Window_spec *win_spec= win_func->window_spec;
-    if (!win_spec)
-      continue;
-    if (win_spec->save_partition_list)
-    {
-      win_spec->partition_list= win_spec->save_partition_list;
-      win_spec->save_partition_list= NULL;
-    }
-    if (win_spec->save_order_list)
-    {
-      win_spec->order_list= win_spec->save_order_list;
-      win_spec->save_order_list= NULL;
-    }
-  }
-}
-
-
 bool st_select_lex::cleanup()
 {
   bool error= FALSE;
@@ -2788,8 +2738,6 @@ bool st_select_lex::cleanup()
   cleanup_order(group_list.first);
   cleanup_ftfuncs(this);
 
-  cleanup_window_funcs(window_funcs);
-
   if (join)
   {
     List_iterator<TABLE_LIST> ti(leaf_tables);
@@ -2816,8 +2764,7 @@ bool st_select_lex::cleanup()
   for (SELECT_LEX_UNIT *lex_unit= first_inner_unit(); lex_unit ;
        lex_unit= lex_unit->next_unit())
   {
-    if (lex_unit->with_element && lex_unit->with_element->is_recursive &&
-        lex_unit->with_element->rec_outer_references)
+    if (lex_unit->with_element && lex_unit->with_element->is_recursive)
       continue;
     error= (bool) ((uint) error | (uint) lex_unit->cleanup());
   }
