@@ -1,5 +1,5 @@
 /* Copyright (c) 2000, 2016, Oracle and/or its affiliates.
-   Copyright (c) 2011, 2022, MariaDB
+   Copyright (c) 2011, 2021, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -272,7 +272,7 @@ static void prepare_record_for_error_message(int error, TABLE *table)
 
   /* Create unique_map with all fields used by that index. */
   my_bitmap_init(&unique_map, unique_map_buf, table->s->fields, FALSE);
-  table->mark_index_columns(keynr, &unique_map);
+  table->mark_columns_used_by_index(keynr, &unique_map);
 
   /* Subtract read_set and write_set. */
   bitmap_subtract(&unique_map, table->read_set);
@@ -370,8 +370,7 @@ int mysql_update(THD *thd,
                  ha_rows *found_return, ha_rows *updated_return)
 {
   bool		using_limit= limit != HA_POS_ERROR;
-  bool          safe_update= (thd->variables.option_bits & OPTION_SAFE_UPDATES)
-                             && !thd->lex->describe;
+  bool          safe_update= thd->variables.option_bits & OPTION_SAFE_UPDATES;
   bool          used_key_is_modified= FALSE, transactional_table;
   bool          will_batch= FALSE;
   bool		can_compare_record;
@@ -427,7 +426,7 @@ int mysql_update(THD *thd,
     DBUG_ASSERT(update_source_table || table_list->view != 0);
     DBUG_PRINT("info", ("Switch to multi-update"));
     /* pass counter value */
-    thd->lex->table_count_update= table_count;
+    thd->lex->table_count= table_count;
     if (thd->lex->period_conditions.is_set())
     {
       my_error(ER_NOT_SUPPORTED_YET, MYF(0),
@@ -469,8 +468,6 @@ int mysql_update(THD *thd,
   want_privilege= (table_list->view ? UPDATE_ACL :
                    table_list->grant.want_privilege);
 #endif
-  promote_select_describe_flag_if_needed(thd->lex);
-
   if (mysql_prepare_update(thd, table_list, &conds, order_num, order))
     DBUG_RETURN(1);
 
@@ -598,7 +595,7 @@ int mysql_update(THD *thd,
   }
 
   /* If running in safe sql mode, don't allow updates without keys */
-  if (!select || !select->quick)
+  if (table->opt_range_keys.is_clear_all())
   {
     thd->set_status_no_index_used();
     if (safe_update && !using_limit)
@@ -755,20 +752,15 @@ int mysql_update(THD *thd,
       !table->check_virtual_columns_marked_for_write())
   {
     DBUG_PRINT("info", ("Trying direct update"));
-    bool use_direct_update= !select || !select->cond;
-    if (!use_direct_update &&
-        (select->cond->used_tables() & ~RAND_TABLE_BIT) == table->map)
+    if (select && select->cond &&
+        (select->cond->used_tables() == table->map))
     {
       DBUG_ASSERT(!table->file->pushed_cond);
       if (!table->file->cond_push(select->cond))
-      {
-        use_direct_update= TRUE;
         table->file->pushed_cond= select->cond;
-      }
     }
 
-    if (use_direct_update &&
-        !table->file->info_push(INFO_KIND_UPDATE_FIELDS, &fields) &&
+    if (!table->file->info_push(INFO_KIND_UPDATE_FIELDS, &fields) &&
         !table->file->info_push(INFO_KIND_UPDATE_VALUES, &values) &&
         !table->file->direct_update_rows_init(&fields))
     {
@@ -1721,8 +1713,14 @@ bool Multiupdate_prelocking_strategy::handle_end(THD *thd)
   */
 
   if (setup_tables_and_check_access(thd, &select_lex->context,
-      &select_lex->top_join_list, table_list, select_lex->leaf_tables,
-      FALSE, UPDATE_ACL, SELECT_ACL, TRUE))
+        &select_lex->top_join_list, table_list, select_lex->leaf_tables,
+        FALSE, UPDATE_ACL, SELECT_ACL, FALSE))
+    DBUG_RETURN(1);
+
+  if (select_lex->handle_derived(thd->lex, DT_MERGE))
+    DBUG_RETURN(1);
+
+  if (thd->lex->save_prep_leaf_tables())
     DBUG_RETURN(1);
 
   List<Item> *fields= &lex->first_select_lex()->item_list;
@@ -1851,7 +1849,7 @@ int mysql_multi_update_prepare(THD *thd)
   TABLE_LIST *table_list= lex->query_tables;
   TABLE_LIST *tl;
   Multiupdate_prelocking_strategy prelocking_strategy;
-  uint table_count= lex->table_count_update;
+  uint table_count= lex->table_count;
   DBUG_ENTER("mysql_multi_update_prepare");
 
   /*
@@ -1946,16 +1944,9 @@ bool mysql_multi_update(THD *thd, TABLE_LIST *table_list, List<Item> *fields,
     DBUG_RETURN(TRUE);
   }
 
-  if ((*result)->init(thd))
-    DBUG_RETURN(1);
-
   thd->abort_on_warning= !ignore && thd->is_strict_mode();
   List<Item> total_list;
 
-  if (setup_tables(thd, &select_lex->context, &select_lex->top_join_list,
-                   table_list, select_lex->leaf_tables, FALSE, FALSE))
-    DBUG_RETURN(1);
-
   if (select_lex->vers_setup_conds(thd, table_list))
     DBUG_RETURN(1);
 
@@ -1997,24 +1988,6 @@ multi_update::multi_update(THD *thd_arg, TABLE_LIST *table_list,
 }
 
 
-bool multi_update::init(THD *thd)
-{
-  table_map tables_to_update= get_table_map(fields);
-  List_iterator_fast<TABLE_LIST> li(*leaves);
-  TABLE_LIST *tbl;
-  while ((tbl =li++))
-  {
-    if (tbl->is_jtbm())
-      continue;
-    if (!(tbl->table->map & tables_to_update))
-      continue;
-    if (updated_leaves.push_back(tbl, thd->mem_root))
-      return true;
-  }
-  return false;
-}
-
-
 /*
   Connect fields with tables and create list of tables that are updated
 */
@@ -2031,7 +2004,7 @@ int multi_update::prepare(List<Item> &not_used_values,
   List_iterator_fast<Item> value_it(*values);
   uint i, max_fields;
   uint leaf_table_count= 0;
-  List_iterator<TABLE_LIST> ti(updated_leaves);
+  List_iterator<TABLE_LIST> ti(*leaves);
   DBUG_ENTER("multi_update::prepare");
 
   if (prepared)
@@ -2116,9 +2089,10 @@ int multi_update::prepare(List<Item> &not_used_values,
       if (!tl)
 	DBUG_RETURN(1);
       update.link_in_list(tl, &tl->next_local);
-      table_ref->shared= tl->shared= table_count++;
+      tl->shared= table_count++;
       table->no_keyread=1;
       table->covering_keys.clear_all();
+      table->pos_in_table_list= tl;
       table->prepare_triggers_for_update_stmt_or_event();
       table->reset_default_fields();
     }
@@ -2288,11 +2262,6 @@ multi_update::initialize_tables(JOIN *join)
   if (unlikely((thd->variables.option_bits & OPTION_SAFE_UPDATES) &&
                error_if_full_join(join)))
     DBUG_RETURN(1);
-  if (join->implicit_grouping)
-  {
-    my_error(ER_INVALID_GROUP_FUNC_USE, MYF(0));
-    DBUG_RETURN(1);
-  }
   main_table=join->join_tab->table;
   table_to_update= 0;
 
