@@ -1,4 +1,4 @@
-/* Copyright (C) 2015-2021 Codership Oy <info@codership.com>
+/* Copyright (C) 2015-2019 Codership Oy <info@codership.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,7 +30,6 @@
 #include "wsrep_high_priority_service.h"
 #include "wsrep_storage_service.h"
 #include "wsrep_thd.h"
-#include "wsrep_server_state.h"
 
 #include <string>
 #include <sstream>
@@ -55,7 +54,7 @@ static const std::string create_cluster_table_str=
   "view_seqno BIGINT NOT NULL,"
   "protocol_version INT NOT NULL,"
   "capabilities INT NOT NULL"
-  ") ENGINE=InnoDB STATS_PERSISTENT=0";
+  ") ENGINE=InnoDB";
 
 static const std::string create_members_table_str=
   "CREATE TABLE IF NOT EXISTS " + wsrep_schema_str + "." + members_table_str +
@@ -64,7 +63,7 @@ static const std::string create_members_table_str=
   "cluster_uuid CHAR(36) NOT NULL,"
   "node_name CHAR(32) NOT NULL,"
   "node_incoming_address VARCHAR(256) NOT NULL"
-  ") ENGINE=InnoDB STATS_PERSISTENT=0";
+  ") ENGINE=InnoDB";
 
 #ifdef WSREP_SCHEMA_MEMBERS_HISTORY
 static const std::string cluster_member_history_table_str= "wsrep_cluster_member_history";
@@ -77,7 +76,7 @@ static const std::string create_members_history_table_str=
   "last_view_seqno BIGINT NOT NULL,"
   "node_name CHAR(32) NOT NULL,"
   "node_incoming_address VARCHAR(256) NOT NULL"
-  ") ENGINE=InnoDB STATS_PERSISTENT=0";
+  ") ENGINE=InnoDB";
 #endif /* WSREP_SCHEMA_MEMBERS_HISTORY */
 
 static const std::string create_frag_table_str=
@@ -89,7 +88,7 @@ static const std::string create_frag_table_str=
   "flags INT NOT NULL, "
   "frag LONGBLOB NOT NULL, "
   "PRIMARY KEY (node_uuid, trx_id, seqno)"
-  ") ENGINE=InnoDB STATS_PERSISTENT=0";
+  ") ENGINE=InnoDB";
 
 static const std::string delete_from_cluster_table=
   "DELETE FROM " + wsrep_schema_str + "." + cluster_table_str;
@@ -97,26 +96,6 @@ static const std::string delete_from_cluster_table=
 static const std::string delete_from_members_table=
   "DELETE FROM " + wsrep_schema_str + "." + members_table_str;
 
-/* For rolling upgrade we need to use ALTER. We do not want
-persistent statistics to be collected from these tables. */
-static const std::string alter_cluster_table=
-  "ALTER TABLE " + wsrep_schema_str + "." + cluster_table_str +
-  " STATS_PERSISTENT=0";
-
-static const std::string alter_members_table=
-  "ALTER TABLE " + wsrep_schema_str + "." + members_table_str +
-  " STATS_PERSISTENT=0";
-
-#ifdef WSREP_SCHEMA_MEMBERS_HISTORY
-static const std::string alter_members_history_table=
-  "ALTER TABLE " + wsrep_schema_str + "." + members_history_table_str +
-  " STATS_PERSISTENT=0";
-#endif
-
-static const std::string alter_frag_table=
-  "ALTER TABLE " + wsrep_schema_str + "." + sr_table_str +
-  " STATS_PERSISTENT=0";
-
 namespace Wsrep_schema_impl
 {
 
@@ -235,7 +214,7 @@ static int execute_SQL(THD* thd, const char* sql, uint length) {
     thd->set_query((char*)sql, length);
     thd->set_query_id(next_query_id());
 
-    mysql_parse(thd, (char*)sql, length, & parser_state, FALSE, FALSE);
+    mysql_parse(thd, (char*)sql, length, & parser_state);
 
     if (thd->is_error()) {
       WSREP_WARN("Wsrep_schema::execute_sql() failed, %d %s\nSQL: %s",
@@ -290,15 +269,14 @@ static int open_table(THD* thd,
                         NULL, lock_type);
   thd->lex->query_tables_own_last= 0;
 
-  // No need to open table if the query was bf aborted,
-  // thd client will get ER_LOCK_DEADLOCK in the end.
-  const bool interrupted= thd->killed ||
-       (thd->is_error() &&
-       (thd->get_stmt_da()->sql_errno() == ER_QUERY_INTERRUPTED));
-
-  if (interrupted ||
-      !open_n_lock_single_table(thd, &tables, tables.lock_type, flags)) {
+  if (!open_n_lock_single_table(thd, &tables, tables.lock_type, flags)) {
+    if (thd->is_error()) {
+      WSREP_WARN("Can't lock table %s.%s : %d (%s)",
+                 schema_name->str, table_name->str,
+                 thd->get_stmt_da()->sql_errno(), thd->get_stmt_da()->message());
+    }
     close_thread_tables(thd);
+    my_error(ER_NO_SUCH_TABLE, MYF(0), schema_name->str, table_name->str);
     DBUG_RETURN(1);
   }
 
@@ -314,15 +292,8 @@ static int open_for_write(THD* thd, const char* table_name, TABLE** table) {
   LEX_CSTRING table_str= { table_name, strlen(table_name) };
   if (Wsrep_schema_impl::open_table(thd, &schema_str, &table_str, TL_WRITE,
                                     table)) {
-    // No need to log an error if the query was bf aborted,
-    // thd client will get ER_LOCK_DEADLOCK in the end.
-    const bool interrupted= thd->killed ||
-      (thd->is_error() &&
-       (thd->get_stmt_da()->sql_errno() == ER_QUERY_INTERRUPTED));
-    if (!interrupted) {
-      WSREP_ERROR("Failed to open table %s.%s for writing",
-                  schema_str.str, table_name);
-    }
+    WSREP_ERROR("Failed to open table %s.%s for writing",
+                schema_str.str, table_name);
     return 1;
   }
   empty_record(*table);
@@ -624,11 +595,9 @@ static int init_for_index_scan(TABLE* table, const uchar* key,
  */
 static int end_index_scan(TABLE* table) {
   int error;
-  if (table->file->inited) {
-    if ((error= table->file->ha_index_end())) {
-      WSREP_ERROR("Failed to end scan: %d", error);
-      return 1;
-    }
+  if ((error= table->file->ha_index_end())) {
+    WSREP_ERROR("Failed to end scan: %d", error);
+    return 1;
   }
   return 0;
 }
@@ -705,27 +674,13 @@ int Wsrep_schema::init()
       Wsrep_schema_impl::execute_SQL(thd,
                                      create_members_history_table_str.c_str(),
                                      create_members_history_table_str.size()) ||
-      Wsrep_schema_impl::execute_SQL(thd,
-                                     alter_members_history_table.c_str(),
-                                     alter_members_history_table.size()) ||
 #endif /* WSREP_SCHEMA_MEMBERS_HISTORY */
       Wsrep_schema_impl::execute_SQL(thd,
                                      create_frag_table_str.c_str(),
-                                     create_frag_table_str.size()) ||
-      Wsrep_schema_impl::execute_SQL(thd,
-                                     alter_cluster_table.c_str(),
-                                     alter_cluster_table.size()) ||
-      Wsrep_schema_impl::execute_SQL(thd,
-                                     alter_members_table.c_str(),
-                                     alter_members_table.size()) ||
-      Wsrep_schema_impl::execute_SQL(thd,
-                                     alter_frag_table.c_str(),
-	                             alter_frag_table.size()))
-  {
+                                     create_frag_table_str.size())) {
     ret= 1;
   }
-  else
-  {
+  else {
     ret= 0;
   }
 
