@@ -1,5 +1,4 @@
-/* Copyright 2008-2022 Codership Oy <http://www.codership.com>
-   Copyright (c) 2008, 2022, MariaDB
+/* Copyright 2008-2020 Codership Oy <http://www.codership.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,11 +29,10 @@
 #include "wsrep_utils.h"
 #include "wsrep_xid.h"
 #include "wsrep_thd.h"
-#include "wsrep_server_state.h"
+#include "wsrep_mysqld.h"
 
 #include <cstdio>
 #include <cstdlib>
-#include "debug_sync.h"
 
 #include <my_service_manager.h>
 
@@ -731,49 +729,6 @@ static int sst_append_env_var(wsp::env&   env,
   return -env.error();
 }
 
-#ifdef __WIN__
-/*
-  Space, single quote, ampersand, backquote, I/O redirection
-  characters, caret, all brackets, plus, exclamation and comma
-  characters require text to be enclosed in double quotes:
-*/
-#define IS_SPECIAL(c) \
-  (isspace(c) || c == '\'' || c == '&' || c == '`' || c == '|' || \
-                 c ==  '>' || c == '<' || c == ';' || c == '^' || \
-                 c ==  '[' || c == ']' || c == '{' || c == '}' || \
-                 c ==  '(' || c == ')' || c == '+' || c == '!' || \
-                 c ==  ',')
-/*
-  Inside values, equals character are interpreted as special
-  character and requires quotation:
-*/
-#define IS_SPECIAL_V(c) (IS_SPECIAL(c) || c == '=')
-/*
-  Double quotation mark and percent characters require escaping:
-*/
-#define IS_REQ_ESCAPING(c) (c == '""' || c == '%')
-#else
-/*
-  Space, single quote, ampersand, backquote, and I/O redirection
-  characters require text to be enclosed in double quotes. The
-  semicolon is used to separate shell commands, so it must be
-  enclosed in double quotes as well:
-*/
-#define IS_SPECIAL(c) \
-  (isspace(c) || c == '\'' || c == '&' || c == '`' || c == '|' || \
-                 c ==  '>' || c == '<' || c == ';')
-/*
-  Inside values, characters are interpreted as in parameter names:
-*/
-#define IS_SPECIAL_V(c) IS_SPECIAL(c)
-/*
-  Double quotation mark and backslash characters require
-  backslash prefixing, the dollar symbol is used to substitute
-  a variable value, therefore it also requires escaping:
-*/
-#define IS_REQ_ESCAPING(c) (c == '"' || c == '\\' || c == '$')
-#endif
-
 static size_t estimate_cmd_len (bool* extra_args)
 {
   /*
@@ -798,16 +753,22 @@ static size_t estimate_cmd_len (bool* extra_args)
       char c;
       while ((c = *arg++) != 0)
       {
-        if (IS_SPECIAL(c))
-        {
-          quotation= true;
-        }
-        else if (IS_REQ_ESCAPING(c))
-        {
-          cmd_len++;
+        /*
+          Space, single quote, ampersand, and I/O redirection characters
+          require text to be enclosed in double quotes:
+        */
+        if (isspace(c) || c == '\'' || c == '&' || c == '|' ||
 #ifdef __WIN__
-          quotation= true;
+                          c == '>'  || c == '<')
+#else
+        /*
+          The semicolon is used to separate shell commands, so it must be
+          enclosed in double quotes as well:
+        */
+                          c == '>'  || c == '<' || c == ';')
 #endif
+        {
+          quotation= true;
         }
         /*
           If the equals symbol is encountered, then we need to separately
@@ -827,20 +788,58 @@ static size_t estimate_cmd_len (bool* extra_args)
           }
           while ((c = *arg++) != 0)
           {
-            if (IS_SPECIAL_V(c))
+            /*
+              Space, single quote, ampersand, and I/O redirection characters
+              require text to be enclosed in double quotes:
+            */
+            if (isspace(c) || c == '\'' || c == '&' || c == '|' ||
+#ifdef __WIN__
+                              c == '>'  || c == '<')
+#else
+            /*
+              The semicolon is used to separate shell commands, so it must be
+              enclosed in double quotes as well:
+            */
+                              c == '>'  || c == '<' || c == ';')
+#endif
             {
               quotation= true;
             }
-            else if (IS_REQ_ESCAPING(c))
+            /*
+              Double quotation mark or backslash symbol requires backslash
+              prefixing:
+            */
+#ifdef __WIN__
+            else if (c == '"' || c == '\\')
+#else
+            /*
+              The dollar symbol is used to substitute a variable, therefore
+              it also requires escaping:
+            */
+            else if (c == '"' || c == '\\' || c == '$')
+#endif
             {
               cmd_len++;
-#ifdef __WIN__
-              quotation= true;
-#endif
             }
           }
           break;
         }
+        /*
+          Double quotation mark or backslash symbol requires backslash
+          prefixing:
+        */
+#ifdef __WIN__
+        else if (c == '"' || c == '\\')
+#else
+        /*
+          The dollar symbol is used to substitute a variable, therefore
+          it also requires escaping:
+        */
+        else if (c == '"' || c == '\\' || c == '$')
+#endif
+        {
+          cmd_len++;
+        }
       }
       /* Perhaps we need to quote the entire argument or its right part: */
       if (quotation)
@@ -883,16 +882,22 @@ static void copy_orig_argv (char* cmd_str)
       char c;
       while ((c = *arg_scan++) != 0)
       {
-        if (IS_SPECIAL(c))
-        {
-          quotation= true;
-        }
-        else if (IS_REQ_ESCAPING(c))
-        {
-          plain= false;
+        /*
+          Space, single quote, ampersand, and I/O redirection characters
+          require text to be enclosed in double quotes:
+        */
+        if (isspace(c) || c == '\'' || c == '&' || c == '|' ||
 #ifdef __WIN__
-          quotation= true;
+                          c == '>'  || c == '<')
+#else
+        /*
+          The semicolon is used to separate shell commands, so it must be
+          enclosed in double quotes as well:
+        */
+                          c == '>'  || c == '<' || c == ';')
 #endif
+        {
+          quotation= true;
         }
         /*
           If the equals symbol is encountered, then we need to separately
@@ -928,13 +933,13 @@ static void copy_orig_argv (char* cmd_str)
               while (m)
               {
                 c = *arg++;
-                if (IS_REQ_ESCAPING(c))
-                {
 #ifdef __WIN__
-                  *cmd_str++ = c;
+                if (c == '"' || c == '\\')
 #else
-                  *cmd_str++ = '\\';
+                if (c == '"' || c == '\\' || c == '$')
 #endif
+                {
+                  *cmd_str++ = '\\';
                 }
                 *cmd_str++ = c;
                 m--;
@@ -963,20 +968,58 @@ static void copy_orig_argv (char* cmd_str)
           /* Let's deal with the left side of the expression: */
           while ((c = *arg_scan++) != 0)
           {
-            if (IS_SPECIAL_V(c))
+            /*
+              Space, single quote, ampersand, and I/O redirection characters
+              require text to be enclosed in double quotes:
+            */
+            if (isspace(c) || c == '\'' || c == '&' || c == '|' ||
+#ifdef __WIN__
+                              c == '>'  || c == '<')
+#else
+            /*
+              The semicolon is used to separate shell commands, so it must be
+              enclosed in double quotes as well:
+            */
+                              c == '>'  || c == '<' || c == ';')
+#endif
             {
               quotation= true;
             }
-            else if (IS_REQ_ESCAPING(c))
+            /*
+              Double quotation mark or backslash symbol requires backslash
+              prefixing:
+            */
+#ifdef __WIN__
+            else if (c == '"' || c == '\\')
+#else
+            /*
+              The dollar symbol is used to substitute a variable, therefore
+              it also requires escaping:
+            */
+            else if (c == '"' || c == '\\' || c == '$')
+#endif
             {
               plain= false;
-#ifdef __WIN__
-              quotation= true;
-#endif
             }
           }
           break;
         }
+        /*
+          Double quotation mark or backslash symbol requires backslash
+          prefixing:
+        */
+#ifdef __WIN__
+        else if (c == '"' || c == '\\')
+#else
+        /*
+          The dollar symbol is used to substitute a variable, therefore
+          it also requires escaping:
+        */
+        else if (c == '"' || c == '\\' || c == '$')
+#endif
+        {
+          plain= false;
+        }
       }
       if (n)
       {
@@ -999,13 +1042,13 @@ static void copy_orig_argv (char* cmd_str)
         {
           while ((c = *arg++) != 0)
           {
-            if (IS_REQ_ESCAPING(c))
-            {
 #ifdef __WIN__
-              *cmd_str++ = c;
+            if (c == '"' || c == '\\')
 #else
-              *cmd_str++ = '\\';
+            if (c == '"' || c == '\\' || c == '$')
 #endif
+            {
+              *cmd_str++ = '\\';
             }
             *cmd_str++ = c;
           }
@@ -1057,8 +1100,6 @@ static ssize_t sst_prepare_other (const char*  method,
   {
     WSREP_ERROR("sst_prepare_other(): generate_binlog_index_opt_val() failed %d",
                 ret);
-    if (binlog_opt_val) my_free(binlog_opt_val);
-    return ret;
   }
 
   make_wsrep_defaults_file();
@@ -1069,16 +1110,13 @@ static ssize_t sst_prepare_other (const char*  method,
                  WSREP_SST_OPT_ADDR " '%s' "
                  WSREP_SST_OPT_DATA " '%s' "
                  "%s"
-                 WSREP_SST_OPT_PARENT " %d "
-                 WSREP_SST_OPT_PROGRESS " %d"
+                 WSREP_SST_OPT_PARENT " '%d'"
                  "%s"
                  "%s",
                  method, addr_in, mysql_real_data_home,
                  wsrep_defaults_file,
                  (int)getpid(),
-                 0,
                  binlog_opt_val, binlog_index_opt_val);
-
   my_free(binlog_opt_val);
   my_free(binlog_index_opt_val);
 
@@ -1227,46 +1265,35 @@ std::string wsrep_sst_prepare()
   /*
     Figure out SST receive address. Common for all SST methods.
   */
-  wsp::Address* addr_in_parser= NULL;
-
   // Attempt 1: wsrep_sst_receive_address
   if (wsrep_sst_receive_address &&
-      strcmp(wsrep_sst_receive_address, WSREP_SST_ADDRESS_AUTO))
+      strcmp (wsrep_sst_receive_address, WSREP_SST_ADDRESS_AUTO))
   {
-    addr_in_parser = new wsp::Address(wsrep_sst_receive_address);
-
-    if (!addr_in_parser->is_valid())
-    {
-      WSREP_ERROR("Could not parse wsrep_sst_receive_address : %s",
-                  wsrep_sst_receive_address);
-      unireg_abort(1);
-    }
+    addr_in= wsrep_sst_receive_address;
   }
-  //Attempt 2: wsrep_node_address
-  else if (wsrep_node_address && *wsrep_node_address)
-  {
-    addr_in_parser = new wsp::Address(wsrep_node_address);
 
-    if (addr_in_parser->is_valid())
-    {
-      // we must not inherit the port number from this address:
-      addr_in_parser->set_port(0);
-    }
-    else
+  //Attempt 2: wsrep_node_address
+  else if (wsrep_node_address && strlen(wsrep_node_address))
+  {
+    wsp::Address addr(wsrep_node_address);
+
+    if (!addr.is_valid())
     {
       WSREP_ERROR("Could not parse wsrep_node_address : %s",
                   wsrep_node_address);
       throw wsrep::runtime_error("Failed to prepare for SST. Unrecoverable");
     }
+    memcpy(ip_buf, addr.get_address(), addr.get_address_len());
+    addr_in= ip_buf;
   }
   // Attempt 3: Try to get the IP from the list of available interfaces.
   else
   {
-    ssize_t ret= wsrep_guess_ip(ip_buf, ip_max);
+    ssize_t ret= wsrep_guess_ip (ip_buf, ip_max);
 
     if (ret && ret < ip_max)
     {
-      addr_in_parser = new wsp::Address(ip_buf);
+      addr_in= ip_buf;
     }
     else
     {
@@ -1276,51 +1303,6 @@ std::string wsrep_sst_prepare()
     }
   }
 
-  assert(addr_in_parser);
-
-  size_t len= addr_in_parser->get_address_len();
-  bool is_ipv6= addr_in_parser->is_ipv6();
-  const char* address= addr_in_parser->get_address();
-
-  if (len > (is_ipv6 ? ip_max - 2 : ip_max))
-  {
-    WSREP_ERROR("Address to accept state transfer is too long: '%s'",
-                address);
-    unireg_abort(1);
-  }
-
-  if (is_ipv6)
-  {
-    /* wsrep_sst_*.sh scripts requite ipv6 addreses to be in square breackets */
-    ip_buf[0] = '[';
-    /* the length (len) already includes the null byte: */
-    memcpy(ip_buf + 1, address, len - 1);
-    ip_buf[len] = ']';
-    ip_buf[len + 1] = 0;
-    len += 2;
-  }
-  else
-  {
-    memcpy(ip_buf, address, len);
-  }
-
-  int port= addr_in_parser->get_port();
-  if (port)
-  {
-    size_t space= ip_max - len;
-    ip_buf[len - 1] = ':';
-    int ret= snprintf(ip_buf + len, ip_max - len, "%d", port);
-    if (ret <= 0 || (size_t) ret > space)
-    {
-      WSREP_ERROR("Address to accept state transfer is too long: '%s:%d'",
-                  address, port);
-      unireg_abort(1);
-    }
-  }
-
-  delete addr_in_parser;
-  addr_in = ip_buf;
-
   ssize_t addr_len= -ENOSYS;
   method = wsrep_sst_method;
   if (!strcmp(method, WSREP_SST_MYSQLDUMP))
@@ -1374,7 +1356,7 @@ std::string wsrep_sst_prepare()
 
   const char* method_ptr(ret.data());
   const char* addr_ptr(ret.data() + strlen(method_ptr) + 1);
-  WSREP_DEBUG("Prepared SST request: %s|%s", method_ptr, addr_ptr);
+  WSREP_INFO ("Prepared SST request: %s|%s", method_ptr, addr_ptr);
 
   if (addr_out != addr_in) /* malloc'ed */ free ((char*)addr_out);
 
@@ -1490,6 +1472,8 @@ static int sst_donate_mysqldump (const char*         addr,
                                         wsrep::seqno::undefined()));
   Wsrep_server_state::instance().sst_sent(sst_sent_gtid, ret);
 
+  wsrep_donor_monitor_end();
+
   return ret;
 }
 
@@ -1554,19 +1538,21 @@ static int run_sql_command(THD *thd, const char *query)
     return -1;
   }
 
-  mysql_parse(thd, thd->query(), thd->query_length(), &ps, FALSE, FALSE);
-
+  mysql_parse(thd, thd->query(), thd->query_length(), &ps);
   if (thd->is_error())
   {
     int const err= thd->get_stmt_da()->sql_errno();
-    WSREP_WARN ("Error executing '%s': %d (%s)",
-                query, err, thd->get_stmt_da()->message());
+    WSREP_WARN ("Error executing '%s': %d (%s)%s",
+                query, err, thd->get_stmt_da()->message(),
+                err == ER_UNKNOWN_SYSTEM_VARIABLE ?
+                ". Was mysqld built with --with-innodb-disallow-writes ?" : "");
     thd->clear_error();
     return -1;
   }
   return 0;
 }
 
+
 static int sst_flush_tables(THD* thd)
 {
   WSREP_INFO("Flushing tables for SST...");
@@ -1626,19 +1612,16 @@ static int sst_flush_tables(THD* thd)
   }
   else
   {
-    ha_disable_internal_writes(true);
-
     WSREP_INFO("Tables flushed.");
-
-    // Create a file with cluster state ID and wsrep_gtid_domain_id.
+    /*
+      Tables have been flushed. Create a file with cluster state ID and
+      wsrep_gtid_domain_id.
+    */
     char content[100];
     snprintf(content, sizeof(content), "%s:%lld %d\n", wsrep_cluster_state_uuid,
              (long long)wsrep_locked_seqno, wsrep_gtid_server.domain_id);
     err= sst_create_file(flush_success, content);
 
-    if (err)
-      WSREP_INFO("Creating file for flush_success failed %d",err);
-
     const char base_name[]= "tables_flushed";
     ssize_t const full_len= strlen(mysql_real_data_home) + strlen(base_name)+2;
     char *real_name= (char*) malloc(full_len);
@@ -1673,13 +1656,39 @@ static int sst_flush_tables(THD* thd)
     }
     free(real_name);
     free(tmp_name);
-    if (err)
-      ha_disable_internal_writes(false);
   }
 
   return err;
 }
 
+
+static void sst_disallow_writes (THD* thd, bool yes)
+{
+  char query_str[64]= { 0, };
+  ssize_t const query_max= sizeof(query_str) - 1;
+  CHARSET_INFO *current_charset;
+
+  current_charset= thd->variables.character_set_client;
+
+  if (!is_supported_parser_charset(current_charset))
+  {
+      /* Do not use non-supported parser character sets */
+      WSREP_WARN("Current client character set is non-supported parser character set: %s", current_charset->csname);
+      thd->variables.character_set_client= &my_charset_latin1;
+      WSREP_WARN("For SST temporally setting character set to : %s",
+                 my_charset_latin1.csname);
+  }
+
+  snprintf (query_str, query_max, "SET GLOBAL innodb_disallow_writes=%d",
+            yes ? 1 : 0);
+
+  if (run_sql_command(thd, query_str))
+  {
+    WSREP_ERROR("Failed to disallow InnoDB writes");
+  }
+  thd->variables.character_set_client= current_charset;
+}
+
 static void* sst_donor_thread (void* a)
 {
   sst_thread_arg* arg= (sst_thread_arg*)a;
@@ -1725,32 +1734,20 @@ wait_signal:
       if (!strcasecmp (out, magic_flush))
       {
         err= sst_flush_tables (thd.ptr);
-
         if (!err)
         {
-          locked= true;
+          sst_disallow_writes (thd.ptr, true);
           /*
             Lets also keep statements that modify binary logs (like RESET LOGS,
             RESET MASTER) from proceeding until the files have been transferred
             to the joiner node.
           */
           if (mysql_bin_log.is_open())
-            mysql_mutex_lock(mysql_bin_log.get_log_lock());
-
-          WSREP_INFO("Donor state reached");
-
-#ifdef ENABLED_DEBUG_SYNC
-          DBUG_EXECUTE_IF("sync.wsrep_donor_state",
           {
-            const char act[]=
-                             "now "
-                             "SIGNAL sync.wsrep_donor_state_reached "
-                             "WAIT_FOR signal.wsrep_donor_state";
-            assert(!debug_sync_set_action(thd.ptr,
-                                          STRING_WITH_LEN(act)));
-          };);
-#endif
+            mysql_mutex_lock(mysql_bin_log.get_log_lock());
+          }
 
+          locked= true;
           goto wait_signal;
         }
       }
@@ -1758,11 +1755,14 @@ wait_signal:
       {
         if (locked)
         {
-          locked= false;
-          ha_disable_internal_writes(false);
           if (mysql_bin_log.is_open())
+          {
+            mysql_mutex_assert_owner(mysql_bin_log.get_log_lock());
             mysql_mutex_unlock(mysql_bin_log.get_log_lock());
+          }
+          sst_disallow_writes (thd.ptr, false);
           thd.ptr->global_read_lock.unlock_global_read_lock(thd.ptr);
+          locked= false;
         }
         err=  0;
         goto wait_signal;
@@ -1793,12 +1793,12 @@ wait_signal:
 
   if (locked) // don't forget to unlock server before return
   {
-    ha_disable_internal_writes(false);
     if (mysql_bin_log.is_open())
     {
       mysql_mutex_assert_owner(mysql_bin_log.get_log_lock());
       mysql_mutex_unlock(mysql_bin_log.get_log_lock());
     }
+    sst_disallow_writes (thd.ptr, false);
     thd.ptr->global_read_lock.unlock_global_read_lock(thd.ptr);
   }
 
@@ -1845,8 +1845,6 @@ static int sst_donate_other (const char*        method,
   {
     WSREP_ERROR("sst_prepare_other(): generate_binlog_index_opt_val() failed %d",
                 ret);
-    if (binlog_opt_val) my_free(binlog_opt_val);
-    return ret;
   }
 
   make_wsrep_defaults_file();
@@ -1857,18 +1855,16 @@ static int sst_donate_other (const char*        method,
                  "wsrep_sst_%s "
                  WSREP_SST_OPT_ROLE " 'donor' "
                  WSREP_SST_OPT_ADDR " '%s' "
-                 WSREP_SST_OPT_LPORT " %u "
+                 WSREP_SST_OPT_LPORT " '%u' "
                  WSREP_SST_OPT_SOCKET " '%s' "
-                 WSREP_SST_OPT_PROGRESS " %d "
                  WSREP_SST_OPT_DATA " '%s' "
                  "%s"
                  WSREP_SST_OPT_GTID " '%s:%lld' "
-                 WSREP_SST_OPT_GTID_DOMAIN_ID " %d"
+                 WSREP_SST_OPT_GTID_DOMAIN_ID " '%d'"
                  "%s"
                  "%s"
                  "%s",
                  method, addr, mysqld_port, mysqld_unix_port,
-                 0,
                  mysql_real_data_home,
                  wsrep_defaults_file,
                  uuid_oss.str().c_str(), gtid.seqno().get(), wsrep_gtid_server.domain_id,
@@ -1942,6 +1938,11 @@ int wsrep_sst_donate(const std::string& msg,
                      const wsrep::gtid& current_gtid,
                      const bool         bypass)
 {
+  /* This will be reset when sync callback is called.
+   * Should we set wsrep_ready to FALSE here too? */
+
+  wsrep_config_state->set(wsrep::server_state::s_donor);
+
   const char* method= msg.data();
   size_t method_len= strlen (method);
 
