@@ -83,7 +83,7 @@ BDOC::BDOC(PGLOBAL G) : BJSON(G, NULL)
 PBVAL BDOC::ParseJson(PGLOBAL g, char* js, size_t lng)
 {
   size_t i;
-  bool  b = false, ptyp = (bool *)pty;
+  bool  b = false;
   PBVAL bvp = NULL;
 
   s = js;
@@ -145,7 +145,7 @@ PBVAL BDOC::ParseJson(PGLOBAL g, char* js, size_t lng)
           b = false;
           break;
         } // endif b
-
+        /* fall through */
       default:
         if (bvp->Type != TYPE_UNKNOWN) {
           bvp->To_Val = ParseAsArray(i);
@@ -683,7 +683,7 @@ bool BDOC::SerializeArray(OFFSET arp, bool b)
   } else if (jp->WriteChr('['))
     return true;
 
-  for (vp; vp; vp = MVP(vp->Next)) {
+  for (; vp; vp = MVP(vp->Next)) {
     if (first)
       first = false;
     else if ((!b || jp->Prty()) && jp->WriteChr(','))
@@ -718,7 +718,7 @@ bool BDOC::SerializeObject(OFFSET obp)
   if (jp->WriteChr('{'))
     return true;
 
-  for (prp; prp; prp = GetNext(prp)) {
+  for (; prp; prp = GetNext(prp)) {
     if (first)
       first = false;
     else if (jp->WriteChr(','))
@@ -1138,9 +1138,6 @@ PBVAL BJSON::GetArrayValue(PBVAL bap, int n)
   CheckType(bap, TYPE_JAR);
   int i = 0;
 
-  if (n < 0)
-    n += GetArraySize(bap);
-
   for (PBVAL bvp = GetArray(bap); bvp; bvp = GetNext(bvp), i++)
     if (i == n)
       return bvp;
@@ -1351,17 +1348,12 @@ PBVAL BJSON::NewVal(PVAL valp)
 /***********************************************************************/
 /* Sub-allocate and initialize a BVAL from another BVAL.               */
 /***********************************************************************/
-PBVAL BJSON::DupVal(PBVAL bvlp)
-{
-  if (bvlp) {
-    PBVAL bvp = NewVal();
-
-    *bvp = *bvlp;
-    bvp->Next = 0;
-    return bvp;
-  } else
-    return NULL;
+PBVAL BJSON::DupVal(PBVAL bvlp) {
+  PBVAL bvp = NewVal();
 
+  *bvp = *bvlp;
+  bvp->Next = 0;
+  return bvp;
 } // end of DupVal
 
 /***********************************************************************/
