@@ -1,6 +1,6 @@
 /************* tabbson C++ Program Source Code File (.CPP) *************/
-/* PROGRAM NAME: tabbson     Version 1.1                               */
-/*  (C) Copyright to the author Olivier BERTRAND          2020 - 2021  */
+/* PROGRAM NAME: tabbson     Version 1.0                               */
+/*  (C) Copyright to the author Olivier BERTRAND          2020         */
 /*  This program are the BSON class DB execution routines.             */
 /***********************************************************************/
 
@@ -158,9 +158,8 @@ BSONDISC::BSONDISC(PGLOBAL g, uint* lg)
   bp = NULL;
   row = NULL;
   sep = NULL;
-  strfy = NULL;
   i = n = bf = ncol = lvl = sz = limit = 0;
-  all = false;
+  all = strfy = false;
 } // end of BSONDISC constructor
 
 int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
@@ -174,7 +173,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
   sep = GetStringTableOption(g, topt, "Separator", ".");
   sz = GetIntegerTableOption(g, topt, "Jsize", 1024);
   limit = GetIntegerTableOption(g, topt, "Limit", 10);
-  strfy = GetStringTableOption(g, topt, "Stringify", NULL);
+  strfy = GetBooleanTableOption(g, topt, "Stringify", false);
 
   /*********************************************************************/
   /*  Open the input file.                                             */
@@ -187,9 +186,6 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 #endif   // ZIP_SUPPORT
   tdp->Fn = GetStringTableOption(g, topt, "Filename", NULL);
 
-  if (!tdp->Fn && topt->http)
-    tdp->Fn = GetStringTableOption(g, topt, "Subtype", NULL);
-
   if (!(tdp->Database = SetPath(g, db)))
     return 0;
 
@@ -203,8 +199,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
   if (!tdp->Fn && !tdp->Uri) {
     strcpy(g->Message, MSG(MISSING_FNAME));
     return 0;
-  } else
-    topt->subtype = NULL;
+  } // endif Fn
 
   if (tdp->Fn) {
     //  We used the file name relative to recorded datapath
@@ -433,7 +428,7 @@ bool BSONDISC::Find(PGLOBAL g, PBVAL jvp, PCSZ key, int j)
     jcol.Type = TYPE_UNKNOWN;
     jcol.Len = jcol.Scale = 0;
     jcol.Cbn = true;
-  } else  if (j < lvl && !(strfy && !stricmp(strfy, colname))) {
+  } else  if (j < lvl) {
     if (!fmt[bf])
       strcat(fmt, colname);
 
@@ -504,7 +499,7 @@ bool BSONDISC::Find(PGLOBAL g, PBVAL jvp, PCSZ key, int j)
     } // endswitch Type
 
   } else if (lvl >= 0) {
-    if (strfy && !stricmp(strfy, colname)) {
+    if (strfy) {
       if (!fmt[bf])
         strcat(fmt, colname);
 
@@ -736,6 +731,7 @@ void BCUTIL::SetJsonValue(PGLOBAL g, PVAL vp, PBVAL jvp)
     case TYPE_FLOAT:
       switch (vp->GetType()) {
         case TYPE_STRING:
+        case TYPE_DATE:
         case TYPE_DECIM:
           vp->SetValue_psz(GetString(jvp));
           break;
@@ -753,16 +749,6 @@ void BCUTIL::SetJsonValue(PGLOBAL g, PVAL vp, PBVAL jvp)
           if (jvp->Type == TYPE_DBL || jvp->Type == TYPE_FLOAT)
             vp->SetPrec(jvp->Nd);
 
-          break;
-        case TYPE_DATE:
-          if (jvp->Type == TYPE_STRG) {
-            if (!((DTVAL*)vp)->IsFormatted())
-              ((DTVAL*)vp)->SetFormat(g, "YYYY-MM-DDThh:mm:ssZ", 20, 0);
-
-            vp->SetValue_psz(GetString(jvp));
-          } else
-            vp->SetValue(GetInteger(jvp));
-
           break;
         default:
           sprintf(G->Message, "Unsupported column type %d", vp->GetType());
@@ -838,7 +824,7 @@ PBVAL BCUTIL::MakeBson(PGLOBAL g, PBVAL jsp, int n)
 /***********************************************************************/
 PBVAL BCUTIL::GetRowValue(PGLOBAL g, PBVAL row, int i)
 {
-  int    nod = Cp->Nod, n = nod - 1;
+  int    nod = Cp->Nod;
   JNODE *nodes = Cp->Nodes;
   PBVAL  arp;
   PBVAL  bvp = NULL;
@@ -895,7 +881,7 @@ PBVAL BCUTIL::GetRowValue(PGLOBAL g, PBVAL row, int i)
   } // endfor i
 
   return bvp;
-} // end of GetRowValue
+} // end of GetColumnValue
 
 /***********************************************************************/
 /*  GetColumnValue:                                                    */
@@ -1970,6 +1956,7 @@ PSZ BSONCOL::GetJpath(PGLOBAL g, bool proj)
       return NULL;
 
     for (p1 = p2 = mgopath; *p1; p1++)
+    {
       if (i) {                 // Inside []
         if (isdigit(*p1)) {
           if (!proj)
@@ -2007,14 +1994,14 @@ PSZ BSONCOL::GetJpath(PGLOBAL g, bool proj)
           p2--;              // Suppress last :*
           break;
         } // endif p2
-
+        /* fall through */
       default:
         *p2++ = *p1;
         break;
       } // endswitch p1;
-
-      *p2 = 0;
-      return mgopath;
+    }
+    *p2 = 0;
+    return mgopath;
   } else
     return NULL;
 
