@@ -2,7 +2,7 @@
 
 Copyright (c) 1996, 2018, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2012, Facebook Inc.
-Copyright (c) 2013, 2022, MariaDB Corporation.
+Copyright (c) 2013, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -35,7 +35,6 @@ Created 1/8/1996 Heikki Tuuri
 #include "dict0dict.h"
 #include "fts0priv.h"
 #include "lock0lock.h"
-#include "sync0sync.h"
 #include "row0row.h"
 #include "sql_string.h"
 #include <iostream>
@@ -188,7 +187,7 @@ dict_table_t *dict_mem_table_create(const char *name, fil_space_t *space,
 	}
 
 	table->autoinc_lock = static_cast<ib_lock_t*>(
-		mem_heap_alloc(heap, lock_get_size()));
+		mem_heap_alloc(heap, sizeof *table->autoinc_lock));
 
 	/* If the table has an FTS index or we are in the process
 	of building one, create the table->fts */
@@ -791,8 +790,8 @@ dict_mem_index_create(
 		index->rtr_track = new
 			(mem_heap_alloc(heap, sizeof *index->rtr_track))
 			rtr_info_track_t();
-		mutex_create(LATCH_ID_RTR_ACTIVE_MUTEX,
-			     &index->rtr_track->rtr_active_mutex);
+		mysql_mutex_init(rtr_active_mutex_key,
+				 &index->rtr_track->rtr_active_mutex, nullptr);
 	}
 
 	return(index);
@@ -1101,7 +1100,7 @@ dict_mem_index_free(
 			rtr_info->index = NULL;
 		}
 
-		mutex_destroy(&index->rtr_track->rtr_active_mutex);
+		mysql_mutex_destroy(&index->rtr_track->rtr_active_mutex);
 		index->rtr_track->~rtr_info_track_t();
 	}
 
@@ -1210,14 +1209,11 @@ bool dict_foreign_t::affects_fulltext() const
   return false;
 }
 
-/** Reconstruct the clustered index fields.
-@return whether metadata is incorrect */
-inline bool dict_index_t::reconstruct_fields()
+/** Reconstruct the clustered index fields. */
+inline void dict_index_t::reconstruct_fields()
 {
 	DBUG_ASSERT(is_primary());
 
-	const auto old_n_fields = n_fields;
-
 	n_fields = (n_fields + table->instant->n_dropped)
 		& dict_index_t::MAX_N_FIELDS;
 	n_def = (n_def + table->instant->n_dropped)
@@ -1245,17 +1241,13 @@ inline bool dict_index_t::reconstruct_fields()
 		} else {
 			DBUG_ASSERT(!c.is_not_null());
 			const auto old = std::find_if(
-				fields + n_first, fields + old_n_fields,
+				fields + n_first, fields + n_fields,
 				[c](const dict_field_t& o)
 				{ return o.col->ind == c.ind(); });
-
-			if (old >= fields + old_n_fields
-			    || old->prefix_len
-			    || old->col != &table->cols[c.ind()]) {
-				return true;
-			}
-
 			ut_ad(old >= &fields[n_first]);
+			ut_ad(old < &fields[n_fields]);
+			DBUG_ASSERT(!old->prefix_len);
+			DBUG_ASSERT(old->col == &table->cols[c.ind()]);
 			f = *old;
 		}
 
@@ -1268,8 +1260,6 @@ inline bool dict_index_t::reconstruct_fields()
 
 	fields = tfields;
 	n_core_null_bytes = static_cast<byte>(UT_BITS_IN_BYTES(n_core_null));
-
-	return false;
 }
 
 /** Reconstruct dropped or reordered columns.
@@ -1334,7 +1324,8 @@ bool dict_table_t::deserialise_columns(const byte* metadata, ulint len)
 	}
 	DBUG_ASSERT(col == &dropped_cols[n_dropped_cols]);
 
-	return UT_LIST_GET_FIRST(indexes)->reconstruct_fields();
+	UT_LIST_GET_FIRST(indexes)->reconstruct_fields();
+	return false;
 }
 
 /** Check if record in clustered index is historical row.
