@@ -32,9 +32,9 @@ Created 11/5/1995 Heikki Tuuri
 
 #include "fil0fil.h"
 #include "mtr0types.h"
-#include "buf0types.h"
 #include "span.h"
 #include "assume_aligned.h"
+#include "buf0types.h"
 #ifndef UNIV_INNOCHECKSUM
 #include "hash0hash.h"
 #include "ut0byte.h"
@@ -68,12 +68,6 @@ struct fil_addr_t;
 #define BUF_EVICT_IF_IN_POOL	20	/*!< evict a clean block if found */
 /* @} */
 
-/** If LRU list of a buf_pool is less than this size then LRU eviction
-should not happen. This is because when we do LRU flushing we also put
-the blocks on free list. If LRU list is very small then we can end up
-in thrashing. */
-#define BUF_LRU_MIN_LEN		256
-
 # ifdef UNIV_DEBUG
 extern my_bool	buf_disable_resize_buffer_pool_debug; /*!< if TRUE, resizing
 					buffer pool is not allowed. */
@@ -216,7 +210,7 @@ NOTE! The following macros should be used instead of buf_page_get_gen,
 to improve debugging. Only values RW_S_LATCH and RW_X_LATCH are allowed
 in LA! */
 #define buf_page_get(ID, SIZE, LA, MTR)					\
-	buf_page_get_gen(ID, SIZE, LA, NULL, BUF_GET, __FILE__, __LINE__, MTR)
+	buf_page_get_gen(ID, SIZE, LA, NULL, BUF_GET, MTR)
 
 /**************************************************************//**
 Use these macros to bufferfix a page with no latching. Remember not to
@@ -225,8 +219,7 @@ the contents of the page! We have separated this case, because it is
 error-prone programming not to set a latch, and it should be used
 with care. */
 #define buf_page_get_with_no_latch(ID, SIZE, MTR)	\
-	buf_page_get_gen(ID, SIZE, RW_NO_LATCH, NULL, BUF_GET_NO_LATCH, \
-			 __FILE__, __LINE__, MTR)
+	buf_page_get_gen(ID, SIZE, RW_NO_LATCH, NULL, BUF_GET_NO_LATCH, MTR)
 /********************************************************************//**
 This is the general function used to get optimistic access to a database
 page.
@@ -237,33 +230,15 @@ buf_page_optimistic_get(
 	ulint		rw_latch,/*!< in: RW_S_LATCH, RW_X_LATCH */
 	buf_block_t*	block,	/*!< in: guessed block */
 	ib_uint64_t	modify_clock,/*!< in: modify clock value */
-	const char*	file,	/*!< in: file name */
-	unsigned	line,	/*!< in: line where called */
 	mtr_t*		mtr);	/*!< in: mini-transaction */
 
-/** Given a tablespace id and page number tries to get that page. If the
-page is not in the buffer pool it is not loaded and NULL is returned.
-Suitable for using when holding the lock_sys_t::mutex.
-@param[in]	page_id	page id
-@param[in]	file	file name
-@param[in]	line	line where called
-@param[in]	mtr	mini-transaction
-@return pointer to a page or NULL */
-buf_block_t*
-buf_page_try_get_func(
-	const page_id_t		page_id,
-	const char*		file,
-	unsigned		line,
-	mtr_t*			mtr);
-
-/** Tries to get a page.
-If the page is not in the buffer pool it is not loaded. Suitable for using
-when holding the lock_sys_t::mutex.
+/** Try to S-latch a page.
+Suitable for using when holding the lock_sys latches (as it avoids deadlock).
 @param[in]	page_id	page identifier
-@param[in]	mtr	mini-transaction
-@return the page if in buffer pool, NULL if not */
-#define buf_page_try_get(page_id, mtr)	\
-	buf_page_try_get_func((page_id), __FILE__, __LINE__, mtr);
+@param[in,out]	mtr	mini-transaction
+@return the block
+@retval nullptr if an S-latch cannot be granted immediately */
+buf_block_t *buf_page_try_get(const page_id_t page_id, mtr_t *mtr);
 
 /** Get read access to a compressed page (usually of type
 FIL_PAGE_TYPE_ZBLOB or FIL_PAGE_TYPE_ZBLOB2).
@@ -284,8 +259,6 @@ buf_page_t* buf_page_get_zip(const page_id_t page_id, ulint zip_size);
 @param[in]	guess			guessed block or NULL
 @param[in]	mode			BUF_GET, BUF_GET_IF_IN_POOL,
 BUF_PEEK_IF_IN_POOL, BUF_GET_NO_LATCH, or BUF_GET_IF_IN_POOL_OR_WATCH
-@param[in]	file			file name
-@param[in]	line			line where called
 @param[in]	mtr			mini-transaction
 @param[out]	err			DB_SUCCESS or error code
 @param[in]	allow_ibuf_merge	Allow change buffer merge while
@@ -298,8 +271,6 @@ buf_page_get_gen(
 	ulint			rw_latch,
 	buf_block_t*		guess,
 	ulint			mode,
-	const char*		file,
-	unsigned		line,
 	mtr_t*			mtr,
 	dberr_t*		err = NULL,
 	bool			allow_ibuf_merge = false);
@@ -311,8 +282,6 @@ buf_page_get_gen(
 @param[in]	guess			guessed block or NULL
 @param[in]	mode			BUF_GET, BUF_GET_IF_IN_POOL,
 BUF_PEEK_IF_IN_POOL, BUF_GET_NO_LATCH, or BUF_GET_IF_IN_POOL_OR_WATCH
-@param[in]	file			file name
-@param[in]	line			line where called
 @param[in]	mtr			mini-transaction
 @param[out]	err			DB_SUCCESS or error code
 @param[in]	allow_ibuf_merge	Allow change buffer merge to happen
@@ -327,8 +296,6 @@ buf_page_get_low(
 	ulint			rw_latch,
 	buf_block_t*		guess,
 	ulint			mode,
-	const char*		file,
-	unsigned		line,
 	mtr_t*			mtr,
 	dberr_t*		err,
 	bool			allow_ibuf_merge);
@@ -365,15 +332,11 @@ buf_page_release_latch(
 					RW_NO_LATCH */
 /** Move a block to the start of the LRU list. */
 void buf_page_make_young(buf_page_t *bpage);
-/** Mark the page status as FREED for the given tablespace id and
-page number. If the page is not in buffer pool then ignore it.
+/** Mark the page status as FREED for the given tablespace and page number.
 @param[in,out]	space	tablespace
 @param[in]	page	page number
-@param[in,out]	mtr	mini-transaction
-@param[in]	file	file name
-@param[in]	line	line where called */
-void buf_page_free(fil_space_t *space, uint32_t page, mtr_t *mtr,
-                   const char *file, unsigned line);
+@param[in,out]	mtr	mini-transaction */
+void buf_page_free(fil_space_t *space, uint32_t page, mtr_t *mtr);
 
 /********************************************************************//**
 Reads the freed_page_clock of a buffer block.
@@ -438,30 +401,11 @@ buf_block_get_modify_clock(
 	buf_block_t*	block);	/*!< in: block */
 /*******************************************************************//**
 Increments the bufferfix count. */
-UNIV_INLINE
-void
-buf_block_buf_fix_inc_func(
-/*=======================*/
-# ifdef UNIV_DEBUG
-	const char*	file,	/*!< in: file name */
-	unsigned	line,	/*!< in: line */
-# endif /* UNIV_DEBUG */
-	buf_block_t*	block)	/*!< in/out: block to bufferfix */
-	MY_ATTRIBUTE((nonnull));
+# define buf_block_buf_fix_inc(block) (block)->fix()
 
-# ifdef UNIV_DEBUG
-/** Increments the bufferfix count.
-@param[in,out]	b	block to bufferfix
-@param[in]	f	file name where requested
-@param[in]	l	line number where requested */
-#  define buf_block_buf_fix_inc(b,f,l) buf_block_buf_fix_inc_func(f,l,b)
-# else /* UNIV_DEBUG */
-/** Increments the bufferfix count.
-@param[in,out]	b	block to bufferfix
-@param[in]	f	file name where requested
-@param[in]	l	line number where requested */
-#  define buf_block_buf_fix_inc(b,f,l) buf_block_buf_fix_inc_func(b)
-# endif /* UNIV_DEBUG */
+/*******************************************************************//**
+Decrements the bufferfix count. */
+# define buf_block_buf_fix_dec(block) (block)->unfix()
 #endif /* !UNIV_INNOCHECKSUM */
 
 /** Check if a buffer is all zeroes.
@@ -469,42 +413,6 @@ buf_block_buf_fix_inc_func(
 @return whether the buffer is all zeroes */
 bool buf_is_zeroes(st_::span<const byte> buf);
 
-/** Checks if the page is in crc32 checksum format.
-@param[in]	read_buf		database page
-@param[in]	checksum_field1		new checksum field
-@param[in]	checksum_field2		old checksum field
-@return true if the page is in crc32 checksum format. */
-bool
-buf_page_is_checksum_valid_crc32(
-	const byte*			read_buf,
-	ulint				checksum_field1,
-	ulint				checksum_field2)
-	MY_ATTRIBUTE((nonnull(1), warn_unused_result));
-
-/** Checks if the page is in innodb checksum format.
-@param[in]	read_buf	database page
-@param[in]	checksum_field1	new checksum field
-@param[in]	checksum_field2	old checksum field
-@return true if the page is in innodb checksum format. */
-bool
-buf_page_is_checksum_valid_innodb(
-	const byte*			read_buf,
-	ulint				checksum_field1,
-	ulint				checksum_field2)
-	MY_ATTRIBUTE((nonnull(1), warn_unused_result));
-
-/** Checks if the page is in none checksum format.
-@param[in]	read_buf	database page
-@param[in]	checksum_field1	new checksum field
-@param[in]	checksum_field2	old checksum field
-@return true if the page is in none checksum format. */
-bool
-buf_page_is_checksum_valid_none(
-	const byte*			read_buf,
-	ulint				checksum_field1,
-	ulint				checksum_field2)
-	MY_ATTRIBUTE((nonnull(1), warn_unused_result));
-
 /** Check if a page is corrupt.
 @param[in]	check_lsn	whether the LSN should be checked
 @param[in]	read_buf	database page
@@ -636,22 +544,6 @@ void buf_pool_invalidate();
 --------------------------- LOWER LEVEL ROUTINES -------------------------
 =========================================================================*/
 
-#ifdef UNIV_DEBUG
-/*********************************************************************//**
-Adds latch level info for the rw-lock protecting the buffer frame. This
-should be called in the debug version after a successful latching of a
-page if we know the latching order level of the acquired latch. */
-UNIV_INLINE
-void
-buf_block_dbg_add_level(
-/*====================*/
-	buf_block_t*	block,	/*!< in: buffer page
-				where we have acquired latch */
-	latch_level_t	level);	/*!< in: latching order level */
-#else /* UNIV_DEBUG */
-# define buf_block_dbg_add_level(block, level) /* nothing */
-#endif /* UNIV_DEBUG */
-
 #ifdef UNIV_DEBUG
 /*********************************************************************//**
 Gets a pointer to the memory frame of a block.
@@ -694,6 +586,7 @@ dberr_t buf_page_read_complete(buf_page_t *bpage, const fil_node_t &node);
 if needed.
 @param[in]	size	size in bytes
 @return	aligned size */
+UNIV_INLINE
 ulint
 buf_pool_size_align(
 	ulint	size);
@@ -768,16 +661,6 @@ private:
   /** Count of how manyfold this block is currently bufferfixed. */
   Atomic_counter<uint32_t> buf_fix_count_;
 
-  /** log sequence number of the START of the log entry written of the
-  oldest modification to this block which has not yet been written
-  to the data file;
-
-  0 if no modifications are pending;
-  1 if no modifications are pending, but the block is in buf_pool.flush_list;
-  2 if modifications are pending, but the block is not in buf_pool.flush_list
-  (because id().space() is the temporary tablespace). */
-  Atomic_relaxed<lsn_t> oldest_modification_;
-
   /** type of pending I/O operation; protected by buf_pool.mutex
   if in_LRU_list */
   Atomic_relaxed<buf_io_fix> io_fix_;
@@ -827,6 +710,12 @@ public:
   or if state() is BUF_BLOCK_MEMORY or BUF_BLOCK_REMOVE_HASH. */
   UT_LIST_NODE_T(buf_page_t) list;
 
+private:
+  /** log sequence number of the START of the log entry written of the
+  oldest modification to this block which has not yet been written
+  to the data file; 0 if no modifications are pending. */
+  Atomic_counter<lsn_t> oldest_modification_;
+public:
 	/** @name LRU replacement algorithm fields.
 	Protected by buf_pool.mutex. */
 	/* @{ */
@@ -941,26 +830,12 @@ public:
   inline void set_io_fix(buf_io_fix io_fix);
   inline void set_corrupt_id();
 
-  /** @return the log sequence number of the oldest pending modification
-  @retval 0 if the block is being removed from (or not in) buf_pool.flush_list
-  @retval 1 if the block is in buf_pool.flush_list but not modified
-  @retval 2 if the block belongs to the temporary tablespace and
-  has unwritten changes */
+  /** @return the oldest modification */
   lsn_t oldest_modification() const { return oldest_modification_; }
-  /** @return the log sequence number of the oldest pending modification,
-  @retval 0 if the block is definitely not in buf_pool.flush_list
-  @retval 1 if the block is in buf_pool.flush_list but not modified
-  @retval 2 if the block belongs to the temporary tablespace and
-  has unwritten changes */
-  lsn_t oldest_modification_acquire() const
-  { return oldest_modification_.load(std::memory_order_acquire); }
   /** Set oldest_modification when adding to buf_pool.flush_list */
   inline void set_oldest_modification(lsn_t lsn);
-  /** Clear oldest_modification after removing from buf_pool.flush_list */
+  /** Clear oldest_modification when removing from buf_pool.flush_list */
   inline void clear_oldest_modification();
-  /** Note that a block is no longer dirty, while not removing
-  it from buf_pool.flush_list */
-  inline void clear_oldest_modification(bool temporary);
 
   /** Notify that a page in a temporary tablespace has been modified. */
   void set_temp_modified()
@@ -968,15 +843,14 @@ public:
     ut_ad(fsp_is_system_temporary(id().space()));
     ut_ad(state() == BUF_BLOCK_FILE_PAGE);
     ut_ad(!oldest_modification());
-    oldest_modification_= 2;
+    oldest_modification_= 1;
   }
 
   /** Prepare to release a file page to buf_pool.free. */
   void free_file_page()
   {
     ut_ad(state() == BUF_BLOCK_REMOVE_HASH);
-    /* buf_LRU_block_free_non_file_page() asserts !oldest_modification() */
-    ut_d(oldest_modification_= 0;)
+    ut_d(oldest_modification_= 0); /* for buf_LRU_block_free_non_file_page() */
     set_corrupt_id();
     ut_d(set_state(BUF_BLOCK_MEMORY));
   }
@@ -1068,8 +942,8 @@ struct buf_block_t{
 					is of size srv_page_size, and
 					aligned to an address divisible by
 					srv_page_size */
-	rw_lock_t	lock;		/*!< read-write lock of the buffer
-					frame */
+  /** read-write lock covering frame */
+  block_lock lock;
 #ifdef UNIV_DEBUG
   /** whether page.list is in buf_pool.withdraw
   ((state() == BUF_BLOCK_NOT_USED)) and the buffer pool is being shrunk;
@@ -1193,22 +1067,12 @@ struct buf_block_t{
 # define assert_block_ahi_empty_on_init(block) /* nothing */
 # define assert_block_ahi_valid(block) /* nothing */
 #endif /* BTR_CUR_HASH_ADAPT */
-# ifdef UNIV_DEBUG
-	/** @name Debug fields */
-	/* @{ */
-	rw_lock_t*	debug_latch;	/*!< in the debug version, each thread
-					which bufferfixes the block acquires
-					an s-latch here; so we can use the
-					debug utilities in sync0rw */
-	/* @} */
-# endif
   void fix() { page.fix(); }
   uint32_t unfix()
   {
     ut_ad(page.buf_fix_count() || page.io_fix() != BUF_IO_NONE ||
           page.state() == BUF_BLOCK_ZIP_PAGE ||
-          !rw_lock_own_flagged(&lock, RW_LOCK_FLAG_X | RW_LOCK_FLAG_S |
-                               RW_LOCK_FLAG_SX));
+          !lock.have_any());
     return page.unfix();
   }
 
@@ -1356,7 +1220,11 @@ struct buf_buddy_free_t {
 
 /** @brief The buffer pool statistics structure. */
 struct buf_pool_stat_t{
-	ulint	n_page_gets;	/*!< number of page gets performed;
+	/** Initialize the counters */
+	void init() { memset((void*) this, 0, sizeof *this); }
+
+	ib_counter_t<ulint, ib_counter_element_t>	n_page_gets;
+				/*!< number of page gets performed;
 				also successful searches through
 				the adaptive hash index are
 				counted as page gets; this field
@@ -1513,7 +1381,7 @@ public:
   {
     ut_ad(curr_size < old_size);
 #ifdef SAFE_MUTEX
-    if (resizing.load(std::memory_order_relaxed))
+    if (resize_in_progress())
       mysql_mutex_assert_owner(&mutex);
 #endif /* SAFE_MUTEX */
 
@@ -1533,7 +1401,7 @@ public:
   {
     ut_ad(curr_size < old_size);
 #ifdef SAFE_MUTEX
-    if (resizing.load(std::memory_order_relaxed))
+    if (resize_in_progress())
       mysql_mutex_assert_owner(&mutex);
 #endif /* SAFE_MUTEX */
 
@@ -1583,27 +1451,23 @@ public:
   inline buf_block_t *block_from_ahi(const byte *ptr) const;
 #endif /* BTR_CUR_HASH_ADAPT */
 
-  bool is_block_lock(const rw_lock_t *l) const
-  { return is_block_field(static_cast<const void*>(l)); }
+  /** @return the block that was made dirty the longest time ago */
+  const buf_page_t *get_oldest_modified() const
+  {
+    mysql_mutex_assert_owner(&flush_list_mutex);
+    const buf_page_t *bpage= UT_LIST_GET_LAST(flush_list);
+    ut_ad(!bpage || !fsp_is_system_temporary(bpage->id().space()));
+    ut_ad(!bpage || bpage->oldest_modification());
+    return bpage;
+  }
 
   /**
   @return the smallest oldest_modification lsn for any page
   @retval empty_lsn if all modified persistent pages have been flushed */
-  lsn_t get_oldest_modification(lsn_t empty_lsn)
+  lsn_t get_oldest_modification(lsn_t empty_lsn) const
   {
-    mysql_mutex_assert_owner(&flush_list_mutex);
-    while (buf_page_t *bpage= UT_LIST_GET_LAST(flush_list))
-    {
-      ut_ad(!fsp_is_system_temporary(bpage->id().space()));
-      lsn_t lsn= bpage->oldest_modification();
-      if (lsn != 1)
-      {
-        ut_ad(lsn > 2);
-        return lsn;
-      }
-      delete_from_flush_list(bpage);
-    }
-    return empty_lsn;
+    const buf_page_t *bpage= get_oldest_modified();
+    return bpage ? bpage->oldest_modification() : empty_lsn;
   }
 
   /** Determine if a buffer block was created by chunk_t::create().
@@ -1744,7 +1608,33 @@ public:
   /** Stop watching whether a page has been read in.
   watch_set(id) must have returned nullptr before.
   @param id   page identifier */
-  void watch_unset(const page_id_t id);
+  void watch_unset(const page_id_t id)
+  {
+    const ulint fold= id.fold();
+    page_hash_latch *hash_lock= page_hash.lock<true>(fold);
+    /* The page must exist because watch_set() increments buf_fix_count. */
+    buf_page_t *watch= page_hash_get_low(id, fold);
+    if (watch->unfix() == 0 && watch_is_sentinel(*watch))
+    {
+      /* The following is based on watch_remove(). */
+      ut_ad(watch->in_page_hash);
+      ut_d(watch->in_page_hash= false);
+      HASH_DELETE(buf_page_t, hash, &page_hash, fold, watch);
+      hash_lock->write_unlock();
+      // Now that the watch is detached from page_hash, release it to watch[].
+      mysql_mutex_lock(&mutex);
+      /* It is possible that watch_remove() already removed the watch. */
+      if (watch->id_ == id)
+      {
+        ut_ad(!watch->buf_fix_count());
+        ut_ad(watch->state() == BUF_BLOCK_ZIP_PAGE);
+        watch->set_state(BUF_BLOCK_NOT_USED);
+      }
+      mysql_mutex_unlock(&mutex);
+    }
+    else
+      hash_lock->write_unlock();
+  }
 
   /** Remove the sentinel block for the watch before replacing it with a
   real block. watch_unset() or watch_occurred() will notice
@@ -1791,18 +1681,15 @@ public:
 
   /** Buffer pool mutex */
   MY_ALIGNED(CPU_LEVEL1_DCACHE_LINESIZE) mysql_mutex_t mutex;
-  /** Number of pending LRU flush; protected by mutex. */
-  ulint n_flush_LRU_;
+  /** Number of pending LRU flush. */
+  Atomic_counter<ulint> n_flush_LRU;
   /** broadcast when n_flush_LRU reaches 0; protected by mutex */
   pthread_cond_t done_flush_LRU;
-  /** Number of pending flush_list flush; protected by mutex */
-  ulint n_flush_list_;
+  /** Number of pending flush_list flush. */
+  Atomic_counter<ulint> n_flush_list;
   /** broadcast when n_flush_list reaches 0; protected by mutex */
   pthread_cond_t done_flush_list;
 
-  TPOOL_SUPPRESS_TSAN ulint n_flush_LRU() const { return n_flush_LRU_; }
-  TPOOL_SUPPRESS_TSAN ulint n_flush_list() const { return n_flush_list_; }
-
 	/** @name General fields */
 	/* @{ */
 	ulint		curr_pool_size;	/*!< Current pool size in bytes */
@@ -1874,14 +1761,22 @@ public:
     page_hash_latch *lock_get(ulint fold) const
     { return lock_get(fold, n_cells); }
 
-    /** Acquire an array latch.
+    /** Acquire an array latch, tolerating concurrent buf_pool_t::resize()
     @tparam exclusive  whether the latch is to be acquired exclusively
     @param fold    hash bucket key */
     template<bool exclusive> page_hash_latch *lock(ulint fold)
     {
-      page_hash_latch *latch= lock_get(fold, n_cells);
-      latch->acquire<exclusive>();
-      return latch;
+      for (;;)
+      {
+        auto n= n_cells;
+        page_hash_latch *latch= lock_get(fold, n);
+        latch->acquire<exclusive>();
+        /* Our latch prevents n_cells from changing. */
+        if (UNIV_LIKELY(n == n_cells))
+          return latch;
+        /* Retry, because buf_pool_t::resize_hash() affected us. */
+        latch->release<exclusive>();
+      }
     }
 
     /** Exclusively aqcuire all latches */
@@ -1891,6 +1786,19 @@ public:
     inline void write_unlock_all();
   };
 
+private:
+  /** Former page_hash that has been deleted during resize();
+  singly-linked list via freed_page_hash->array[1] */
+  page_hash_table *freed_page_hash;
+
+  /** Lock all page_hash, also freed_page_hash. */
+  inline void write_lock_all_page_hash();
+  /** Release all page_hash, also freed_page_hash. */
+  inline void write_unlock_all_page_hash();
+  /** Resize page_hash and zip_hash. */
+  inline void resize_hash();
+
+public:
   /** Hash table of file pages (buf_page_t::in_file() holds),
   indexed by page_id_t. Protected by both mutex and page_hash.lock_get(). */
   page_hash_table page_hash;
@@ -1940,7 +1848,7 @@ public:
     return page_cleaner_is_idle;
   }
   /** Wake up the page cleaner if needed */
-  void page_cleaner_wakeup();
+  inline void page_cleaner_wakeup();
 
   /** Register whether an explicit wakeup of the page cleaner is needed */
   void page_cleaner_set_idle(bool deep_sleep)
@@ -1956,8 +1864,8 @@ public:
     last_activity_count= activity_count;
   }
 
-  // n_flush_LRU() + n_flush_list()
-  // is approximately COUNT(io_fix()==BUF_IO_WRITE) in flush_list
+  // n_flush_LRU + n_flush_list is approximately COUNT(io_fix()==BUF_IO_WRITE)
+  // in flush_list
 
 	unsigned	freed_page_clock;/*!< a sequence number used
 					to count the number of buffer
@@ -1986,8 +1894,6 @@ public:
 	UT_LIST_BASE_NODE_T(buf_page_t) free;
 					/*!< base node of the free
 					block list */
-  /** signaled each time when the free list grows; protected by mutex */
-  pthread_cond_t done_free;
 
 	UT_LIST_BASE_NODE_T(buf_page_t) withdraw;
 					/*!< base node of the withdraw
@@ -2044,35 +1950,13 @@ public:
   /** @return whether any I/O is pending */
   bool any_io_pending() const
   {
-    return n_pend_reads || n_flush_LRU() || n_flush_list();
+    return n_pend_reads || n_flush_LRU || n_flush_list;
   }
   /** @return total amount of pending I/O */
   ulint io_pending() const
   {
-    return n_pend_reads + n_flush_LRU() + n_flush_list();
+    return n_pend_reads + n_flush_LRU + n_flush_list;
   }
-
-private:
-  /** Remove a block from the flush list. */
-  inline void delete_from_flush_list_low(buf_page_t *bpage);
-  /** Remove a block from flush_list.
-  @param bpage   buffer pool page
-  @param clear   whether to invoke buf_page_t::clear_oldest_modification() */
-  void delete_from_flush_list(buf_page_t *bpage, bool clear);
-public:
-  /** Remove a block from flush_list.
-  @param bpage   buffer pool page */
-  void delete_from_flush_list(buf_page_t *bpage)
-  { delete_from_flush_list(bpage, true); }
-
-  /** Insert a modified block into the flush list.
-  @param block    modified block
-  @param lsn      start LSN of the mini-transaction that modified the block */
-  void insert_into_flush_list(buf_block_t *block, lsn_t lsn);
-
-  /** Free a page whose underlying file page has been freed. */
-  inline void release_freed_page(buf_page_t *bpage);
-
 private:
   /** Temporary memory for page_compressed and encrypted I/O */
   struct io_buf_t
@@ -2185,10 +2069,9 @@ inline void buf_page_t::set_corrupt_id()
   switch (oldest_modification()) {
   case 0:
     break;
-  case 2:
+  case 1:
     ut_ad(fsp_is_system_temporary(id().space()));
-    /* buf_LRU_block_free_non_file_page() asserts !oldest_modification() */
-    ut_d(oldest_modification_= 0;)
+    ut_d(oldest_modification_= 0); /* for buf_LRU_block_free_non_file_page() */
     break;
   default:
     ut_ad("block is dirty" == 0);
@@ -2212,12 +2095,11 @@ inline void buf_page_t::set_corrupt_id()
 inline void buf_page_t::set_oldest_modification(lsn_t lsn)
 {
   mysql_mutex_assert_owner(&buf_pool.flush_list_mutex);
-  ut_ad(oldest_modification() <= 1);
-  ut_ad(lsn > 2);
+  ut_ad(!oldest_modification());
   oldest_modification_= lsn;
 }
 
-/** Clear oldest_modification after removing from buf_pool.flush_list */
+/** Clear oldest_modification when removing from buf_pool.flush_list */
 inline void buf_page_t::clear_oldest_modification()
 {
   mysql_mutex_assert_owner(&buf_pool.flush_list_mutex);
@@ -2225,32 +2107,7 @@ inline void buf_page_t::clear_oldest_modification()
   ut_ad(state == BUF_BLOCK_FILE_PAGE || state == BUF_BLOCK_ZIP_PAGE ||
         state == BUF_BLOCK_REMOVE_HASH);
   ut_ad(oldest_modification());
-  ut_ad(!list.prev);
-  ut_ad(!list.next);
-  /* We must use release memory order to guarantee that callers of
-  oldest_modification_acquire() will observe the block as
-  being detached from buf_pool.flush_list, after reading the value 0. */
-  oldest_modification_.store(0, std::memory_order_release);
-}
-
-/** Note that a block is no longer dirty, while not removing
-it from buf_pool.flush_list */
-inline void buf_page_t::clear_oldest_modification(bool temporary)
-{
-  ut_ad(temporary == fsp_is_system_temporary(id().space()));
-  if (temporary)
-  {
-    ut_ad(oldest_modification() == 2);
-    oldest_modification_= 0;
-  }
-  else
-  {
-    /* We use release memory order to guarantee that callers of
-    oldest_modification_acquire() will observe the block as
-    being detached from buf_pool.flush_list, after reading the value 0. */
-    ut_ad(oldest_modification() > 2);
-    oldest_modification_.store(1, std::memory_order_release);
-  }
+  oldest_modification_= 0;
 }
 
 /** @return whether the block is modified and ready for flushing */
@@ -2259,10 +2116,7 @@ inline bool buf_page_t::ready_for_flush() const
   mysql_mutex_assert_owner(&buf_pool.mutex);
   ut_ad(in_LRU_list);
   ut_a(in_file());
-  ut_ad(fsp_is_system_temporary(id().space())
-        ? oldest_modification() == 2
-        : oldest_modification() > 2);
-  return io_fix_ == BUF_IO_NONE;
+  return oldest_modification() && io_fix_ == BUF_IO_NONE;
 }
 
 /** @return whether the block can be relocated in memory.
@@ -2339,7 +2193,7 @@ MEMORY:		is not in free list, LRU list, or flush list, nor page
 		hash table
 FILE_PAGE:	space and offset are defined, is in page hash table
 		if io_fix == BUF_IO_WRITE,
-			buf_pool.n_flush_LRU() || buf_pool.n_flush_list()
+			buf_pool.n_flush_LRU > 0 || buf_pool.n_flush_list > 0
 
 		(1) if buf_fix_count == 0, then
 			is in LRU list, not in free list
@@ -2426,7 +2280,7 @@ struct	CheckUnzipLRUAndLRUList {
 };
 #endif /* UNIV_DEBUG */
 
-#include "buf0buf.inl"
+#include "buf0buf.ic"
 
 #endif /* !UNIV_INNOCHECKSUM */
 
