@@ -2,7 +2,7 @@
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2009, Percona Inc.
-Copyright (c) 2013, 2022, MariaDB Corporation.
+Copyright (c) 2013, 2021, MariaDB Corporation.
 
 Portions of this file contain modifications contributed and copyrighted
 by Percona Inc.. Those modifications are
@@ -53,6 +53,8 @@ typedef ib_uint64_t os_offset_t;
 
 #ifdef _WIN32
 
+typedef HANDLE	os_file_dir_t;	/*!< directory stream */
+
 /** We define always WIN_ASYNC_IO, and check at run-time whether
 the OS actually supports it: Win 95 does not, NT does. */
 # define WIN_ASYNC_IO
@@ -66,6 +68,8 @@ typedef native_file_handle os_file_t;
 
 #else /* _WIN32 */
 
+typedef DIR*	os_file_dir_t;	/*!< directory stream */
+
 /** File handle */
 typedef int	os_file_t;
 
@@ -316,6 +320,43 @@ parameter (--tmpdir).
 FILE*
 os_file_create_tmpfile();
 
+/** The os_file_opendir() function opens a directory stream corresponding to the
+directory named by the dirname argument. The directory stream is positioned
+at the first entry. In both Unix and Windows we automatically skip the '.'
+and '..' items at the start of the directory listing.
+
+@param[in]	dirname		directory name; it must not contain a trailing
+				'\' or '/'
+@param[in]	is_fatal	true if we should treat an error as a fatal
+				error; if we try to open symlinks then we do
+				not wish a fatal error if it happens not to be
+				a directory
+@return directory stream, NULL if error */
+os_file_dir_t
+os_file_opendir(
+	const char*	dirname,
+	bool		is_fatal);
+
+/**
+Closes a directory stream.
+@param[in] dir	directory stream
+@return 0 if success, -1 if failure */
+int
+os_file_closedir(
+	os_file_dir_t	dir);
+
+/** This function returns information of the next file in the directory. We jump
+over the '.' and '..' entries in the directory.
+@param[in]	dirname		directory name or path
+@param[in]	dir		directory stream
+@param[out]	info		buffer where the info is returned
+@return 0 if ok, -1 if error, 1 if at the end of the directory */
+int
+os_file_readdir_next_file(
+	const char*	dirname,
+	os_file_dir_t	dir,
+	os_file_stat_t*	info);
+
 /**
 This function attempts to create a directory named pathname. The new directory
 gets default permissions. On Unix, the permissions are (0770 & ~umask). If the
@@ -387,14 +428,6 @@ os_file_set_nocache(
 	const char*	operation_name);
 #endif
 
-#ifndef _WIN32 /* On Microsoft Windows, mandatory locking is used */
-/** Obtain an exclusive lock on a file.
-@param fd      file descriptor
-@param name    file name
-@return 0 on success */
-int os_file_lock(int fd, const char *name);
-#endif
-
 /** NOTE! Use the corresponding macro os_file_create(), not directly
 this function!
 Opens an existing file or creates a new.
@@ -1119,14 +1152,14 @@ void os_aio_free();
 @retval DB_IO_ERROR on I/O error */
 dberr_t os_aio(const IORequest &type, void *buf, os_offset_t offset, size_t n);
 
-/** Wait until there are no pending asynchronous writes. */
-void os_aio_wait_until_no_pending_writes();
+/** Waits until there are no pending writes in os_aio_write_array. There can
+be other, synchronous, pending writes. */
+void
+os_aio_wait_until_no_pending_writes();
 
-
-/** Wait until there are no pending asynchronous reads. */
+/** Wait until all pending asynchronous reads have completed. */
 void os_aio_wait_until_no_pending_reads();
 
-
 /** Prints info of the aio arrays.
 @param[in/out]	file		file where to print */
 void
@@ -1207,33 +1240,36 @@ os_file_punch_hole(
 	os_offset_t	len)
 	MY_ATTRIBUTE((warn_unused_result));
 
-/** Normalizes a directory path for the current OS:
-On Windows, we convert '/' to '\', else we convert '\' to '/'.
-@param[in,out] str A null-terminated directory and file path */
-void os_normalize_path(char*	str);
-
 /* Determine if a path is an absolute path or not.
 @param[in]	OS directory or file path to evaluate
 @retval true if an absolute path
 @retval false if a relative path */
-UNIV_INLINE
-bool
-is_absolute_path(
-	const char*	path)
+inline bool is_absolute_path(const char *path)
 {
-	if (path[0] == OS_PATH_SEPARATOR) {
-		return(true);
-	}
+  switch (path[0]) {
+#ifdef _WIN32
+  case '\0':
+    return false;
+  case '\\':
+#endif
+  case '/':
+    return true;
+  }
 
 #ifdef _WIN32
-	if (path[1] == ':' && path[2] == OS_PATH_SEPARATOR) {
-		return(true);
-	}
+  if (path[1] == ':')
+  {
+    switch (path[2]) {
+    case '/':
+    case '\\':
+      return true;
+    }
+  }
 #endif /* _WIN32 */
 
-	return(false);
+  return false;
 }
 
-#include "os0file.inl"
+#include "os0file.ic"
 
 #endif /* os0file_h */
