@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1997, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2022, MariaDB Corporation.
+Copyright (c) 2017, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -25,8 +25,10 @@ Created 3/14/1997 Heikki Tuuri
 *******************************************************/
 
 #include "row0purge.h"
+#include "btr0cur.h"
 #include "fsp0fsp.h"
 #include "mach0data.h"
+#include "dict0crea.h"
 #include "dict0stats.h"
 #include "trx0rseg.h"
 #include "trx0trx.h"
@@ -72,9 +74,7 @@ row_purge_reposition_pcur(
 	if (node->found_clust) {
 		ut_ad(node->validate_pcur());
 
-		node->found_clust =
-		  btr_pcur_restore_position(mode, &node->pcur, mtr) ==
-		    btr_pcur_t::SAME_ALL;
+		node->found_clust = btr_pcur_restore_position(mode, &node->pcur, mtr);
 
 	} else {
 		node->found_clust = row_search_on_row_ref(
@@ -118,6 +118,21 @@ row_purge_remove_clust_if_poss_low(
 		return true;
 	}
 
+	if (node->table->id == DICT_INDEXES_ID) {
+		/* If this is a record of the SYS_INDEXES table, then
+		we have to free the file segments of the index tree
+		associated with the index */
+		dict_drop_index_tree(&node->pcur, nullptr, &mtr);
+		mtr.commit();
+		mtr.start();
+		index->set_modified(mtr);
+
+		if (!row_purge_reposition_pcur(mode, node, &mtr)) {
+			mtr.commit();
+			return true;
+		}
+	}
+
 	rec_t* rec = btr_pcur_get_rec(&node->pcur);
 	rec_offs offsets_[REC_OFFS_NORMAL_SIZE];
 	rec_offs_init(offsets_);
@@ -197,7 +212,7 @@ row_purge_remove_clust_if_poss(
 			return(true);
 		}
 
-		os_thread_sleep(BTR_CUR_RETRY_SLEEP_TIME);
+		std::this_thread::sleep_for(BTR_CUR_RETRY_SLEEP_TIME);
 	}
 
 	return(false);
@@ -445,21 +460,19 @@ row_purge_remove_sec_if_poss_leaf(
 
 	/* Set the purge node for the call to row_purge_poss_sec(). */
 	pcur.btr_cur.purge_node = node;
-	if (dict_index_is_spatial(index)) {
-		rw_lock_sx_lock(dict_index_get_lock(index));
+	if (index->is_spatial()) {
 		pcur.btr_cur.thr = NULL;
+		index->lock.u_lock(SRW_LOCK_CALL);
+		search_result = row_search_index_entry(
+			index, entry, mode, &pcur, &mtr);
+		index->lock.u_unlock();
 	} else {
 		/* Set the query thread, so that ibuf_insert_low() will be
 		able to invoke thd_get_trx(). */
 		pcur.btr_cur.thr = static_cast<que_thr_t*>(
 			que_node_get_parent(node));
-	}
-
-	search_result = row_search_index_entry(
-		index, entry, mode, &pcur, &mtr);
-
-	if (dict_index_is_spatial(index)) {
-		rw_lock_sx_unlock(dict_index_get_lock(index));
+		search_result = row_search_index_entry(
+			index, entry, mode, &pcur, &mtr);
 	}
 
 	switch (search_result) {
@@ -581,7 +594,7 @@ retry:
 
 		n_tries++;
 
-		os_thread_sleep(BTR_CUR_RETRY_SLEEP_TIME);
+		std::this_thread::sleep_for(BTR_CUR_RETRY_SLEEP_TIME);
 
 		goto retry;
 	}
@@ -838,8 +851,6 @@ skip_secondaries:
 				page_id_t(rseg->space->id, page_no),
 				0, RW_X_LATCH, &mtr);
 
-			buf_block_dbg_add_level(block, SYNC_TRX_UNDO_PAGE);
-
 			data_field = buf_block_get_frame(block)
 				+ offset + internal_offset;
 
@@ -865,138 +876,6 @@ skip_secondaries:
 	row_purge_upd_exist_or_extern_func(node,undo_rec)
 #endif /* UNIV_DEBUG */
 
-/** Build a partial row from an update undo log record for purge.
-Any columns which occur as ordering in any index of the table are present.
-Any missing columns are indicated by col->mtype == DATA_MISSING.
-
-@param ptr    remaining part of the undo log record
-@param index  clustered index
-@param node   purge node
-@return pointer to remaining part of undo record */
-static byte *row_purge_get_partial(const byte *ptr, const dict_index_t &index,
-                                   purge_node_t *node)
-{
-  bool first_v_col= true;
-  bool is_undo_log= true;
-
-  ut_ad(index.is_primary());
-  ut_ad(index.n_uniq == node->ref->n_fields);
-
-  node->row= dtuple_create_with_vcol(node->heap, index.table->n_cols,
-                                     index.table->n_v_cols);
-
-  /* Mark all columns in the row uninitialized, so that
-  we can distinguish missing fields from fields that are SQL NULL. */
-  for (ulint i= 0; i < index.table->n_cols; i++)
-    node->row->fields[i].type.mtype= DATA_MISSING;
-
-  dtuple_init_v_fld(node->row);
-
-  for (const upd_field_t *uf= node->update->fields, *const ue=
-         node->update->fields + node->update->n_fields; uf != ue; uf++)
-  {
-    if (!uf->old_v_val)
-    {
-      const dict_col_t &c= *dict_index_get_nth_col(&index, uf->field_no);
-      if (!c.is_dropped())
-        node->row->fields[c.ind]= uf->new_val;
-    }
-  }
-
-  const byte *end_ptr= ptr + mach_read_from_2(ptr);
-  ptr+= 2;
-
-  while (ptr != end_ptr)
-  {
-    dfield_t *dfield;
-    const byte *field;
-    const dict_col_t *col;
-    uint32_t len, orig_len, field_no= mach_read_next_compressed(&ptr);
-
-    if (field_no >= REC_MAX_N_FIELDS)
-    {
-      ptr= trx_undo_read_v_idx(index.table, ptr, first_v_col, &is_undo_log,
-                               &field_no);
-      first_v_col= false;
-
-      ptr= trx_undo_rec_get_col_val(ptr, &field, &len, &orig_len);
-
-      if (field_no == FIL_NULL)
-        continue; /* there no longer is an index on the virtual column */
-
-      dict_v_col_t *vcol= dict_table_get_nth_v_col(index.table, field_no);
-      col =&vcol->m_col;
-      dfield= dtuple_get_nth_v_field(node->row, vcol->v_pos);
-      dict_col_copy_type(&vcol->m_col, &dfield->type);
-    }
-    else
-    {
-      ptr= trx_undo_rec_get_col_val(ptr, &field, &len, &orig_len);
-      col= dict_index_get_nth_col(&index, field_no);
-      if (col->is_dropped())
-        continue;
-      dfield= dtuple_get_nth_field(node->row, col->ind);
-      ut_ad(dfield->type.mtype == DATA_MISSING ||
-            dict_col_type_assert_equal(col, &dfield->type));
-      ut_ad(dfield->type.mtype == DATA_MISSING ||
-            dfield->len == len ||
-            (len != UNIV_SQL_NULL && len >= UNIV_EXTERN_STORAGE_FIELD));
-      dict_col_copy_type(dict_table_get_nth_col(index.table, col->ind),
-                         &dfield->type);
-    }
-
-    dfield_set_data(dfield, field, len);
-
-    if (len == UNIV_SQL_NULL || len < UNIV_EXTERN_STORAGE_FIELD)
-      continue;
-
-    spatial_status_t spatial_status= static_cast<spatial_status_t>
-      ((len & SPATIAL_STATUS_MASK) >> SPATIAL_STATUS_SHIFT);
-    len&= ~SPATIAL_STATUS_MASK;
-
-    /* Keep compatible with 5.7.9 format. */
-    if (spatial_status == SPATIAL_UNKNOWN)
-      spatial_status= dict_col_get_spatial_status(col);
-
-    switch (UNIV_EXPECT(spatial_status, SPATIAL_NONE)) {
-    case SPATIAL_ONLY:
-      ut_ad(len - UNIV_EXTERN_STORAGE_FIELD == DATA_MBR_LEN);
-      dfield_set_len(dfield, len - UNIV_EXTERN_STORAGE_FIELD);
-      break;
-
-    case SPATIAL_MIXED:
-      dfield_set_len(dfield, len - UNIV_EXTERN_STORAGE_FIELD - DATA_MBR_LEN);
-      break;
-
-    default:
-      dfield_set_len(dfield, len - UNIV_EXTERN_STORAGE_FIELD);
-      break;
-    }
-
-    dfield_set_ext(dfield);
-    dfield_set_spatial_status(dfield, spatial_status);
-
-    if (!col->ord_part || spatial_status == SPATIAL_ONLY ||
-        node->rec_type == TRX_UNDO_UPD_DEL_REC)
-      continue;
-    /* If the prefix of this BLOB column is indexed, ensure that enough
-    prefix is stored in the undo log record. */
-    ut_a(dfield_get_len(dfield) >= BTR_EXTERN_FIELD_REF_SIZE);
-    ut_a(dict_table_has_atomic_blobs(index.table) ||
-         dfield_get_len(dfield) >=
-         REC_ANTELOPE_MAX_INDEX_COL_LEN + BTR_EXTERN_FIELD_REF_SIZE);
-  }
-
-  for (ulint i= 0; i < index.n_uniq; i++)
-  {
-    dfield_t &field= node->row->fields[index.fields[i].col->ind];
-    if (field.type.mtype == DATA_MISSING)
-      field= node->ref->fields[i];
-  }
-
-  return const_cast<byte*>(ptr);
-}
-
 /** Parses the row reference and other info in a modify undo log record.
 @param[in]	node		row undo node
 @param[in]	undo_rec	record to purge
@@ -1032,6 +911,7 @@ row_purge_parse_undo_rec(
 	switch (type) {
 	case TRX_UNDO_RENAME_TABLE:
 		return false;
+	case TRX_UNDO_EMPTY:
 	case TRX_UNDO_INSERT_METADATA:
 	case TRX_UNDO_INSERT_REC:
 		/* These records do not store any transaction identifier.
@@ -1065,11 +945,11 @@ row_purge_parse_undo_rec(
 	}
 
 try_again:
-	node->table = dict_table_open_on_id<true>(
+	node->table = dict_table_open_on_id(
 		table_id, false, DICT_TABLE_OP_NORMAL, node->purge_thd,
 		&node->mdl_ticket);
 
-	if (!node->table) {
+	if (node->table == NULL || node->table->name.is_temporary()) {
 		/* The table has been dropped: no need to do purge and
 		release mdl happened as a part of open process itself */
 		goto err_exit;
@@ -1091,10 +971,10 @@ already_locked:
 		if (!mysqld_server_started) {
 
 			node->close_table();
-			if (srv_shutdown_state > SRV_SHUTDOWN_NONE) {
+			if (srv_shutdown_state > SRV_SHUTDOWN_INITIATED) {
 				return(false);
 			}
-			os_thread_sleep(1000000);
+			std::this_thread::sleep_for(std::chrono::seconds(1));
 			goto try_again;
 		}
 	}
@@ -1122,6 +1002,9 @@ err_exit:
 	if (type == TRX_UNDO_INSERT_METADATA) {
 		node->ref = &trx_undo_metadata;
 		return(true);
+	} else if (type == TRX_UNDO_EMPTY) {
+		node->ref = nullptr;
+		return true;
 	}
 
 	ptr = trx_undo_rec_get_row_ref(ptr, clust_index, &(node->ref),
@@ -1140,7 +1023,10 @@ err_exit:
 
 	if (!(node->cmpl_info & UPD_NODE_NO_ORD_CHANGE)) {
 		ut_ad(!(node->update->info_bits & REC_INFO_MIN_REC_FLAG));
-		ptr = row_purge_get_partial(ptr, *clust_index, node);
+		ptr = trx_undo_rec_get_partial_row(
+			ptr, clust_index, node->update, &node->row,
+			type == TRX_UNDO_UPD_DEL_REC,
+			node->heap);
 	} else if (node->update->info_bits & REC_INFO_MIN_REC_FLAG) {
 		node->ref = &trx_undo_metadata;
 	}
@@ -1176,6 +1062,8 @@ row_purge_record_func(
 	ut_ad(!trx_undo_roll_ptr_is_insert(node->roll_ptr));
 
 	switch (node->rec_type) {
+	case TRX_UNDO_EMPTY:
+		break;
 	case TRX_UNDO_DEL_MARK_REC:
 		purged = row_purge_del_mark(node);
 		if (purged) {
@@ -1247,7 +1135,7 @@ row_purge(
 			}
 
 			/* Retry the purge in a second. */
-			os_thread_sleep(1000000);
+			std::this_thread::sleep_for(std::chrono::seconds(1));
 		}
 	}
 }
