@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1997, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -110,7 +110,8 @@ row_undo_mod_clust_low(
 	ut_ad(success);
 	ut_ad(rec_get_trx_id(btr_cur_get_rec(btr_cur),
 			     btr_cur_get_index(btr_cur))
-	      == thr_get_trx(thr)->id);
+	      == thr_get_trx(thr)->id
+	      || btr_cur_get_index(btr_cur)->table->is_temporary());
 	ut_ad(node->ref != &trx_undo_metadata
 	      || node->update->info_bits == REC_INFO_METADATA_ADD
 	      || node->update->info_bits == REC_INFO_METADATA_ALTER);
@@ -189,8 +190,9 @@ static ulint row_trx_id_offset(const rec_t* rec, const dict_index_t* index)
 		rec_offs_init(offsets_);
 		mem_heap_t* heap = NULL;
 		const ulint trx_id_pos = index->n_uniq ? index->n_uniq : 1;
-		rec_offs* offsets = rec_get_offsets(rec, index, offsets_, true,
-						 trx_id_pos + 1, &heap);
+		rec_offs* offsets = rec_get_offsets(rec, index, offsets_,
+						    index->n_core_fields,
+						    trx_id_pos + 1, &heap);
 		ut_ad(!heap);
 		ulint len;
 		trx_id_offset = rec_get_nth_field_offs(
@@ -214,8 +216,6 @@ static bool row_undo_mod_must_purge(undo_node_t* node, mtr_t* mtr)
 	ut_ad(btr_cur->index->is_primary());
 	DEBUG_SYNC_C("rollback_purge_clust");
 
-	mtr->s_lock(&purge_sys.latch, __FILE__, __LINE__);
-
 	if (!purge_sys.changes_visible(node->new_trx_id, node->table->name)) {
 		return false;
 	}
@@ -239,6 +239,7 @@ row_undo_mod_clust(
 {
 	btr_pcur_t*	pcur;
 	mtr_t		mtr;
+	bool		have_latch = false;
 	dberr_t		err;
 	dict_index_t*	index;
 	bool		online;
@@ -246,8 +247,6 @@ row_undo_mod_clust(
 	ut_ad(thr_get_trx(thr) == node->trx);
 	ut_ad(node->trx->dict_operation_lock_mode);
 	ut_ad(node->trx->in_rollback);
-	ut_ad(rw_lock_own_flagged(&dict_sys.latch,
-				  RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 
 	log_free_check();
 	pcur = &node->pcur;
@@ -308,11 +307,7 @@ row_undo_mod_clust(
 	ut_ad(online || !dict_index_is_online_ddl(index));
 
 	if (err == DB_SUCCESS && online) {
-
-		ut_ad(rw_lock_own_flagged(
-				&index->lock,
-				RW_LOCK_FLAG_S | RW_LOCK_FLAG_X
-				| RW_LOCK_FLAG_SX));
+		ut_ad(index->lock.have_any());
 
 		switch (node->rec_type) {
 		case TRX_UNDO_DEL_MARK_REC:
@@ -365,23 +360,32 @@ row_undo_mod_clust(
 			goto mtr_commit_exit;
 		}
 
+		ut_ad(rec_get_deleted_flag(btr_pcur_get_rec(pcur),
+					   dict_table_is_comp(node->table)));
+
 		if (index->table->is_temporary()) {
 			mtr.set_log_mode(MTR_LOG_NO_REDO);
+			if (btr_cur_optimistic_delete(&pcur->btr_cur, 0,
+						      &mtr)) {
+				goto mtr_commit_exit;
+			}
+			btr_pcur_commit_specify_mtr(pcur, &mtr);
 		} else {
+			index->set_modified(mtr);
+			have_latch = true;
+			purge_sys.latch.rd_lock(SRW_LOCK_CALL);
 			if (!row_undo_mod_must_purge(node, &mtr)) {
 				goto mtr_commit_exit;
 			}
-			index->set_modified(mtr);
+			if (btr_cur_optimistic_delete(&pcur->btr_cur, 0,
+						      &mtr)) {
+				goto mtr_commit_exit;
+			}
+			purge_sys.latch.rd_unlock();
+			btr_pcur_commit_specify_mtr(pcur, &mtr);
+			have_latch = false;
 		}
 
-		ut_ad(rec_get_deleted_flag(btr_pcur_get_rec(pcur),
-					   dict_table_is_comp(node->table)));
-		if (btr_cur_optimistic_delete(&pcur->btr_cur, 0, &mtr)) {
-			goto mtr_commit_exit;
-		}
-
-		btr_pcur_commit_specify_mtr(pcur, &mtr);
-
 		mtr.start();
 		if (!btr_pcur_restore_position(
 			    BTR_MODIFY_TREE | BTR_LATCH_FOR_DELETE,
@@ -389,18 +393,20 @@ row_undo_mod_clust(
 			goto mtr_commit_exit;
 		}
 
+		ut_ad(rec_get_deleted_flag(btr_pcur_get_rec(pcur),
+					   dict_table_is_comp(node->table)));
+
 		if (index->table->is_temporary()) {
 			mtr.set_log_mode(MTR_LOG_NO_REDO);
 		} else {
+			have_latch = true;
+			purge_sys.latch.rd_lock(SRW_LOCK_CALL);
 			if (!row_undo_mod_must_purge(node, &mtr)) {
 				goto mtr_commit_exit;
 			}
 			index->set_modified(mtr);
 		}
 
-		ut_ad(rec_get_deleted_flag(btr_pcur_get_rec(pcur),
-					   dict_table_is_comp(node->table)));
-
 		/* This operation is analogous to purge, we can free
 		also inherited externally stored fields. We can also
 		assume that the record was complete (including BLOBs),
@@ -409,8 +415,7 @@ row_undo_mod_clust(
 		rollback=false, just like purge does. */
 		btr_cur_pessimistic_delete(&err, FALSE, &pcur->btr_cur, 0,
 					   false, &mtr);
-		ut_ad(err == DB_SUCCESS
-		      || err == DB_OUT_OF_FILE_SPACE);
+		ut_ad(err == DB_SUCCESS || err == DB_OUT_OF_FILE_SPACE);
 	} else if (!index->table->is_temporary() && node->new_trx_id) {
 		/* We rolled back a record so that it still exists.
 		We must reset the DB_TRX_ID if the history is no
@@ -421,7 +426,8 @@ row_undo_mod_clust(
 			goto mtr_commit_exit;
 		}
 		rec_t* rec = btr_pcur_get_rec(pcur);
-		mtr.s_lock(&purge_sys.latch, __FILE__, __LINE__);
+		have_latch = true;
+		purge_sys.latch.rd_lock(SRW_LOCK_CALL);
 		if (!purge_sys.changes_visible(node->new_trx_id,
 					       node->table->name)) {
 			goto mtr_commit_exit;
@@ -457,9 +463,9 @@ row_undo_mod_clust(
 		} else {
 			ut_ad(index->n_uniq <= MAX_REF_PARTS);
 			rec_offs_init(offsets_);
-			offsets = rec_get_offsets(
-				rec, index, offsets_, true, trx_id_pos + 2,
-				&heap);
+			offsets = rec_get_offsets(rec, index, offsets_,
+						  index->n_core_fields,
+						  trx_id_pos + 2, &heap);
 			ulint len;
 			trx_id_offset = rec_get_nth_field_offs(
 				offsets, trx_id_pos, &len);
@@ -493,6 +499,10 @@ row_undo_mod_clust(
 	}
 
 mtr_commit_exit:
+	if (have_latch) {
+		purge_sys.latch.rd_unlock();
+	}
+
 	btr_pcur_commit_specify_mtr(pcur, &mtr);
 
 func_exit:
@@ -845,7 +855,8 @@ try_again:
 		offsets_heap = NULL;
 		offsets = rec_get_offsets(
 			btr_cur_get_rec(btr_cur),
-			index, NULL, true, ULINT_UNDEFINED, &offsets_heap);
+			index, nullptr, index->n_core_fields, ULINT_UNDEFINED,
+			&offsets_heap);
 		update = row_upd_build_sec_rec_difference_binary(
 			btr_cur_get_rec(btr_cur), index, offsets, entry, heap);
 		if (upd_get_n_fields(update) == 0) {
@@ -904,9 +915,9 @@ row_undo_mod_sec_flag_corrupted(
 		on the data dictionary during normal rollback,
 		we can only mark the index corrupted in the
 		data dictionary cache. TODO: fix this somehow.*/
-		mutex_enter(&dict_sys.mutex);
+		dict_sys.mutex_lock();
 		dict_set_corrupted_index_cache_only(index);
-		mutex_exit(&dict_sys.mutex);
+		dict_sys.mutex_unlock();
 		break;
 	default:
 		ut_ad(0);
@@ -1222,9 +1233,9 @@ static bool row_undo_mod_parse_undo_rec(undo_node_t* node, bool dict_locked)
 		node->table = dict_table_open_on_id(table_id, dict_locked,
 						    DICT_TABLE_OP_NORMAL);
 	} else if (!dict_locked) {
-		mutex_enter(&dict_sys.mutex);
+		dict_sys.mutex_lock();
 		node->table = dict_sys.get_temporary_table(table_id);
-		mutex_exit(&dict_sys.mutex);
+		dict_sys.mutex_unlock();
 	} else {
 		node->table = dict_sys.get_temporary_table(table_id);
 	}
