@@ -64,12 +64,6 @@ const byte timestamp_max_bytes[7] = {
 
 static const ulint MAX_DETAILED_ERROR_LEN = 256;
 
-/** Set of table_id */
-typedef std::set<
-	table_id_t,
-	std::less<table_id_t>,
-	ut_allocator<table_id_t> >	table_id_set;
-
 /*************************************************************//**
 Set detailed error message for the transaction. */
 void
@@ -137,20 +131,20 @@ trx_init(
 
 	trx->auto_commit = false;
 
-	trx->will_lock = false;
+	trx->will_lock = 0;
 
 	trx->ddl = false;
 
 	trx->internal = false;
 
+	trx->bulk_insert = false;
+
 	ut_d(trx->start_file = 0);
 
 	ut_d(trx->start_line = 0);
 
 	trx->magic_n = TRX_MAGIC_N;
 
-	trx->lock.que_state = TRX_QUE_RUNNING;
-
 	trx->last_sql_stat_start.least_undo_no = 0;
 
 	ut_ad(!trx->read_view.is_open());
@@ -160,7 +154,6 @@ trx_init(
 	trx->lock.table_cached = 0;
 #ifdef WITH_WSREP
 	ut_ad(!trx->wsrep);
-	ut_ad(!trx->wsrep_UK_scan);
 #endif /* WITH_WSREP */
 }
 
@@ -196,9 +189,9 @@ struct TrxFactory {
 
 		trx->lock.lock_heap = mem_heap_create_typed(
 			1024, MEM_HEAP_FOR_LOCK_HEAP);
+		pthread_cond_init(&trx->lock.cond, nullptr);
 
-		lock_trx_lock_list_init(&trx->lock.trx_locks);
-
+		UT_LIST_INIT(trx->lock.trx_locks, &lock_t::trx_locks);
 		UT_LIST_INIT(trx->lock.evicted_tables,
 			     &dict_table_t::table_LRU);
 
@@ -206,7 +199,7 @@ struct TrxFactory {
 			trx->trx_savepoints,
 			&trx_named_savept_t::trx_savepoints);
 
-		mutex_create(LATCH_ID_TRX, &trx->mutex);
+		trx->mutex_init();
 	}
 
 	/** Release resources held by the transaction object.
@@ -237,13 +230,15 @@ struct TrxFactory {
 			trx->lock.lock_heap = NULL;
 		}
 
+		pthread_cond_destroy(&trx->lock.cond);
+
 		ut_a(UT_LIST_GET_LEN(trx->lock.trx_locks) == 0);
 		ut_ad(UT_LIST_GET_LEN(trx->lock.evicted_tables) == 0);
 
 		UT_DELETE(trx->xid);
 		ut_free(trx->detailed_error);
 
-		mutex_free(&trx->mutex);
+		trx->mutex_destroy();
 
 		trx->mod_tables.~trx_mod_tables_t();
 
@@ -256,49 +251,47 @@ struct TrxFactory {
 };
 
 /** The lock strategy for TrxPool */
-struct TrxPoolLock {
-	TrxPoolLock() { }
+class TrxPoolLock
+{
+  mysql_mutex_t mutex;
 
-	/** Create the mutex */
-	void create()
-	{
-		mutex_create(LATCH_ID_TRX_POOL, &m_mutex);
-	}
+public:
+  /** Create the mutex */
+  void create()
+  {
+    mysql_mutex_init(trx_pool_mutex_key, &mutex, nullptr);
+  }
 
-	/** Acquire the mutex */
-	void enter() { mutex_enter(&m_mutex); }
+  /** Acquire the mutex */
+  void enter() { mysql_mutex_lock(&mutex); }
 
-	/** Release the mutex */
-	void exit() { mutex_exit(&m_mutex); }
+  /** Release the mutex */
+  void exit() { mysql_mutex_unlock(&mutex); }
 
-	/** Free the mutex */
-	void destroy() { mutex_free(&m_mutex); }
-
-	/** Mutex to use */
-	ib_mutex_t	m_mutex;
+  /** Free the mutex */
+  void destroy() { mysql_mutex_destroy(&mutex); }
 };
 
 /** The lock strategy for the TrxPoolManager */
-struct TrxPoolManagerLock {
-	TrxPoolManagerLock() { }
+class TrxPoolManagerLock
+{
+  mysql_mutex_t mutex;
 
-	/** Create the mutex */
-	void create()
-	{
-		mutex_create(LATCH_ID_TRX_POOL_MANAGER, &m_mutex);
-	}
+public:
+  /** Create the mutex */
+  void create()
+  {
+    mysql_mutex_init(trx_pool_manager_mutex_key, &mutex, nullptr);
+  }
 
-	/** Acquire the mutex */
-	void enter() { mutex_enter(&m_mutex); }
+  /** Acquire the mutex */
+  void enter() { mysql_mutex_lock(&mutex); }
 
-	/** Release the mutex */
-	void exit() { mutex_exit(&m_mutex); }
+  /** Release the mutex */
+  void exit() { mysql_mutex_unlock(&mutex); }
 
-	/** Free the mutex */
-	void destroy() { mutex_free(&m_mutex); }
-
-	/** Mutex to use */
-	ib_mutex_t	m_mutex;
+  /** Free the mutex */
+  void destroy() { mysql_mutex_destroy(&mutex); }
 };
 
 /** Use explicit mutexes for the trx_t pool and its manager. */
@@ -353,7 +346,7 @@ trx_t *trx_create()
 	ib_alloc_t*	alloc;
 
 	/* We just got trx from pool, it should be non locking */
-	ut_ad(!trx->will_lock);
+	ut_ad(trx->will_lock == 0);
 	ut_ad(!trx->rw_trx_hash_pins);
 
 	DBUG_LOG("trx", "Create: " << trx);
@@ -370,10 +363,6 @@ trx_t *trx_create()
 	ut_ad(trx->lock.rec_cached == 0);
 	ut_ad(UT_LIST_GET_LEN(trx->lock.evicted_tables) == 0);
 
-#ifdef WITH_WSREP
-	ut_ad(!trx->wsrep_UK_scan);
-#endif /* WITH_WSREP */
-
 	trx_sys.register_trx(trx);
 
 	return(trx);
@@ -410,10 +399,7 @@ void trx_t::free()
     autoinc_locks= NULL;
   }
 
-  mod_tables.clear();
-
-  MEM_NOACCESS(&skip_lock_inheritance_and_n_ref,
-               sizeof skip_lock_inheritance_and_n_ref);
+  MEM_NOACCESS(&n_ref, sizeof n_ref);
   /* do not poison mutex */
   MEM_NOACCESS(&id, sizeof id);
   MEM_NOACCESS(&state, sizeof state);
@@ -471,10 +457,6 @@ void trx_t::free()
   MEM_NOACCESS(&xid, sizeof xid);
   MEM_NOACCESS(&mod_tables, sizeof mod_tables);
   MEM_NOACCESS(&detailed_error, sizeof detailed_error);
-#ifdef WITH_WSREP
-  ut_ad(!wsrep_UK_scan);
-  MEM_NOACCESS(&wsrep_UK_scan, sizeof wsrep_UK_scan);
-#endif /* WITH_WSREP */
   MEM_NOACCESS(&magic_n, sizeof magic_n);
   trx_pools->mem_free(this);
 }
@@ -497,9 +479,9 @@ inline void trx_t::commit_state()
   makes modifications to the database, will get an lsn larger than the
   committing transaction T. In the case where the log flush fails, and
   T never gets committed, also T2 will never get committed. */
-  trx_mutex_enter(this);
+  mutex.wr_lock();
   state= TRX_STATE_COMMITTED_IN_MEMORY;
-  trx_mutex_exit(this);
+  mutex.wr_unlock();
   ut_ad(id || !is_referenced());
 }
 
@@ -512,14 +494,13 @@ inline void trx_t::release_locks()
   if (UT_LIST_GET_LEN(lock.trx_locks))
   {
     lock_release(this);
-    lock.n_rec_locks = 0;
+    ut_ad(!lock.n_rec_locks);
     ut_ad(UT_LIST_GET_LEN(lock.trx_locks) == 0);
     ut_ad(ib_vector_is_empty(autoinc_locks));
     mem_heap_empty(lock.lock_heap);
   }
 
   lock.table_locks.clear();
-  reset_skip_lock_inheritance();
 }
 
 /** At shutdown, frees a transaction object. */
@@ -541,6 +522,7 @@ trx_free_at_shutdown(trx_t *trx)
 
 	trx->commit_state();
 	trx->release_locks();
+	trx->mod_tables.clear();
 	trx_undo_free_at_shutdown(trx);
 
 	ut_a(!trx->read_only);
@@ -563,12 +545,10 @@ void trx_disconnect_prepared(trx_t *trx)
   ut_ad(trx->mysql_thd);
   ut_ad(!trx->mysql_log_file_name);
   trx->read_view.close();
-  trx_sys.trx_list.freeze();
   trx->is_recovered= true;
   trx->mysql_thd= NULL;
-  trx_sys.trx_list.unfreeze();
   /* todo/fixme: suggest to do it at innodb prepare */
-  trx->will_lock= false;
+  trx->will_lock= 0;
   trx_sys.rw_trx_hash.put_pins(trx);
 }
 
@@ -581,9 +561,6 @@ trx_resurrect_table_locks(
 	trx_t*			trx,	/*!< in/out: transaction */
 	const trx_undo_t*	undo)	/*!< in: undo log */
 {
-	mtr_t			mtr;
-	table_id_set		tables;
-
 	ut_ad(trx_state_eq(trx, TRX_STATE_ACTIVE) ||
 	      trx_state_eq(trx, TRX_STATE_PREPARED));
 	ut_ad(undo->rseg == trx->rsegs.m_redo.rseg);
@@ -592,7 +569,9 @@ trx_resurrect_table_locks(
 		return;
 	}
 
-	mtr_start(&mtr);
+	mtr_t mtr;
+	std::map<table_id_t, bool> tables;
+	mtr.start();
 
 	/* trx_rseg_mem_create() may have acquired an X-latch on this
 	page, so we cannot acquire an S-latch. */
@@ -617,37 +596,37 @@ trx_resurrect_table_locks(
 		trx_undo_rec_get_pars(
 			undo_rec, &type, &cmpl_info,
 			&updated_extern, &undo_no, &table_id);
-		tables.insert(table_id);
+
+		tables.emplace(table_id, type == TRX_UNDO_EMPTY);
 
 		undo_rec = trx_undo_get_prev_rec(
 			block, page_offset(undo_rec), undo->hdr_page_no,
 			undo->hdr_offset, false, &mtr);
 	} while (undo_rec);
 
-	mtr_commit(&mtr);
+	mtr.commit();
 
-	for (table_id_set::const_iterator i = tables.begin();
-	     i != tables.end(); i++) {
+	for (auto p : tables) {
 		if (dict_table_t* table = dict_table_open_on_id(
-			    *i, FALSE, DICT_TABLE_OP_LOAD_TABLESPACE)) {
+			    p.first, FALSE, DICT_TABLE_OP_LOAD_TABLESPACE)) {
 			if (!table->is_readable()) {
-				mutex_enter(&dict_sys.mutex);
+				dict_sys.mutex_lock();
 				dict_table_close(table, TRUE, FALSE);
 				dict_sys.remove(table);
-				mutex_exit(&dict_sys.mutex);
+				dict_sys.mutex_unlock();
 				continue;
 			}
 
 			if (trx->state == TRX_STATE_PREPARED) {
-				trx->mod_tables.insert(
-					trx_mod_tables_t::value_type(table,
-								     0));
+				trx->mod_tables.emplace(table, 0);
 			}
-			lock_table_ix_resurrect(table, trx);
+
+			lock_table_resurrect(table, trx,
+					     p.second ? LOCK_X : LOCK_IX);
 
 			DBUG_LOG("ib_trx",
 				 "resurrect " << ib::hex(trx->id)
-				 << " IX lock on " << table->name);
+				 << " lock on " << table->name);
 
 			dict_table_close(table, FALSE, FALSE);
 		}
@@ -662,7 +641,8 @@ trx_resurrect_table_locks(
 
 static void trx_resurrect(trx_undo_t *undo, trx_rseg_t *rseg,
                           time_t start_time, ulonglong start_time_micro,
-                          uint64_t *rows_to_undo)
+                          uint64_t *rows_to_undo,
+                          bool is_old_insert)
 {
   trx_state_t state;
   /*
@@ -685,6 +665,8 @@ static void trx_resurrect(trx_undo_t *undo, trx_rseg_t *rseg,
     state= TRX_STATE_PREPARED;
     break;
   default:
+    if (is_old_insert && srv_force_recovery < SRV_FORCE_NO_TRX_UNDO)
+      trx_undo_commit_cleanup(undo, false);
     return;
   }
 
@@ -693,7 +675,11 @@ static void trx_resurrect(trx_undo_t *undo, trx_rseg_t *rseg,
   ut_d(trx->start_file= __FILE__);
   ut_d(trx->start_line= __LINE__);
 
-  trx->rsegs.m_redo.undo= undo;
+  if (is_old_insert)
+    trx->rsegs.m_redo.old_insert= undo;
+  else
+    trx->rsegs.m_redo.undo= undo;
+
   trx->undo_no= undo->top_undo_no + 1;
   trx->rsegs.m_redo.rseg= rseg;
   /*
@@ -724,7 +710,8 @@ static void trx_resurrect(trx_undo_t *undo, trx_rseg_t *rseg,
 
 
 /** Initialize (resurrect) transactions at startup. */
-dberr_t trx_lists_init_at_db_start()
+void
+trx_lists_init_at_db_start()
 {
 	ut_a(srv_is_being_started);
 	ut_ad(!srv_was_started);
@@ -733,18 +720,16 @@ dberr_t trx_lists_init_at_db_start()
 		/* mariabackup --prepare only deals with
 		the redo log and the data files, not with
 		transactions or the data dictionary. */
-		return trx_rseg_array_init();
+		trx_rseg_array_init();
+		return;
 	}
 
 	if (srv_force_recovery >= SRV_FORCE_NO_UNDO_LOG_SCAN) {
-		return DB_SUCCESS;
+		return;
 	}
 
 	purge_sys.create();
-	if (dberr_t err = trx_rseg_array_init()) {
-		ib::info() << "Retry with innodb_force_recovery=5";
-		return err;
-	}
+	trx_rseg_array_init();
 
 	/* Look from the rollback segments if there exist undo logs for
 	transactions. */
@@ -762,6 +747,17 @@ dberr_t trx_lists_init_at_db_start()
 		if (rseg == NULL) {
 			continue;
 		}
+
+		/* Resurrect transactions that were doing inserts
+		using the old separate insert_undo log. */
+		undo = UT_LIST_GET_FIRST(rseg->old_insert_list);
+		while (undo) {
+			trx_undo_t* next = UT_LIST_GET_NEXT(undo_list, undo);
+			trx_resurrect(undo, rseg, start_time, start_time_micro,
+				      &rows_to_undo, true);
+			undo = next;
+		}
+
 		/* Ressurrect other transactions. */
 		for (undo = UT_LIST_GET_FIRST(rseg->undo_list);
 		     undo != NULL;
@@ -769,7 +765,8 @@ dberr_t trx_lists_init_at_db_start()
 			trx_t *trx = trx_sys.find(0, undo->trx_id, false);
 			if (!trx) {
 				trx_resurrect(undo, rseg, start_time,
-					      start_time_micro, &rows_to_undo);
+					      start_time_micro,
+					      &rows_to_undo, false);
 			} else {
 				ut_ad(trx_state_eq(trx, TRX_STATE_ACTIVE) ||
 				      trx_state_eq(trx, TRX_STATE_PREPARED));
@@ -794,16 +791,16 @@ dberr_t trx_lists_init_at_db_start()
 		}
 	}
 
-	if (const auto size = trx_sys.rw_trx_hash.size()) {
-		ib::info() << size
+	if (trx_sys.rw_trx_hash.size()) {
+
+		ib::info() << trx_sys.rw_trx_hash.size()
 			<< " transaction(s) which must be rolled back or"
 			" cleaned up in total " << rows_to_undo
 			<< " row operations to undo";
+
 		ib::info() << "Trx id counter is " << trx_sys.get_max_trx_id();
 	}
-
 	purge_sys.clone_oldest_view();
-	return DB_SUCCESS;
 }
 
 /** Assign a persistent rollback segment in a round-robin fashion,
@@ -882,12 +879,12 @@ static trx_rseg_t* trx_assign_rseg_low()
 		/* By now we have only selected the rseg but not marked it
 		allocated. By marking it allocated we are ensuring that it will
 		never be selected for UNDO truncate purge. */
-		mutex_enter(&rseg->mutex);
+		mysql_mutex_lock(&rseg->mutex);
 		if (!rseg->skip_allocation) {
 			rseg->trx_ref_count++;
 			allocated = true;
 		}
-		mutex_exit(&rseg->mutex);
+		mysql_mutex_unlock(&rseg->mutex);
 	} while (!allocated);
 
 	ut_ad(rseg->trx_ref_count > 0);
@@ -897,10 +894,11 @@ static trx_rseg_t* trx_assign_rseg_low()
 
 /** Assign a rollback segment for modifying temporary tables.
 @return the assigned rollback segment */
-trx_rseg_t *trx_t::assign_temp_rseg()
+trx_rseg_t*
+trx_t::assign_temp_rseg()
 {
 	ut_ad(!rsegs.m_noredo.rseg);
-	ut_ad(!is_autocommit_non_locking());
+	ut_ad(!trx_is_autocommit_non_locking(this));
 	compile_time_assert(ut_is_2pow(TRX_SYS_N_RSEGS));
 
 	/* Choose a temporary rollback segment between 0 and 127
@@ -947,8 +945,8 @@ trx_start_low(
 		    && thd_trx_is_read_only(trx->mysql_thd));
 
 	if (!trx->auto_commit) {
-		trx->will_lock = true;
-	} else if (!trx->will_lock) {
+		++trx->will_lock;
+	} else if (trx->will_lock == 0) {
 		trx->read_only = true;
 	}
 
@@ -986,7 +984,7 @@ trx_start_low(
 
 		trx_sys.register_rw(trx);
 	} else {
-		if (!trx->is_autocommit_non_locking()) {
+		if (!trx_is_autocommit_non_locking(trx)) {
 
 			/* If this is a read-only transaction that is writing
 			to a temporary table then it needs a transaction id
@@ -1019,10 +1017,10 @@ trx_serialise(trx_t* trx)
 {
 	trx_rseg_t *rseg = trx->rsegs.m_redo.rseg;
 	ut_ad(rseg);
-	ut_ad(mutex_own(&rseg->mutex));
+	mysql_mutex_assert_owner(&rseg->mutex);
 
 	if (rseg->last_page_no == FIL_NULL) {
-		mutex_enter(&purge_sys.pq_mutex);
+		mysql_mutex_lock(&purge_sys.pq_mutex);
 	}
 
 	trx_sys.assign_new_trx_no(trx);
@@ -1034,7 +1032,7 @@ trx_serialise(trx_t* trx)
 	if (rseg->last_page_no == FIL_NULL) {
 		purge_sys.purge_queue.push(TrxUndoRsegs(trx->rw_trx_hash_element->no,
 							*rseg));
-		mutex_exit(&purge_sys.pq_mutex);
+		mysql_mutex_unlock(&purge_sys.pq_mutex);
 	}
 }
 
@@ -1067,37 +1065,45 @@ trx_write_serialisation_history(
 		temp_mtr.start();
 		temp_mtr.set_log_mode(MTR_LOG_NO_REDO);
 
-		mutex_enter(&trx->rsegs.m_noredo.rseg->mutex);
+		mysql_mutex_lock(&trx->rsegs.m_noredo.rseg->mutex);
 		trx_undo_set_state_at_finish(undo, &temp_mtr);
-		mutex_exit(&trx->rsegs.m_noredo.rseg->mutex);
+		mysql_mutex_unlock(&trx->rsegs.m_noredo.rseg->mutex);
 		temp_mtr.commit();
 	}
 
 	trx_rseg_t*	rseg = trx->rsegs.m_redo.rseg;
 	if (!rseg) {
 		ut_ad(!trx->rsegs.m_redo.undo);
+		ut_ad(!trx->rsegs.m_redo.old_insert);
 		return;
 	}
 
 	trx_undo_t*& undo = trx->rsegs.m_redo.undo;
+	trx_undo_t*& old_insert = trx->rsegs.m_redo.old_insert;
 
-	if (!undo) {
+	if (!undo && !old_insert) {
 		return;
 	}
 
 	ut_ad(!trx->read_only);
 	ut_ad(!undo || undo->rseg == rseg);
-	mutex_enter(&rseg->mutex);
+	ut_ad(!old_insert || old_insert->rseg == rseg);
+	mysql_mutex_lock(&rseg->mutex);
 
 	/* Assign the transaction serialisation number and add any
 	undo log to the purge queue. */
 	trx_serialise(trx);
+
+	if (UNIV_LIKELY_NULL(old_insert)) {
+		UT_LIST_REMOVE(rseg->old_insert_list, old_insert);
+		trx_purge_add_undo_to_history(trx, old_insert, mtr);
+	}
 	if (undo) {
 		UT_LIST_REMOVE(rseg->undo_list, undo);
 		trx_purge_add_undo_to_history(trx, undo, mtr);
 	}
 
-	mutex_exit(&rseg->mutex);
+	mysql_mutex_unlock(&rseg->mutex);
 
 	MONITOR_INC(MONITOR_TRX_COMMIT_UNDO);
 }
@@ -1160,35 +1166,49 @@ trx_finalize_for_fts(
 	trx->fts_trx = NULL;
 }
 
-/**********************************************************************//**
-If required, flushes the log to disk based on the value of
-innodb_flush_log_at_trx_commit. */
-static
-void
-trx_flush_log_if_needed_low(
-/*========================*/
-	lsn_t	lsn)	/*!< in: lsn up to which logs are to be
-			flushed. */
+extern "C" MYSQL_THD thd_increment_pending_ops();
+extern "C" void thd_decrement_pending_ops(MYSQL_THD);
+
+
+#include "../log/log0sync.h"
+
+/*
+  If required, initiates write and optionally flush of the log to
+  disk
+  @param[in] lsn - lsn up to which logs are to be flushed.
+  @param[in] trx_state - if trx_state is PREPARED, the function will
+  also wait for the flush to complete.
+*/
+static void trx_flush_log_if_needed_low(lsn_t lsn, trx_state_t trx_state)
 {
-	bool	flush = srv_file_flush_method != SRV_NOSYNC;
+  if (!srv_flush_log_at_trx_commit)
+    return;
 
-	switch (srv_flush_log_at_trx_commit) {
-	case 2:
-		/* Write the log but do not flush it to disk */
-		flush = false;
-		/* fall through */
-	case 1:
-	case 3:
-		/* Write the log and optionally flush it to disk */
-		log_write_up_to(lsn, flush);
-		srv_inc_activity_count();
-		return;
-	case 0:
-		/* Do nothing */
-		return;
-	}
+  if (log_sys.get_flushed_lsn() > lsn)
+    return;
 
-	ut_error;
+  bool flush= srv_file_flush_method != SRV_NOSYNC &&
+              srv_flush_log_at_trx_commit == 1;
+
+  if (trx_state == TRX_STATE_PREPARED)
+  {
+    /* XA, which is used with binlog as well.
+    Be conservative, use synchronous wait.*/
+    log_write_up_to(lsn, flush);
+    return;
+  }
+
+  completion_callback cb;
+  if ((cb.m_param = thd_increment_pending_ops()))
+  {
+    cb.m_callback = (void (*)(void *)) thd_decrement_pending_ops;
+    log_write_up_to(lsn, flush, false, &cb);
+  }
+  else
+  {
+    /* No THD, synchronous write */
+    log_write_up_to(lsn, flush);
+  }
 }
 
 /**********************************************************************//**
@@ -1203,43 +1223,25 @@ trx_flush_log_if_needed(
 	trx_t*	trx)	/*!< in/out: transaction */
 {
 	trx->op_info = "flushing log";
-	trx_flush_log_if_needed_low(lsn);
+	trx_flush_log_if_needed_low(lsn,trx->state);
 	trx->op_info = "";
 }
 
-/**********************************************************************//**
-For each table that has been modified by the given transaction: update
-its dict_table_t::update_time with the current timestamp. Clear the list
-of the modified tables at the end. */
-static
-void
-trx_update_mod_tables_timestamp(
-/*============================*/
-	trx_t*	trx)	/*!< in: transaction */
+/** Process tables that were modified by the committing transaction. */
+inline void trx_t::commit_tables()
 {
-	/* consider using trx->start_time if calling time() is too
-	expensive here */
-	const time_t now = time(NULL);
+  if (undo_no && !mod_tables.empty())
+  {
+    const trx_id_t max_trx_id= trx_sys.get_max_trx_id();
+    const auto now= start_time;
 
-	trx_mod_tables_t::const_iterator	end = trx->mod_tables.end();
-
-	for (trx_mod_tables_t::const_iterator it = trx->mod_tables.begin();
-	     it != end;
-	     ++it) {
-
-		/* This could be executed by multiple threads concurrently
-		on the same table object. This is fine because time_t is
-		word size or less. And _purely_ _theoretically_, even if
-		time_t write is not atomic, likely the value of 'now' is
-		the same in all threads and even if it is not, getting a
-		"garbage" in table->update_time is justified because
-		protecting it with a latch here would be too performance
-		intrusive. */
-		dict_table_t* table = it->first;
-		table->update_time = now;
-	}
-
-	trx->mod_tables.clear();
+    for (const auto &p : mod_tables)
+    {
+      dict_table_t *table= p.first;
+      table->update_time= now;
+      table->query_cache_inv_trx_id= max_trx_id;
+    }
+  }
 }
 
 /** Evict a table definition due to the rollback of ALTER TABLE.
@@ -1279,20 +1281,15 @@ inline void trx_t::commit_in_memory(const mtr_t *mtr)
   must_flush_log_later= false;
   read_view.close();
 
-  if (is_autocommit_non_locking())
+  if (trx_is_autocommit_non_locking(this))
   {
     ut_ad(id == 0);
     ut_ad(read_only);
-    ut_ad(!will_lock);
     ut_a(!is_recovered);
     ut_ad(!rsegs.m_redo.rseg);
-    ut_ad(mysql_thd);
-    ut_ad(state == TRX_STATE_ACTIVE);
 
-    /* Note: We are asserting without holding the lock mutex. But
-    that is OK because this transaction is not waiting and cannot
-    be rolled back and no new locks can (or should) be added
-    because it is flagged as a non-locking read-only transaction. */
+    /* Note: We do not have to hold any lock_sys latch here, because
+    this is a non-locking transaction. */
     ut_a(UT_LIST_GET_LEN(lock.trx_locks) == 0);
 
     /* This state change is not protected by any mutex, therefore
@@ -1301,11 +1298,12 @@ inline void trx_t::commit_in_memory(const mtr_t *mtr)
     However, the freezing of trx_sys.trx_list will protect the trx_t
     instance and it cannot be removed from the trx_list and freed
     without first unfreezing trx_list. */
-    state= TRX_STATE_NOT_STARTED;
+    ut_ad(trx_state_eq(this, TRX_STATE_ACTIVE));
 
     MONITOR_INC(MONITOR_TRX_NL_RO_COMMIT);
 
     DBUG_LOG("trx", "Autocommit in memory: " << this);
+    state= TRX_STATE_NOT_STARTED;
   }
   else
   {
@@ -1334,7 +1332,7 @@ inline void trx_t::commit_in_memory(const mtr_t *mtr)
     }
     else
     {
-      trx_update_mod_tables_timestamp(this);
+      commit_tables();
       MONITOR_INC(MONITOR_TRX_RW_COMMIT);
       is_recovered= false;
     }
@@ -1353,17 +1351,34 @@ inline void trx_t::commit_in_memory(const mtr_t *mtr)
   ut_ad(!rsegs.m_redo.undo);
   ut_ad(UT_LIST_GET_LEN(lock.evicted_tables) == 0);
 
+  if (trx_rseg_t *rseg= rsegs.m_redo.rseg)
+  {
+    mysql_mutex_lock(&rseg->mutex);
+    ut_ad(rseg->trx_ref_count > 0);
+    --rseg->trx_ref_count;
+    mysql_mutex_unlock(&rseg->mutex);
+
+    if (trx_undo_t *&insert= rsegs.m_redo.old_insert)
+    {
+      ut_ad(insert->rseg == rseg);
+      trx_undo_commit_cleanup(insert, false);
+      insert= nullptr;
+    }
+  }
+
+  ut_ad(!rsegs.m_redo.old_insert);
+
   if (mtr)
   {
     if (trx_undo_t *&undo= rsegs.m_noredo.undo)
     {
       ut_ad(undo->rseg == rsegs.m_noredo.rseg);
-      trx_undo_commit_cleanup(undo);
+      trx_undo_commit_cleanup(undo, true);
       undo= nullptr;
     }
 
     /* NOTE that we could possibly make a group commit more efficient
-    here: call os_thread_yield here to allow also other trxs to come
+    here: call std::this_thread::yield() here to allow also other trxs to come
     to commit! */
 
     /*-------------------------------------*/
@@ -1401,20 +1416,7 @@ inline void trx_t::commit_in_memory(const mtr_t *mtr)
 
   ut_ad(!rsegs.m_noredo.undo);
 
-  /* Only after trx_undo_commit_cleanup() it is safe to release
-  our rseg reference. */
-  if (trx_rseg_t *rseg= rsegs.m_redo.rseg)
-  {
-    mutex_enter(&rseg->mutex);
-    ut_ad(rseg->trx_ref_count > 0);
-    --rseg->trx_ref_count;
-    mutex_exit(&rseg->mutex);
-  }
-
-  /* Free all savepoints, starting from the first. */
-  trx_named_savept_t *savep= UT_LIST_GET_FIRST(trx_savepoints);
-
-  trx_roll_savepoints_free(this, savep);
+  savepoints_discard();
 
   if (fts_trx)
     trx_finalize_for_fts(this, undo_no != 0);
@@ -1428,17 +1430,19 @@ inline void trx_t::commit_in_memory(const mtr_t *mtr)
     wsrep= false;
     wsrep_commit_ordered(mysql_thd);
   }
-  lock.was_chosen_as_wsrep_victim= false;
+  ut_ad(!(lock.was_chosen_as_deadlock_victim & byte(~2U)));
+  lock.was_chosen_as_deadlock_victim= false;
 #endif /* WITH_WSREP */
-  trx_mutex_enter(this);
+  mutex.wr_lock();
   dict_operation= TRX_DICT_OP_NONE;
 
   DBUG_LOG("trx", "Commit in memory: " << this);
   state= TRX_STATE_NOT_STARTED;
+  mod_tables.clear();
 
   assert_freed();
   trx_init(this);
-  trx_mutex_exit(this);
+  mutex.wr_unlock();
 
   ut_a(error_state == DB_SUCCESS);
   if (!srv_read_only_mode)
@@ -1449,21 +1453,23 @@ inline void trx_t::commit_in_memory(const mtr_t *mtr)
 @param mtr  mini-transaction (if there are any persistent modifications) */
 void trx_t::commit_low(mtr_t *mtr)
 {
+  assert_trx_nonlocking_or_in_list(this);
+  ut_ad(!trx_state_eq(this, TRX_STATE_COMMITTED_IN_MEMORY));
   ut_ad(!mtr || mtr->is_active());
   ut_d(bool aborted = in_rollback && error_state == DB_DEADLOCK);
-  ut_ad(!mtr == (aborted || !has_logged()));
+  ut_ad(!mtr == (aborted || !has_logged_or_recovered()));
   ut_ad(!mtr || !aborted);
 
   /* undo_no is non-zero if we're doing the final commit. */
   if (fts_trx && undo_no)
   {
-    ut_a(!is_autocommit_non_locking());
+    ut_a(!trx_is_autocommit_non_locking(this));
+    dberr_t error= fts_commit(this);
     /* FTS-FIXME: Temporarily tolerate DB_DUPLICATE_KEY instead of
     dying. This is a possible scenario if there is a crash between
     insert to DELETED table committing and transaction committing. The
     fix would be able to return error from this function */
-    if (dberr_t error= fts_commit(this))
-      ut_a(error == DB_DUPLICATE_KEY);
+    ut_a(error == DB_SUCCESS || error == DB_DUPLICATE_KEY);
   }
 
 #ifndef DBUG_OFF
@@ -1505,7 +1511,7 @@ void trx_t::commit()
   mtr_t *mtr= nullptr;
   mtr_t local_mtr;
 
-  if (has_logged())
+  if (has_logged_or_recovered())
   {
     mtr= &local_mtr;
     local_mtr.start();
@@ -1533,19 +1539,7 @@ trx_commit_or_rollback_prepare(
 	case TRX_STATE_ACTIVE:
 	case TRX_STATE_PREPARED:
 	case TRX_STATE_PREPARED_RECOVERED:
-		/* If the trx is in a lock wait state, moves the waiting
-		query thread to the suspended state */
-
-		if (trx->lock.que_state == TRX_QUE_LOCK_WAIT) {
-
-			ut_a(trx->lock.wait_thr != NULL);
-			trx->lock.wait_thr->state = QUE_THR_SUSPENDED;
-			trx->lock.wait_thr = NULL;
-
-			trx->lock.que_state = TRX_QUE_RUNNING;
-		}
-
-		ut_ad(trx->lock.n_active_thrs == 1);
+		trx->lock.wait_thr = NULL;
 		return;
 
 	case TRX_STATE_COMMITTED_IN_MEMORY:
@@ -1598,14 +1592,11 @@ trx_commit_step(
 		trx = thr_get_trx(thr);
 
 		ut_a(trx->lock.wait_thr == NULL);
-		ut_a(trx->lock.que_state != TRX_QUE_LOCK_WAIT);
 
 		trx_commit_or_rollback_prepare(trx);
 
-		trx->lock.que_state = TRX_QUE_COMMITTING;
 		trx->commit();
 		ut_ad(trx->lock.wait_thr == NULL);
-		trx->lock.que_state = TRX_QUE_RUNNING;
 
 		thr = NULL;
 	} else {
@@ -1691,12 +1682,18 @@ trx_mark_sql_stat_end(
 		trx->undo_no = 0;
 		/* fall through */
 	case TRX_STATE_ACTIVE:
-		trx->last_sql_stat_start.least_undo_no = trx->undo_no;
-
 		if (trx->fts_trx != NULL) {
 			fts_savepoint_laststmt_refresh(trx);
 		}
 
+		if (trx->is_bulk_insert()) {
+			/* Allow a subsequent INSERT into an empty table
+			if !unique_checks && !foreign_key_checks. */
+			return;
+		}
+
+		trx->last_sql_stat_start.least_undo_no = trx->undo_no;
+		trx->end_bulk_insert();
 		return;
 	}
 
@@ -1716,15 +1713,20 @@ trx_print_low(
 			/*!< in: max query length to print,
 			or 0 to use the default max length */
 	ulint		n_rec_locks,
-			/*!< in: lock_number_of_rows_locked(&trx->lock) */
+			/*!< in: trx->lock.n_rec_locks */
 	ulint		n_trx_locks,
 			/*!< in: length of trx->lock.trx_locks */
 	ulint		heap_size)
 			/*!< in: mem_heap_get_size(trx->lock.lock_heap) */
 {
 	ibool		newline;
+	const char*	op_info;
 
-	fprintf(f, "TRANSACTION " TRX_ID_FMT, trx_get_id_for_print(trx));
+	if (const trx_id_t id = trx->id) {
+		fprintf(f, "TRANSACTION " TRX_ID_FMT, trx->id);
+	} else {
+		fprintf(f, "TRANSACTION (%p)", trx);
+	}
 
 	switch (trx->state) {
 	case TRX_STATE_NOT_STARTED:
@@ -1746,7 +1748,9 @@ trx_print_low(
 	fprintf(f, ", state %lu", (ulong) trx->state);
 	ut_ad(0);
 state_ok:
-	const char* op_info = trx->op_info;
+
+	/* prevent a race condition */
+	op_info = trx->op_info;
 
 	if (*op_info) {
 		putc(' ', f);
@@ -1767,21 +1771,12 @@ state_ok:
 
 	newline = TRUE;
 
-	/* trx->lock.que_state of an ACTIVE transaction may change
-	while we are not holding trx->mutex. We perform a dirty read
-	for performance reasons. */
-
-	switch (trx->lock.que_state) {
-	case TRX_QUE_RUNNING:
-		newline = FALSE; break;
-	case TRX_QUE_LOCK_WAIT:
-		fputs("LOCK WAIT ", f); break;
-	case TRX_QUE_ROLLING_BACK:
-		fputs("ROLLING BACK ", f); break;
-	case TRX_QUE_COMMITTING:
-		fputs("COMMITTING ", f); break;
-	default:
-		fprintf(f, "que state %lu ", (ulong) trx->lock.que_state);
+	if (trx->in_rollback) { /* dirty read for performance reasons */
+		fputs("ROLLING BACK ", f);
+	} else if (trx->lock.wait_lock) {
+		fputs("LOCK WAIT ", f);
+	} else {
+		newline = FALSE;
 	}
 
 	if (n_trx_locks > 0 || heap_size > 400) {
@@ -1811,7 +1806,7 @@ state_ok:
 
 /**********************************************************************//**
 Prints info about a transaction.
-The caller must hold lock_sys.mutex.
+The caller must hold lock_sys.latch.
 When possible, use trx_print() instead. */
 void
 trx_print_latched(
@@ -1821,17 +1816,17 @@ trx_print_latched(
 	ulint		max_query_len)	/*!< in: max query length to print,
 					or 0 to use the default max length */
 {
-	ut_ad(lock_mutex_own());
+	lock_sys.assert_locked();
 
 	trx_print_low(f, trx, max_query_len,
-		      lock_number_of_rows_locked(&trx->lock),
+		      trx->lock.n_rec_locks,
 		      UT_LIST_GET_LEN(trx->lock.trx_locks),
 		      mem_heap_get_size(trx->lock.lock_heap));
 }
 
 /**********************************************************************//**
 Prints info about a transaction.
-Acquires and releases lock_sys.mutex. */
+Acquires and releases lock_sys.latch. */
 void
 trx_print(
 /*======*/
@@ -1840,60 +1835,25 @@ trx_print(
 	ulint		max_query_len)	/*!< in: max query length to print,
 					or 0 to use the default max length */
 {
-	ulint	n_rec_locks;
-	ulint	n_trx_locks;
-	ulint	heap_size;
+  ulint n_rec_locks, n_trx_locks, heap_size;
+  {
+    LockMutexGuard g{SRW_LOCK_CALL};
+    n_rec_locks= trx->lock.n_rec_locks;
+    n_trx_locks= UT_LIST_GET_LEN(trx->lock.trx_locks);
+    heap_size= mem_heap_get_size(trx->lock.lock_heap);
+  }
 
-	lock_mutex_enter();
-	n_rec_locks = lock_number_of_rows_locked(&trx->lock);
-	n_trx_locks = UT_LIST_GET_LEN(trx->lock.trx_locks);
-	heap_size = mem_heap_get_size(trx->lock.lock_heap);
-	lock_mutex_exit();
-
-	trx_print_low(f, trx, max_query_len,
-		      n_rec_locks, n_trx_locks, heap_size);
-}
-
-/*******************************************************************//**
-Compares the "weight" (or size) of two transactions. Transactions that
-have edited non-transactional tables are considered heavier than ones
-that have not.
-@return TRUE if weight(a) >= weight(b) */
-bool
-trx_weight_ge(
-/*==========*/
-	const trx_t*	a,	/*!< in: transaction to be compared */
-	const trx_t*	b)	/*!< in: transaction to be compared */
-{
-	ibool	a_notrans_edit;
-	ibool	b_notrans_edit;
-
-	/* If mysql_thd is NULL for a transaction we assume that it has
-	not edited non-transactional tables. */
-
-	a_notrans_edit = a->mysql_thd != NULL
-		&& thd_has_edited_nontrans_tables(a->mysql_thd);
-
-	b_notrans_edit = b->mysql_thd != NULL
-		&& thd_has_edited_nontrans_tables(b->mysql_thd);
-
-	if (a_notrans_edit != b_notrans_edit) {
-
-		return(a_notrans_edit);
-	}
-
-	/* Either both had edited non-transactional tables or both had
-	not, we fall back to comparing the number of altered/locked
-	rows. */
-
-	return(TRX_WEIGHT(a) >= TRX_WEIGHT(b));
+  trx_print_low(f, trx, max_query_len, n_rec_locks, n_trx_locks, heap_size);
 }
 
 /** Prepare a transaction.
 @return	log sequence number that makes the XA PREPARE durable
 @retval	0	if no changes needed to be made durable */
-static lsn_t trx_prepare_low(trx_t *trx)
+static
+lsn_t
+trx_prepare_low(trx_t* trx)
 {
+	ut_ad(!trx->rsegs.m_redo.old_insert);
 	ut_ad(!trx->is_recovered);
 
 	mtr_t	mtr;
@@ -1904,9 +1864,9 @@ static lsn_t trx_prepare_low(trx_t *trx)
 		mtr.start();
 		mtr.set_log_mode(MTR_LOG_NO_REDO);
 
-		mutex_enter(&undo->rseg->mutex);
+		mysql_mutex_lock(&undo->rseg->mutex);
 		trx_undo_set_state_at_prepare(trx, undo, false, &mtr);
-		mutex_exit(&undo->rseg->mutex);
+		mysql_mutex_unlock(&undo->rseg->mutex);
 
 		mtr.commit();
 	}
@@ -1928,9 +1888,9 @@ static lsn_t trx_prepare_low(trx_t *trx)
 	structure define the transaction as prepared in the file-based
 	world, at the serialization point of lsn. */
 
-	mutex_enter(&rseg->mutex);
+	mysql_mutex_lock(&rseg->mutex);
 	trx_undo_set_state_at_prepare(trx, undo, false, &mtr);
-	mutex_exit(&rseg->mutex);
+	mysql_mutex_unlock(&rseg->mutex);
 
 	/* Make the XA PREPARE durable. */
 	mtr.commit();
@@ -1955,9 +1915,9 @@ trx_prepare(
 	DBUG_EXECUTE_IF("ib_trx_crash_during_xa_prepare_step", DBUG_SUICIDE(););
 
 	ut_a(trx->state == TRX_STATE_ACTIVE);
-	trx_mutex_enter(trx);
+	trx->mutex_lock();
 	trx->state = TRX_STATE_PREPARED;
-	trx_mutex_exit(trx);
+	trx->mutex_unlock();
 
 	if (lsn) {
 		/* Depending on the my.cnf options, we may now write the log
@@ -1977,20 +1937,6 @@ trx_prepare(
 		We must not be holding any mutexes or latches here. */
 
 		trx_flush_log_if_needed(lsn, trx);
-
-		if (!UT_LIST_GET_LEN(trx->lock.trx_locks)
-		    || trx->isolation_level == TRX_ISO_SERIALIZABLE) {
-			/* Do not release any locks at the
-			SERIALIZABLE isolation level. */
-		} else if (!trx->mysql_thd
-			   || thd_sql_command(trx->mysql_thd)
-			   != SQLCOM_XA_PREPARE) {
-			/* Do not release locks for XA COMMIT ONE PHASE
-			or for internal distributed transactions
-			(XID::get_my_xid() would be nonzero). */
-		} else {
-			lock_release_on_prepare(trx);
-		}
 	}
 }
 
@@ -2020,7 +1966,7 @@ static my_bool trx_recover_for_mysql_callback(rw_trx_hash_element_t *element,
   trx_recover_for_mysql_callback_arg *arg)
 {
   DBUG_ASSERT(arg->len > 0);
-  mutex_enter(&element->mutex);
+  mysql_mutex_lock(&element->mutex);
   if (trx_t *trx= element->trx)
   {
     /*
@@ -2046,7 +1992,7 @@ static my_bool trx_recover_for_mysql_callback(rw_trx_hash_element_t *element,
       }
     }
   }
-  mutex_exit(&element->mutex);
+  mysql_mutex_unlock(&element->mutex);
   /* Do not terminate upon reaching arg->len; count all transactions */
   return false;
 }
@@ -2055,13 +2001,13 @@ static my_bool trx_recover_for_mysql_callback(rw_trx_hash_element_t *element,
 static my_bool trx_recover_reset_callback(rw_trx_hash_element_t *element,
   void*)
 {
-  mutex_enter(&element->mutex);
+  mysql_mutex_lock(&element->mutex);
   if (trx_t *trx= element->trx)
   {
     if (trx_state_eq(trx, TRX_STATE_PREPARED_RECOVERED))
       trx->state= TRX_STATE_PREPARED;
   }
-  mutex_exit(&element->mutex);
+  mysql_mutex_unlock(&element->mutex);
   return false;
 }
 
@@ -2110,10 +2056,10 @@ static my_bool trx_get_trx_by_xid_callback(rw_trx_hash_element_t *element,
   trx_get_trx_by_xid_callback_arg *arg)
 {
   my_bool found= 0;
-  mutex_enter(&element->mutex);
+  mysql_mutex_lock(&element->mutex);
   if (trx_t *trx= element->trx)
   {
-    trx_mutex_enter(trx);
+    trx->mutex_lock();
     if (trx->is_recovered &&
 	(trx_state_eq(trx, TRX_STATE_PREPARED) ||
 	 trx_state_eq(trx, TRX_STATE_PREPARED_RECOVERED)) &&
@@ -2130,9 +2076,9 @@ static my_bool trx_get_trx_by_xid_callback(rw_trx_hash_element_t *element,
       arg->trx= trx;
       found= 1;
     }
-    trx_mutex_exit(trx);
+    trx->mutex_unlock();
   }
-  mutex_exit(&element->mutex);
+  mysql_mutex_unlock(&element->mutex);
   return found;
 }
 
@@ -2223,7 +2169,7 @@ trx_start_internal_low(
 	/* Ensure it is not flagged as an auto-commit-non-locking
 	transaction. */
 
-	trx->will_lock = true;
+	trx->will_lock = 1;
 
 	trx->internal = true;
 
@@ -2239,7 +2185,7 @@ trx_start_internal_read_only_low(
 	/* Ensure it is not flagged as an auto-commit-non-locking
 	transaction. */
 
-	trx->will_lock = true;
+	trx->will_lock = 1;
 
 	trx->internal = true;
 
@@ -2260,7 +2206,13 @@ trx_start_for_ddl_low(
 		the data dictionary will be locked in crash recovery. */
 
 		trx_set_dict_operation(trx, op);
+
+		/* Ensure it is not flagged as an auto-commit-non-locking
+		transation. */
+		trx->will_lock = 1;
+
 		trx->ddl= true;
+
 		trx_start_internal_low(trx);
 		return;
 
@@ -2287,7 +2239,7 @@ trx_set_rw_mode(
 	trx_t*		trx)		/*!< in/out: transaction that is RW */
 {
 	ut_ad(trx->rsegs.m_redo.rseg == 0);
-	ut_ad(!trx->is_autocommit_non_locking());
+	ut_ad(!trx_is_autocommit_non_locking(trx));
 	ut_ad(!trx->read_only);
 	ut_ad(trx->id == 0);
 
@@ -2305,16 +2257,3 @@ trx_set_rw_mode(
 		trx->read_view.set_creator_trx_id(trx->id);
 	}
 }
-
-bool trx_t::has_stats_table_lock() const
-{
-  for (lock_list::const_iterator it= lock.table_locks.begin(),
-       end= lock.table_locks.end(); it != end; ++it)
-  {
-     const lock_t *lock= *it;
-     if (lock && lock->un_member.tab_lock.table->is_stats_table())
-       return true;
-  }
-
-  return false;
-}
