@@ -34,7 +34,6 @@ Created 3/26/1996 Heikki Tuuri
 #include "trx0purge.h"
 #include "trx0rec.h"
 #include "trx0rseg.h"
-#include "log.h"
 
 /* How should the old versions in the history list be managed?
    ----------------------------------------------------------
@@ -176,7 +175,6 @@ trx_undo_get_prev_rec_from_prev_page(buf_block_t *&block, uint16_t rec,
 
   block= buf_page_get(page_id_t(block->page.id().space(), prev_page_no),
                       0, shared ? RW_S_LATCH : RW_X_LATCH, mtr);
-  buf_block_dbg_add_level(block, SYNC_TRX_UNDO_PAGE);
 
   return trx_undo_page_get_last_rec(block, page_no, offset);
 }
@@ -244,7 +242,6 @@ trx_undo_get_next_rec_from_next_page(buf_block_t *&block, uint32_t page_no,
     return NULL;
 
   block= buf_page_get(page_id_t(block->page.id().space(), next), 0, mode, mtr);
-  buf_block_dbg_add_level(block, SYNC_TRX_UNDO_PAGE);
 
   return trx_undo_page_get_first_rec(block, page_no, offset);
 }
@@ -282,7 +279,6 @@ trx_undo_get_first_rec(const fil_space_t &space, uint32_t page_no,
                        mtr_t *mtr)
 {
   block = buf_page_get(page_id_t(space.id, page_no), 0, mode, mtr);
-  buf_block_dbg_add_level(block, SYNC_TRX_UNDO_PAGE);
 
   if (trx_undo_rec_t *rec= trx_undo_page_get_first_rec(block, page_no, offset))
     return rec;
@@ -394,8 +390,6 @@ trx_undo_seg_create(fil_space_t *space, buf_block_t *rseg_hdr, ulint *id,
 		return NULL;
 	}
 
-	buf_block_dbg_add_level(block, SYNC_TRX_UNDO_PAGE);
-
 	mtr->undo_create(*block);
 	trx_undo_page_init(*block);
 
@@ -435,7 +429,7 @@ static uint16_t trx_undo_header_create(buf_block_t *undo_page, trx_id_t trx_id,
   repurposed after upgrading to MariaDB 10.3. */
   byte *undo_type= my_assume_aligned<2>
     (TRX_UNDO_PAGE_HDR + TRX_UNDO_PAGE_TYPE + undo_page->frame);
-  ut_ad(mach_read_from_2(undo_type) <= 2);
+  ut_ad(mach_read_from_2(undo_type) <= TRX_UNDO_UPDATE);
   mtr->write<2,mtr_t::MAYBE_NOP>(*undo_page, undo_type, 0U);
   byte *start= my_assume_aligned<4>(TRX_UNDO_PAGE_HDR + TRX_UNDO_PAGE_START +
                                     undo_page->frame);
@@ -559,7 +553,7 @@ buf_block_t* trx_undo_add_page(trx_undo_t* undo, mtr_t* mtr)
 	a pessimistic insert in a B-tree, and we must reserve the
 	counterpart of the tree latch, which is the rseg mutex. */
 
-	mutex_enter(&rseg->mutex);
+	mysql_mutex_lock(&rseg->mutex);
 
 	buf_block_t* header_block = trx_undo_page_get(
 		page_id_t(undo->rseg->space->id, undo->hdr_page_no), mtr);
@@ -580,8 +574,6 @@ buf_block_t* trx_undo_add_page(trx_undo_t* undo, mtr_t* mtr)
 		goto func_exit;
 	}
 
-	ut_ad(rw_lock_get_x_lock_count(&new_block->lock) == 1);
-	buf_block_dbg_add_level(new_block, SYNC_TRX_UNDO_PAGE);
 	undo->last_page_no = new_block->page.id().page_no();
 
 	mtr->undo_create(*new_block);
@@ -593,7 +585,7 @@ buf_block_t* trx_undo_add_page(trx_undo_t* undo, mtr_t* mtr)
 	rseg->curr_size++;
 
 func_exit:
-	mutex_exit(&rseg->mutex);
+	mysql_mutex_unlock(&rseg->mutex);
 	return(new_block);
 }
 
@@ -617,7 +609,7 @@ trx_undo_free_page(
 	const ulint	space = rseg->space->id;
 
 	ut_a(hdr_page_no != page_no);
-	ut_ad(mutex_own(&(rseg->mutex)));
+	mysql_mutex_assert_owner(&rseg->mutex);
 
 	buf_block_t* undo_block = trx_undo_page_get(page_id_t(space, page_no),
 						    mtr);
@@ -631,7 +623,7 @@ trx_undo_free_page(
 	fseg_free_page(TRX_UNDO_SEG_HDR + TRX_UNDO_FSEG_HEADER
 		       + header_block->frame,
 		       rseg->space, page_no, mtr);
-	buf_page_free(rseg->space, page_no, mtr, __FILE__, __LINE__);
+	buf_page_free(rseg->space, page_no, mtr);
 
 	const fil_addr_t last_addr = flst_get_last(
 		TRX_UNDO_SEG_HDR + TRX_UNDO_PAGE_LIST + header_block->frame);
@@ -682,7 +674,7 @@ void trx_undo_truncate_end(trx_undo_t& undo, undo_no_t limit, bool is_temp)
 		}
 
 		trx_undo_rec_t* trunc_here = NULL;
-		mutex_enter(&undo.rseg->mutex);
+		mysql_mutex_lock(&undo.rseg->mutex);
 		buf_block_t* undo_block = trx_undo_page_get(
 			page_id_t(undo.rseg->space->id, undo.last_page_no),
 			&mtr);
@@ -702,13 +694,13 @@ void trx_undo_truncate_end(trx_undo_t& undo, undo_no_t limit, bool is_temp)
 
 		if (undo.last_page_no != undo.hdr_page_no) {
 			trx_undo_free_last_page(&undo, &mtr);
-			mutex_exit(&undo.rseg->mutex);
+			mysql_mutex_unlock(&undo.rseg->mutex);
 			mtr.commit();
 			continue;
 		}
 
 func_exit:
-		mutex_exit(&undo.rseg->mutex);
+		mysql_mutex_unlock(&undo.rseg->mutex);
 
 		if (trunc_here) {
 			mtr.write<2>(*undo_block,
@@ -741,7 +733,7 @@ trx_undo_truncate_start(
 	trx_undo_rec_t* last_rec;
 	mtr_t		mtr;
 
-	ut_ad(mutex_own(&(rseg->mutex)));
+	mysql_mutex_assert_owner(&rseg->mutex);
 
 	if (!limit) {
 		return;
@@ -791,22 +783,27 @@ done:
 }
 
 /** Frees an undo log segment which is not in the history list.
-@param undo	temporary undo log */
-static void trx_undo_seg_free(const trx_undo_t *undo)
+@param[in]	undo	undo log
+@param[in]	noredo	whether the undo tablespace is redo logged */
+static void trx_undo_seg_free(const trx_undo_t* undo, bool noredo)
 {
 	ut_ad(undo->id < TRX_RSEG_N_SLOTS);
 
 	trx_rseg_t* const rseg = undo->rseg;
 	bool		finished;
 	mtr_t		mtr;
-	ut_ad(rseg->space == fil_system.temp_space);
 
 	do {
 		mtr.start();
-		mtr.set_log_mode(MTR_LOG_NO_REDO);
+
+		if (noredo) {
+			mtr.set_log_mode(MTR_LOG_NO_REDO);
+		}
+
+		mysql_mutex_lock(&rseg->mutex);
 
 		buf_block_t* block = trx_undo_page_get(
-			page_id_t(SRV_TMP_SPACE_ID, undo->hdr_page_no), &mtr);
+			page_id_t(rseg->space->id, undo->hdr_page_no), &mtr);
 
 		fseg_header_t* file_seg = TRX_UNDO_SEG_HDR
 			+ TRX_UNDO_FSEG_HEADER + block->frame;
@@ -818,12 +815,12 @@ static void trx_undo_seg_free(const trx_undo_t *undo)
 			buf_block_t* rseg_header = trx_rsegf_get(
 				rseg->space, rseg->page_no, &mtr);
 			compile_time_assert(FIL_NULL == 0xffffffff);
-			memset(TRX_RSEG + TRX_RSEG_UNDO_SLOTS
-			       + undo->id * TRX_RSEG_SLOT_SIZE +
-			       rseg_header->frame, 0xff, 4);
+			mtr.memset(rseg_header, TRX_RSEG + TRX_RSEG_UNDO_SLOTS
+				   + undo->id * TRX_RSEG_SLOT_SIZE, 4, 0xff);
 			MONITOR_DEC(MONITOR_NUM_UNDO_SLOT_USED);
 		}
 
+		mysql_mutex_unlock(&rseg->mutex);
 		mtr.commit();
 	} while (!finished);
 }
@@ -835,9 +832,8 @@ static void trx_undo_seg_free(const trx_undo_t *undo)
 @param[in]	id		rollback segment slot
 @param[in]	page_no		undo log segment page number
 @param[in,out]	max_trx_id	the largest observed transaction ID
-@return	the undo log
-@retval nullptr on error */
-trx_undo_t *
+@return	size of the undo log in pages */
+uint32_t
 trx_undo_mem_create_at_db_start(trx_rseg_t *rseg, ulint id, uint32_t page_no,
                                 trx_id_t &max_trx_id)
 {
@@ -849,58 +845,16 @@ trx_undo_mem_create_at_db_start(trx_rseg_t *rseg, ulint id, uint32_t page_no,
 	mtr.start();
 	const buf_block_t* block = trx_undo_page_get(
 		page_id_t(rseg->space->id, page_no), &mtr);
-	const uint16_t type = mach_read_from_2(TRX_UNDO_PAGE_HDR
-					       + TRX_UNDO_PAGE_TYPE
-					       + block->frame);
-	if (UNIV_UNLIKELY(type > 2)) {
-corrupted_type:
-		sql_print_error("InnoDB: unsupported undo header type %u",
-				type);
-corrupted:
-		mtr.commit();
-		return nullptr;
-	}
+	const ulint type = mach_read_from_2(
+		TRX_UNDO_PAGE_HDR + TRX_UNDO_PAGE_TYPE + block->frame);
+	ut_ad(type == 0 || type == TRX_UNDO_INSERT || type == TRX_UNDO_UPDATE);
 
-	uint16_t offset = mach_read_from_2(TRX_UNDO_SEG_HDR + TRX_UNDO_LAST_LOG
-					   + block->frame);
-	if (offset < TRX_UNDO_SEG_HDR + TRX_UNDO_SEG_HDR_SIZE ||
-	    offset >= srv_page_size - TRX_UNDO_LOG_OLD_HDR_SIZE) {
-		sql_print_error("InnoDB: invalid undo header offset %u",
-				offset);
-		goto corrupted;
-	}
-
-	const trx_ulogf_t* const undo_header = block->frame + offset;
 	uint16_t state = mach_read_from_2(TRX_UNDO_SEG_HDR + TRX_UNDO_STATE
 					  + block->frame);
-	switch (state) {
-	case TRX_UNDO_ACTIVE:
-	case TRX_UNDO_PREPARED:
-		if (UNIV_LIKELY(type != 1)) {
-			break;
-		}
-		sql_print_error("InnoDB: upgrade from older version than"
-				" MariaDB 10.3 requires clean shutdown");
-		goto corrupted;
-	default:
-		sql_print_error("InnoDB: unsupported undo header state %u",
-				state);
-		goto corrupted;
-	case TRX_UNDO_TO_PURGE:
-		if (UNIV_UNLIKELY(type == 1)) {
-			goto corrupted_type;
-		}
-		/* fall through */
-	case TRX_UNDO_CACHED:
-		trx_id_t id = mach_read_from_8(TRX_UNDO_TRX_NO + undo_header);
-		if (id >> 48) {
-			sql_print_error("InnoDB: corrupted TRX_NO %llx", id);
-			goto corrupted;
-		}
-		if (id > max_trx_id) {
-			max_trx_id = id;
-		}
-	}
+	uint16_t offset = mach_read_from_2(TRX_UNDO_SEG_HDR + TRX_UNDO_LAST_LOG
+					   + block->frame);
+
+	const trx_ulogf_t*	undo_header = block->frame + offset;
 
 	/* Read X/Open XA transaction identification if it exists, or
 	set it to NULL. */
@@ -912,57 +866,68 @@ corrupted:
 	}
 
 	trx_id_t trx_id = mach_read_from_8(undo_header + TRX_UNDO_TRX_ID);
-	if (trx_id >> 48) {
-		sql_print_error("InnoDB: corrupted TRX_ID %llx", trx_id);
-		goto corrupted;
-	}
 	if (trx_id > max_trx_id) {
 		max_trx_id = trx_id;
 	}
 
-	mutex_enter(&rseg->mutex);
+	mysql_mutex_lock(&rseg->mutex);
 	trx_undo_t* undo = trx_undo_mem_create(
 		rseg, id, trx_id, &xid, page_no, offset);
-	mutex_exit(&rseg->mutex);
-	if (!undo) {
-		return undo;
-	}
+	mysql_mutex_unlock(&rseg->mutex);
 
 	undo->dict_operation = undo_header[TRX_UNDO_DICT_TRANS];
 	undo->table_id = mach_read_from_8(undo_header + TRX_UNDO_TABLE_ID);
 	undo->size = flst_get_len(TRX_UNDO_SEG_HDR + TRX_UNDO_PAGE_LIST
 				  + block->frame);
 
-	fil_addr_t	last_addr = flst_get_last(
-		TRX_UNDO_SEG_HDR + TRX_UNDO_PAGE_LIST + block->frame);
-
-	undo->last_page_no = last_addr.page;
-	undo->top_page_no = last_addr.page;
-
-	const buf_block_t* last = trx_undo_page_get(
-		page_id_t(rseg->space->id, undo->last_page_no), &mtr);
-
-	if (const trx_undo_rec_t* rec = trx_undo_page_get_last_rec(
-		    last, page_no, offset)) {
-		undo->top_offset = static_cast<uint16_t>(rec - last->frame);
-		undo->top_undo_no = trx_undo_rec_get_undo_no(rec);
-		ut_ad(!undo->empty());
+	if (UNIV_UNLIKELY(state == TRX_UNDO_TO_FREE)) {
+		/* This is an old-format insert_undo log segment that
+		is being freed. The page list is inconsistent. */
+		ut_ad(type == TRX_UNDO_INSERT);
+		state = TRX_UNDO_TO_PURGE;
 	} else {
-		undo->top_undo_no = IB_ID_MAX;
-		ut_ad(undo->empty());
+		if (state == TRX_UNDO_TO_PURGE
+		    || state == TRX_UNDO_CACHED) {
+			trx_id_t id = mach_read_from_8(TRX_UNDO_TRX_NO
+						       + undo_header);
+			if (id > max_trx_id) {
+				max_trx_id = id;
+			}
+		}
+
+		fil_addr_t	last_addr = flst_get_last(
+			TRX_UNDO_SEG_HDR + TRX_UNDO_PAGE_LIST + block->frame);
+
+		undo->last_page_no = last_addr.page;
+		undo->top_page_no = last_addr.page;
+
+		const buf_block_t* last = trx_undo_page_get(
+			page_id_t(rseg->space->id, undo->last_page_no), &mtr);
+
+		if (const trx_undo_rec_t* rec = trx_undo_page_get_last_rec(
+			    last, page_no, offset)) {
+			undo->top_offset = uint16_t(rec - last->frame);
+			undo->top_undo_no = trx_undo_rec_get_undo_no(rec);
+			ut_ad(!undo->empty());
+		} else {
+			undo->top_undo_no = IB_ID_MAX;
+			ut_ad(undo->empty());
+		}
 	}
 
 	undo->state = state;
 
 	if (state != TRX_UNDO_CACHED) {
-		UT_LIST_ADD_LAST(rseg->undo_list, undo);
+		UT_LIST_ADD_LAST(type == TRX_UNDO_INSERT
+				 ? rseg->old_insert_list
+				 : rseg->undo_list, undo);
 	} else {
 		UT_LIST_ADD_LAST(rseg->undo_cached, undo);
 		MONITOR_INC(MONITOR_NUM_UNDO_SLOT_CACHED);
 	}
 
 	mtr.commit();
-	return undo;
+	return undo->size;
 }
 
 /********************************************************************//**
@@ -982,7 +947,7 @@ trx_undo_mem_create(
 {
 	trx_undo_t*	undo;
 
-	ut_ad(mutex_own(&(rseg->mutex)));
+	mysql_mutex_assert_owner(&rseg->mutex);
 
 	ut_a(id < TRX_RSEG_N_SLOTS);
 
@@ -1027,7 +992,7 @@ trx_undo_mem_init_for_reuse(
 	const XID*	xid,	/*!< in: X/Open XA transaction identification*/
 	uint16_t	offset)	/*!< in: undo log header byte offset on page */
 {
-	ut_ad(mutex_own(&((undo->rseg)->mutex)));
+	mysql_mutex_assert_owner(&undo->rseg->mutex);
 
 	ut_a(undo->id < TRX_RSEG_N_SLOTS);
 
@@ -1057,7 +1022,7 @@ trx_undo_create(trx_t* trx, trx_rseg_t* rseg, trx_undo_t** undo,
 {
 	ulint		id;
 
-	ut_ad(mutex_own(&(rseg->mutex)));
+	mysql_mutex_assert_owner(&rseg->mutex);
 
 	buf_block_t*	block = trx_undo_seg_create(
 		rseg->space,
@@ -1116,7 +1081,7 @@ buf_block_t*
 trx_undo_reuse_cached(trx_t* trx, trx_rseg_t* rseg, trx_undo_t** pundo,
 		      mtr_t* mtr)
 {
-	ut_ad(mutex_own(&rseg->mutex));
+	mysql_mutex_assert_owner(&rseg->mutex);
 
 	trx_undo_t* undo = UT_LIST_GET_FIRST(rseg->undo_cached);
 	if (!undo) {
@@ -1133,8 +1098,6 @@ trx_undo_reuse_cached(trx_t* trx, trx_rseg_t* rseg, trx_undo_t** pundo,
 		return NULL;
 	}
 
-	buf_block_dbg_add_level(block, SYNC_TRX_UNDO_PAGE);
-
 	UT_LIST_REMOVE(rseg->undo_cached, undo);
 	MONITOR_DEC(MONITOR_NUM_UNDO_SLOT_CACHED);
 
@@ -1186,12 +1149,12 @@ trx_undo_assign(trx_t* trx, dberr_t* err, mtr_t* mtr)
 		return buf_page_get_gen(
 			page_id_t(undo->rseg->space->id, undo->last_page_no),
 			0, RW_X_LATCH, undo->guess_block,
-			BUF_GET, __FILE__, __LINE__, mtr, err);
+			BUF_GET, mtr, err);
 	}
 
 	trx_rseg_t* rseg = trx->rsegs.m_redo.rseg;
 
-	mutex_enter(&rseg->mutex);
+	mysql_mutex_lock(&rseg->mutex);
 	buf_block_t* block = trx_undo_reuse_cached(
 		trx, rseg, &trx->rsegs.m_redo.undo, mtr);
 
@@ -1209,7 +1172,7 @@ trx_undo_assign(trx_t* trx, dberr_t* err, mtr_t* mtr)
 	UT_LIST_ADD_FIRST(rseg->undo_list, trx->rsegs.m_redo.undo);
 
 func_exit:
-	mutex_exit(&rseg->mutex);
+	mysql_mutex_unlock(&rseg->mutex);
 	return block;
 }
 
@@ -1226,8 +1189,7 @@ buf_block_t*
 trx_undo_assign_low(trx_t* trx, trx_rseg_t* rseg, trx_undo_t** undo,
 		    dberr_t* err, mtr_t* mtr)
 {
-  const bool	is_temp __attribute__((unused)) = rseg == trx->rsegs.m_noredo.rseg;
-
+	ut_d(const bool	is_temp = rseg == trx->rsegs.m_noredo.rseg);
 	ut_ad(rseg == trx->rsegs.m_redo.rseg
 	      || rseg == trx->rsegs.m_noredo.rseg);
 	ut_ad(undo == (is_temp
@@ -1240,7 +1202,7 @@ trx_undo_assign_low(trx_t* trx, trx_rseg_t* rseg, trx_undo_t** undo,
 		return buf_page_get_gen(
 			page_id_t(rseg->space->id, (*undo)->last_page_no),
 			0, RW_X_LATCH, (*undo)->guess_block,
-			BUF_GET, __FILE__, __LINE__, mtr, err);
+			BUF_GET, mtr, err);
 	}
 
 	DBUG_EXECUTE_IF(
@@ -1248,7 +1210,7 @@ trx_undo_assign_low(trx_t* trx, trx_rseg_t* rseg, trx_undo_t** undo,
 		*err = DB_TOO_MANY_CONCURRENT_TRXS; return NULL;
 	);
 
-	mutex_enter(&rseg->mutex);
+	mysql_mutex_lock(&rseg->mutex);
 
 	buf_block_t* block = trx_undo_reuse_cached(trx, rseg, undo, mtr);
 
@@ -1265,7 +1227,7 @@ trx_undo_assign_low(trx_t* trx, trx_rseg_t* rseg, trx_undo_t** undo,
 	UT_LIST_ADD_FIRST(rseg->undo_list, *undo);
 
 func_exit:
-	mutex_exit(&rseg->mutex);
+	mysql_mutex_unlock(&rseg->mutex);
 	return block;
 }
 
@@ -1332,18 +1294,22 @@ void trx_undo_set_state_at_prepare(trx_t *trx, trx_undo_t *undo, bool rollback,
 	trx_undo_write_xid(block, offset, undo->xid, mtr);
 }
 
-/** Free temporary undo log after commit or rollback.
+/** Free an old insert or temporary undo log after commit or rollback.
 The information is not needed after a commit or rollback, therefore
 the data can be discarded.
-@param undo     temporary undo log */
-void trx_undo_commit_cleanup(trx_undo_t *undo)
+@param[in,out]	undo	undo log
+@param[in]	is_temp	whether this is temporary undo log */
+void
+trx_undo_commit_cleanup(trx_undo_t* undo, bool is_temp)
 {
 	trx_rseg_t*	rseg	= undo->rseg;
-	ut_ad(rseg->space == fil_system.temp_space);
+	ut_ad(is_temp == !rseg->is_persistent());
+	ut_ad(!is_temp || 0 == UT_LIST_GET_LEN(rseg->old_insert_list));
 
-	mutex_enter(&rseg->mutex);
+	mysql_mutex_lock(&rseg->mutex);
 
-	UT_LIST_REMOVE(rseg->undo_list, undo);
+	UT_LIST_REMOVE(is_temp ? rseg->undo_list : rseg->old_insert_list,
+		       undo);
 
 	if (undo->state == TRX_UNDO_CACHED) {
 		UT_LIST_ADD_FIRST(rseg->undo_cached, undo);
@@ -1352,7 +1318,9 @@ void trx_undo_commit_cleanup(trx_undo_t *undo)
 		ut_ad(undo->state == TRX_UNDO_TO_PURGE);
 
 		/* Delete first the undo log segment in the file */
-		trx_undo_seg_free(undo);
+		mysql_mutex_unlock(&rseg->mutex);
+		trx_undo_seg_free(undo, is_temp);
+		mysql_mutex_lock(&rseg->mutex);
 
 		ut_ad(rseg->curr_size > undo->size);
 		rseg->curr_size -= undo->size;
@@ -1360,17 +1328,19 @@ void trx_undo_commit_cleanup(trx_undo_t *undo)
 		ut_free(undo);
 	}
 
-	mutex_exit(&rseg->mutex);
+	mysql_mutex_unlock(&rseg->mutex);
 }
 
 /** At shutdown, frees the undo logs of a transaction. */
-void trx_undo_free_at_shutdown(trx_t *trx)
+void
+trx_undo_free_at_shutdown(trx_t *trx)
 {
 	if (trx_undo_t*& undo = trx->rsegs.m_redo.undo) {
 		switch (undo->state) {
 		case TRX_UNDO_PREPARED:
 			break;
 		case TRX_UNDO_CACHED:
+		case TRX_UNDO_TO_FREE:
 		case TRX_UNDO_TO_PURGE:
 			ut_ad(trx_state_eq(trx,
 					   TRX_STATE_COMMITTED_IN_MEMORY));
@@ -1391,6 +1361,34 @@ void trx_undo_free_at_shutdown(trx_t *trx)
 		ut_free(undo);
 		undo = NULL;
 	}
+
+	if (trx_undo_t*& undo = trx->rsegs.m_redo.old_insert) {
+		switch (undo->state) {
+		case TRX_UNDO_PREPARED:
+			break;
+		case TRX_UNDO_CACHED:
+		case TRX_UNDO_TO_FREE:
+		case TRX_UNDO_TO_PURGE:
+			ut_ad(trx_state_eq(trx,
+					   TRX_STATE_COMMITTED_IN_MEMORY));
+			/* fall through */
+		case TRX_UNDO_ACTIVE:
+			/* trx_t::commit_state() assigns
+			trx->state = TRX_STATE_COMMITTED_IN_MEMORY. */
+			ut_a(!srv_was_started
+			     || srv_read_only_mode
+			     || srv_force_recovery >= SRV_FORCE_NO_TRX_UNDO
+			     || srv_fast_shutdown);
+			break;
+		default:
+			ut_error;
+		}
+
+		UT_LIST_REMOVE(trx->rsegs.m_redo.rseg->old_insert_list, undo);
+		ut_free(undo);
+		undo = NULL;
+	}
+
 	if (trx_undo_t*& undo = trx->rsegs.m_noredo.undo) {
 		ut_a(undo->state == TRX_UNDO_PREPARED);
 
