@@ -1,5 +1,5 @@
 /* Copyright (C) 2006 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
-   Copyright (c) 2009, 2022, MariaDB Corporation Ab
+   Copyright (c) 2009, 2021, MariaDB Corporation Ab
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -187,9 +187,9 @@ static MARIA_HA *maria_clone_internal(MARIA_SHARE *share,
       maria_delay_key_write)
     share->delay_key_write=1;
 
-  if (!share->now_transactional)       /* If not transctional table */
+  if (!share->base.born_transactional)   /* For transactional ones ... */
   {
-    /* Pagecache requires access to info->trn->rec_lsn */
+    /* ... force crash if no trn given */
     _ma_set_tmp_trn_for_table(&info, &dummy_transaction_object);
     info.state= &share->state.state;	/* Change global values by default */
   }
@@ -276,13 +276,13 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
   char name_buff[FN_REFLEN], org_name[FN_REFLEN], index_name[FN_REFLEN],
        data_name[FN_REFLEN];
   uchar *UNINIT_VAR(disk_cache), *disk_pos, *end_pos;
-  MARIA_HA info, *UNINIT_VAR(m_info), *old_info= NULL;
+  MARIA_HA info, *UNINIT_VAR(m_info), *old_info;
   MARIA_SHARE share_buff,*share;
   double *rec_per_key_part;
   ulong  *nulls_per_key_part;
   my_off_t key_root[HA_MAX_POSSIBLE_KEY];
   ulonglong max_key_file_length, max_data_file_length;
-  my_bool versioning= 1, born_transactional;
+  my_bool versioning= 1;
   File data_file= -1, kfile= -1;
   struct ms3_st *s3_client= 0;
   S3_INFO *share_s3= 0;
@@ -296,9 +296,9 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
 
 #ifndef WITH_S3_STORAGE_ENGINE
   DBUG_ASSERT(!s3);
-#else
-  if (!s3)
 #endif /* WITH_S3_STORAGE_ENGINE */
+
+  if (!s3)
   {
     realpath_err= my_realpath(name_buff, fn_format(org_name, name, "",
                                                    MARIA_NAME_IEXT,
@@ -327,6 +327,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
   }
 #endif /* WITH_S3_STORAGE_ENGINE */
 
+  old_info= 0;
   if (!internal_table)
     mysql_mutex_lock(&THR_LOCK_maria);
   if ((open_flags & HA_OPEN_COPY) ||
@@ -525,7 +526,6 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
     file_version= (share->state.header.not_used == 0);
     if (file_version == 0)
       share->base.language= share->state.header.not_used;
-    born_transactional= share->base.born_transactional;
 
     share->state.state_length=base_pos;
     /* For newly opened tables we reset the error-has-been-printed flag */
@@ -560,7 +560,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
       We can ignore testing uuid if STATE_NOT_MOVABLE is not set, as in this
       case the uuid will be set in _ma_mark_file_changed().
     */
-    if (born_transactional &&
+    if (share->base.born_transactional &&
         ((share->state.create_trid > trnman_get_max_trid() &&
          !maria_in_recovery) ||
          ((share->state.changed & STATE_NOT_MOVABLE) &&
@@ -570,8 +570,8 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
           (STATE_MOVED | STATE_NOT_ZEROFILLED))))
     {
       DBUG_PRINT("warning", ("table is moved from another system.  uuid_diff: %d  create_trid: %lu  max_trid: %lu  moved: %d",
-                            memcmp(share->base.uuid, maria_uuid,
-                                   MY_UUID_SIZE) != 0,
+                             memcmp(share->base.uuid, maria_uuid,
+                                    MY_UUID_SIZE) != 0,
                              (ulong) share->state.create_trid,
                              (ulong) trnman_get_max_trid(),
                              MY_TEST((share->state.changed & STATE_MOVED))));
@@ -601,7 +601,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
     }
 
     /* Ensure we have space in the key buffer for transaction id's */
-    if (born_transactional)
+    if (share->base.born_transactional)
       share->base.max_key_length= ALIGN_SIZE(share->base.max_key_length +
                                              MARIA_MAX_PACK_TRANSID_SIZE);
 
@@ -700,7 +700,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
 
     share->block_size= share->base.block_size;   /* Convenience */
     share->max_index_block_size= share->block_size - KEYPAGE_CHECKSUM_SIZE;
-    share->keypage_header= ((born_transactional ?
+    share->keypage_header= ((share->base.born_transactional ?
                              LSN_STORE_SIZE + TRANSID_SIZE :
                              0) + KEYPAGE_KEYID_SIZE + KEYPAGE_FLAG_SIZE +
                             KEYPAGE_USED_SIZE);
@@ -723,7 +723,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
         /* Calculate length to store a key + nod flag and transaction info */
         keyinfo->max_store_length= (keyinfo->maxlength +
                                     share->base.key_reflength);
-        if (born_transactional)
+        if (share->base.born_transactional)
           keyinfo->max_store_length+= MARIA_INDEX_OVERHEAD_SIZE;
 
         /* See ma_delete.cc::underflow() */
@@ -862,9 +862,9 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
         file for REPAIR. Don't do logging. This base information will not go
         to disk.
       */
-      born_transactional= FALSE;
+      share->base.born_transactional= FALSE;
     }
-    if (born_transactional)
+    if (share->base.born_transactional)
     {
       share->page_type= PAGECACHE_LSN_PAGE;
       if (share->state.create_rename_lsn == LSN_NEEDS_NEW_STATE_LSNS)
@@ -915,7 +915,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
     }
     else
       share->page_type= PAGECACHE_PLAIN_PAGE;
-    share->now_transactional= born_transactional;
+    share->now_transactional= share->base.born_transactional;
 
     /* Use pack_reclength as we don't want to modify base.pack_recklength */
     if (share->state.header.org_data_file_type == DYNAMIC_RECORD)
