@@ -235,6 +235,11 @@ class ha_rocksdb : public my_core::handler {
   */
   uchar *m_pack_buffer;
 
+  /*
+    A buffer long enough to store table record
+   */
+  uchar *m_record_buffer;
+
   /* class to convert between Mysql format and RocksDB format*/
   std::shared_ptr<Rdb_converter> m_converter;
 
@@ -448,13 +453,15 @@ class ha_rocksdb : public my_core::handler {
     }
   }
 
-  /*
-    MariaDB: this function:
-
-    const char *table_type() const
-
-    is non-virtual in class handler, so there's no point to override it.
-  */
+  /** @brief
+    The name that will be used for display purposes.
+   */
+  const char *table_type() const /*override*/ {
+    DBUG_ENTER_FUNC();
+   // MariaDB: this function is not virtual, however ha_innodb
+   // declares it (and then never uses!) psergey-merge-todo:.
+    DBUG_RETURN(rocksdb_hton_name);
+  }
 
   /* The following is only used by SHOW KEYS: */
   const char *index_type(uint inx) override {
@@ -499,12 +506,6 @@ public:
     DBUG_RETURN(&key_map_full);
   }
 
-  bool primary_key_is_clustered() override {
-    DBUG_ENTER_FUNC();
-
-    DBUG_RETURN(true);
-  }
-
   bool should_store_row_debug_checksums() const {
     return m_store_row_debug_checksums && (rand() % 100 < m_checksums_pct);
   }
@@ -638,7 +639,7 @@ public:
       MY_ATTRIBUTE((__warn_unused_result__));
   int close(void) override MY_ATTRIBUTE((__warn_unused_result__));
 
-  int write_row(uchar *const buf) override
+  int write_row(const uchar *const buf) override
       MY_ATTRIBUTE((__warn_unused_result__));
   int update_row(const uchar *const old_data, const uchar *const new_data) override
       MY_ATTRIBUTE((__warn_unused_result__));
@@ -737,6 +738,7 @@ public:
       const TABLE *old_table_arg, const Rdb_tbl_def *old_tbl_def_arg) const
       MY_ATTRIBUTE((__nonnull__));
 
+  using handler::compare_key_parts;
   int compare_key_parts(const KEY *const old_key,
                         const KEY *const new_key) const
       MY_ATTRIBUTE((__nonnull__, __warn_unused_result__));
@@ -905,8 +907,10 @@ public:
   int check(THD *const thd, HA_CHECK_OPT *const check_opt) override
       MY_ATTRIBUTE((__warn_unused_result__));
   int remove_rows(Rdb_tbl_def *const tbl);
-  ha_rows records_in_range(uint inx, key_range *const min_key,
-                           key_range *const max_key) override
+  ha_rows records_in_range(uint inx,
+                           const key_range *const min_key,
+                           const key_range *const max_key,
+                           page_range *pages) override
       MY_ATTRIBUTE((__warn_unused_result__));
 
   int delete_table(Rdb_tbl_def *const tbl);
